/*! 
\file 
\brief Audio Stream

Interface for working with audio streams
\date 21 November 2003
\author Evgeny Karpov
*/

/*
    $Id: audiostream.c,v 1.7 2003/12/21 22:08:23 juhani Exp $
*/

#include <stdlib.h>
#include <string.h>


#include <sndfile.h> /* GNU library */


#include <errorapi.h>


#include "sdevice.h"
#include "audiostream.h"

/*-----------------------------------------------------------------*/

typedef sf_count_t (*SFREAD) (SNDFILE * sndfile, void  * ptr , sf_count_t frames) ;

/*-----------------------------------------------------------------*/

typedef struct {

    SFREAD rw_func ;
    SNDFILE * file_handle ;
    int end_of_stream ;

} SFileStream ;

/*! Keeps internal stream parameters */
typedef struct {

    AStreamType type ;
    AStreamMode mode ;
    AudioParams audio_params ;

    unsigned int samples ;

    union {

        SFileStream * file_stream ;
        HSDEV device_handle ;
    } handle;

} AudioStream ;

/*======================== Functions ==============================*/

AudioStream * create_empty_stream() 
{
    AudioStream * pstream ;

    /* allocate memory */
    if ((pstream = malloc(sizeof(AudioStream)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* set all parameters to zero */
    memset(pstream,0,sizeof(AudioStream)) ;

    return pstream ;
}

HASTREAM asOpen(AStreamType type, AStreamMode mode,
                 AudioParams * audio_params, char * file_name) 
{
    AudioStream * pstream ;
    SF_INFO sfinfo ;

    if (mode == ASTREAM_MODE_WRITE && audio_params == NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    /* allocate memory */
    if ((pstream = create_empty_stream())==NULL) return NULL ;

    pstream->type = type ;
    pstream->mode = mode ;

    /* open actual stream */
    if (pstream->type == ASTREAM_TYPE_FILE)
    {
        if ((pstream->handle.file_stream = malloc(sizeof(SFileStream)))==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
            free(pstream) ;
            return NULL ;
        }

        /* set all fields to zero */
        memset(pstream->handle.file_stream,0,sizeof(SFileStream)) ;

        if (mode == ASTREAM_MODE_READ)
        {
            sfinfo.format = 0 ; /* this is required by libsndfile */

            pstream->handle.file_stream->file_handle
                = sf_open(file_name,SFM_READ,&sfinfo) ;

            /* fill internal parameters */
            pstream->samples = (unsigned int) sfinfo.frames ;

            pstream->audio_params.codec = sfinfo.format ;
            pstream->audio_params.channels = sfinfo.channels ;
            pstream->audio_params.samples_per_sec = sfinfo.samplerate ;

            /* defaults */
            pstream->audio_params.bits_per_sample = 16 ; 
            pstream->audio_params.sign = 1 ; /* set to signed */
            pstream->handle.file_stream->rw_func = (SFREAD)sf_readf_short ;

            if ((sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_PCM_S8 ||
                (sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_PCM_U8)
            {
                pstream->audio_params.bits_per_sample = 8 ;
                pstream->handle.file_stream->rw_func = sf_read_raw ;

            } else if ((sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_PCM_24)
            {
                pstream->audio_params.bits_per_sample = 24 ;
                pstream->handle.file_stream->rw_func = (SFREAD)sf_readf_int ;

            } else if ((sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_PCM_32 ||
                       (sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_FLOAT)
            {
                pstream->audio_params.bits_per_sample = 32 ;
                pstream->handle.file_stream->rw_func = (SFREAD)sf_readf_int ;
            }

            if ((sfinfo.format & SF_FORMAT_SUBMASK) == SF_FORMAT_PCM_U8)
            {
                pstream->audio_params.sign = 0 ;
            }

            if (audio_params)
                memcpy(audio_params,&(pstream->audio_params),sizeof(AudioParams)) ;

        } else if (mode == ASTREAM_MODE_WRITE)
        {
            memset(&sfinfo,0,sizeof(SF_INFO)) ;

            sfinfo.format = audio_params->codec ;
            sfinfo.channels = audio_params->channels ;
            sfinfo.samplerate = audio_params->samples_per_sec ;

            pstream->handle.file_stream->file_handle = sf_open(file_name,SFM_WRITE,&sfinfo) ;
            pstream->handle.file_stream->rw_func = (SFREAD)sf_writef_short ;

        }

        if (pstream->handle.file_stream->file_handle) return (HASTREAM) pstream ;

        /* report what was the error */
        ReportError(ESR_CANT_OPEN_FILE,(char*)sf_strerror(NULL)) ;

    } else if (pstream->type == ASTREAM_TYPE_DEV) 
    {
        /* TODO: open device stream here */

    } else
    {
    }

    asClose(pstream) ;
    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return NULL ;
}

void asClose(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream)
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            sf_close(pstream->handle.file_stream->file_handle) ;
            free(pstream->handle.file_stream) ;

        } else if (pstream->type == ASTREAM_TYPE_DEV) 
        {
            /* TODO: close device stream here */

        } else
        {
            ReportError(ESR_INVALID_PARAMETER,NULL) ;
        }
    }

    free(pstream) ;
}

/*-----------------------------------------------------------------*/

int asRead(HASTREAM as,void * buf,int samples) 
{
    AudioStream * pstream = (AudioStream*) as ;
    int ret_val ;

    if (pstream)
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            ret_val = (int) pstream->handle.file_stream->rw_func(
                              pstream->handle.file_stream->file_handle,
                              buf, samples) ; 

            /* check for end of stream */
            if (ret_val < samples)
                pstream->handle.file_stream->end_of_stream = 1 ;

            return ret_val ;

        } else 
        {
            /* reading from device stream is not implemented yet */
            ReportError(ESR_NOT_IMPLEMENTED,NULL) ;
        }
    }

    return -1 ;
}

/*-----------------------------------------------------------------*/

int asWrite(HASTREAM as,void * buf,int samples) 
{
    AudioStream * pstream = (AudioStream*) as ;
    unsigned int ret_val ;

    if (pstream)
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            ret_val = (int) pstream->handle.file_stream->rw_func(
                              pstream->handle.file_stream->file_handle,
                              buf, samples) ; 
            pstream->samples += ret_val ;

            return ret_val ;

        } else 
        {
            /* writing to device stream is not implemented yet */
            ReportError(ESR_NOT_IMPLEMENTED,NULL) ;
        }
    }

    return -1 ;
}

/*-----------------------------------------------------------------*/

int asEndOfStream(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) 
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            return pstream->handle.file_stream->end_of_stream ;

        } else if (pstream->type == ASTREAM_TYPE_DEV) 
        {
            /* TODO: add device stream here */
        }
    }

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

int asGetSize(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) 
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            return pstream->samples ;

        } else if (pstream->type == ASTREAM_TYPE_DEV) 
        {
            /* TODO: add device stream here */
        }
    }

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

AStreamType asGetType(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) return pstream->type ;

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

AStreamMode asGetMode(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) 
    {
        return pstream->mode ;
    }

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

int asGetChannels(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) return pstream->audio_params.channels ;

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

int asGetRate(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) return pstream->audio_params.samples_per_sec ;

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

int asGetBitsPerSample(HASTREAM as) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) return pstream->audio_params.bits_per_sample ;

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

int asReadDouble(HASTREAM as,double * buff,int samples) 
{
    AudioStream * pstream = (AudioStream*) as ;

    if (pstream) 
    {
        if (pstream->type == ASTREAM_TYPE_FILE)
        {
            return (int) sf_readf_double(
                           pstream->handle.file_stream->file_handle,
                           buff, samples) ;

        } else if (pstream->type == ASTREAM_TYPE_DEV) 
        {
            /* TODO: add device stream here */
        }
    }

    ReportError(ESR_INVALID_PARAMETER,NULL) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

int asStart(HASTREAM as) 
{
    ReportError(ESR_NOT_IMPLEMENTED,NULL) ;

    return ESR_FAILED ;
}

int asPause(HASTREAM as) 
{
    ReportError(ESR_NOT_IMPLEMENTED,NULL) ;

    return ESR_FAILED ;
}

int asResume(HASTREAM as) 
{
    ReportError(ESR_NOT_IMPLEMENTED,NULL) ;

    return ESR_FAILED ;
}

int asStop(HASTREAM as) 
{
    ReportError(ESR_NOT_IMPLEMENTED,NULL) ;

    return ESR_FAILED ;
}

/*======================== Functions ==============================*/
