/*! 
\file 
\brief Comma separated values parsing 

Comma separated values parsing routines

$Id: csv.c,v 1.9 2003/11/05 22:09:09 juhani Exp $

\date 03 November 2003
\author Evgeny Karpov
*/

/*-----------------------------------------------------------------*/

#include <string.h>
#include <errno.h>

#include <errorapi.h>

#include "csv.h"

/*======================== Functions ==============================*/

/*!
\brief Creates CSV structure

Allocates memory for CSV structure

\param fields_num Number of fields in structure
\return Pointer to CSV structure or NULL in the case of error.
\note Calling function is responsible for freeing this memory.
\sa csvDestroy
\sa CSV
\author Evgeny Karpov
*/
CSV * csvCreate(int fields_num) 
{
	CSV * pcsv ;

    /* create structure */
	pcsv = (CSV *) malloc(sizeof(CSV)) ;

    if (pcsv) 
    {
		pcsv->count = fields_num ;
		pcsv->fields = malloc(sizeof(char *)*fields_num) ;

		if (pcsv->fields)
		{
			memset(pcsv->fields,0,sizeof(char *)*fields_num) ;
			return pcsv ;
		}

    }

    free(pcsv) ; 
    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;

    return NULL ;
}

/*!
\brief Frees memory allocated to CSV structure

Frees memory allocated to CSV structure

\param pcsv Pointer to structure
\return Nothing
\note Pointer passed to a function can be NULL and no error 
will be set in this case
\sa csvCreate
\sa CSV
\author Evgeny Karpov
*/
void  csvDestroy(CSV * pcsv) 
{
    int count ;

    if (pcsv==NULL)
	{
		return ;
	}

    if (pcsv->fields!=NULL)
    {
        for (count=0;count<pcsv->count;count++)
		{
            free(pcsv->fields[count]) ;
		}
    }

	free(pcsv->fields) ;
	free(pcsv) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Parses string into CSV structure

Parses string into CSV structure. Substrings
separated by commas are placed in separate CSV fields.

\param str_buff String to parse
\param size Size of string 
\return Pointer to CSV structure or NULL in the case of error.
Caller is responsible for freeing this memory.

\sa csvWriteLine
\sa CSV
\author Evgeny Karpov
*/
CSV * csvParseLine(char * str_buff,int size) 
{
    CSV * pcsv ;
    char ** tmp_fields ;
    char * tmp_str, * new_field ;
    int count, str_len ;

    /* create structure */
	if ((pcsv = (CSV *) malloc(sizeof(CSV)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    pcsv->count = 0 ;
    pcsv->fields = NULL ;

    tmp_str = str_buff ;

    for (count=0; count < size; count++)
    {
        if (str_buff[count]==',' || count >= size - 1)
        {
            /* add new field */
            tmp_fields = realloc(pcsv->fields,
                                 sizeof(char*)*(pcsv->count+1));

            if (tmp_fields == NULL)
            {
                ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                csvDestroy(pcsv) ;
                return NULL ;
            }

            pcsv->count++ ;
            pcsv->fields = tmp_fields ; 
            pcsv->fields[pcsv->count - 1] = NULL ;

            str_len = count - (tmp_str - str_buff) ;

            /* check if field is empty */
            if (str_len == 0)
            {
                continue ;
            }

            /* allocate memory for a field */
            new_field = malloc(sizeof(char)*(str_len + 1));
            if (new_field == NULL)
            {
                ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                csvDestroy(pcsv) ;
                return NULL ;
            }

            pcsv->fields[pcsv->count - 1] = new_field ;

            /* copy string */
            while (*tmp_str != ',' && tmp_str < str_buff + size)
            {
                *(new_field++) = *(tmp_str++) ;
            }

            /* set end of string */
            pcsv->fields[pcsv->count - 1][str_len] = 0 ;

            /* move string pointer */
            tmp_str = str_buff + count + 1 ;
        }
    }

    return pcsv ;
}

/*!
\brief Writes CSV structure into string

Writes fields of CSV structure into string.
Fields of CSV structure will be separated by
commas in the string.

\param pcsv Pointer to SCV structure
\param str_buff String to write
\param size Size of string
\return Number of characters actually written to a string
or zero in case of error

\pre Array \p str_buff must have sufficient space
for at least \p size characters
\note String size must be sufficient to contain all
fields with commas

\sa csvParseLine
\sa CSV
\author Evgeny Karpov
*/
int csvWriteLine(CSV * pcsv, char * str_buff, int size) 
{
    char * tmp_str ;
    int count,field ;

    tmp_str = str_buff ;
    field = 0 ;

    for (count = 0;
         count < size && field < pcsv->count;
         count++, field++)
    {
        tmp_str = pcsv->fields[field] ;

        if (tmp_str == NULL)
        {
            /* just insert comma and continue */
            str_buff[count] = ',' ;

            continue ;
        }

        /* copy string */
        while (*tmp_str != 0 && count < size) 
        {
            str_buff[count++] = *(tmp_str++) ;
        }

        if (*tmp_str != 0)
        {
            return 0 ; /* buffer size is not enough */
        }

        /* insert comma */
        if (count < size && field < pcsv->count - 1)
        {
            str_buff[count] = ',' ;
        }
    }

    if (count == size)
    {
        return 0 ;
    }

    return count ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Parse line into CSV structure

Parse line into CSV structure without allocating new memory. 
Memory for CSV structure must be allocated before. All pointers
in CSV structure will point to the parts of parsed string. All 
commas will be replaced by zero to mark end of string in fields.

\param pcsv Pointer to CSV structure
\param str_buff String to parse
\param size Size of string
\return Return number of fields parsed from string or zero if 
there is not enough fileds in CSV structure

\note String buffer should have one additional character for the end 
of string symbol

\sa CSV
\author Evgeny Karpov
*/
int csvParseLineNoMem(CSV * pcsv, char * str_buff, int size) 
{
    char * tmp_str ;
    int count, field_count ;

    field_count = 0 ;
    tmp_str = str_buff ;

    for (count=0; count < size; count++)
    {
        if (str_buff[count]==',' || count == size - 1)
        {
            /* check do we have enough fields */
            if (field_count == pcsv->count)
			{
				return 0 ;
			}

            /* change comma to end of string */
            str_buff[count] = 0 ;

            /* add new field */
            field_count++ ;

            pcsv->fields[field_count - 1] = tmp_str ;

            /* move string pointer */
            tmp_str = str_buff + count + 1 ;
        }
    }

    return field_count ;
}

/*======================== Functions ==============================*/
