/*! 
\file fftdct.c
\brief A generic FFT and DCT library implementation,
used by MFCC module.

This is an implementation of FFT and DCT used by MFCC module.
The current implementation uses the shortfft library for
FFT and a direct, not FFT-based computation of the DCT.\n
$Id: fftdct.c,v 1.3 2004/01/05 16:22:41 juhani Exp $
\date 2003/12
\author Juhani Saastamoinen
*/
#include <stdlib.h>
#include <math.h>


#include <fftw.h>


#include <errorapi.h>


#include "fftdct.h"


#if !defined(M_PI)
/*! Define pi if not already defined */
#define M_PI 3.14159265358979323846
#endif


FFTProcessor* fftCreate(int input_size)
{
	FFTProcessor* p;

    if (input_size < 2)
    {
        /* It makes no sense to compute 1 point FFT */
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

	p = malloc(sizeof(FFTProcessor));
    if (NULL == p)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL;
    }

    p->input_size = input_size;
    p->fftw_plan = NULL ;
    p->fftw_in = NULL ;
    p->fftw_out = NULL ;

    /* create FFTW plan */
    p->fftw_plan = fftw_create_plan(
                        p->input_size,
                        FFTW_FORWARD, FFTW_ESTIMATE) ;
    if (NULL == p->fftw_plan)
    {
        /* perhaps new error code here? */
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        fftDestroy(p);
        return NULL ;
    }

    /* allocate working arrays */
    p->fftw_in  = (fftw_complex*) malloc(sizeof(fftw_complex)
                                           * input_size) ;
    if (NULL == p->fftw_in)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        fftDestroy(p);
        return NULL ;
    }

    /* Allocate memory for working arrays */
    p->fftw_out = (fftw_complex*) malloc(sizeof(fftw_complex)
                                           * input_size) ;
    if (NULL == p->fftw_out)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        fftDestroy(p);
        return NULL ;
    }

    return p;
}


void fftDestroy(FFTProcessor* fft)
{
    free(fft->fftw_out) ;
    free(fft->fftw_in) ;
    fftw_destroy_plan(fft->fftw_plan) ;

    free(fft);
}


int fftGetInputSize(FFTProcessor* fft)
{
    if (NULL == fft)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return fft->input_size;
}


int fftCompute(FFTProcessor* fft, double* input, double* spectrum)
{
    int i;

    if ((NULL == fft) || (NULL == input) || (NULL == spectrum))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    for (i = 0; i < fft->input_size; i++)
    {
        fft->fftw_in[i].re = input[i];
        fft->fftw_in[i].im = 0.0;
    }

    fftw_one(fft->fftw_plan, fft->fftw_in, fft->fftw_out);

    for (i = 0; i < fft->input_size / 2 + 1; i++)
    {
        spectrum[i] = c_re(fft->fftw_out[i]) * c_re(fft->fftw_out[i])
                        + c_im(fft->fftw_out[i]) * c_im(fft->fftw_out[i]);
    }

    return ESR_SUCCEED;
}


/*!
\brief Create Discrete Cosine Transformation data structure

The discrete cosine transformation (DCT) of vector \p input.
The DCT of a vector \f$ f \in \mathbb{R}^n\f$ is defined as
\f[
	g_w =
	  \sum_{k=1}^n
	    f_k\,\cos\left({k \over n} \pi \left(k+{1 \over 2}\right)\right).
\f]
\param input Pointer to the input vector data
\param input_size The input signal length
\param output Pointer to the DCT output vector data
\param output_size The size of the DCT output vector
\note The standard DCT output size is the same as the input size
but in some applications it is useful to map the input vector
to a lower dimensional space with a DCT-like manner.
\note There must be at least \p output_size
elements available for writing after the output pointer.
\todo The implementation is not optimal:
Use temporary variables for part of the cosine arguments
or even better: wrap DCT in a struct/class,
compute the cosines once, and use
convolution-like folded pointwise multiplication
to perform the DCT efficiently.
\sa FFT
\author Evgeny Karpov
 */
DCTProcessor* dctCreate(int input_size, int spectrum_size)
{
    int i,j ;
    DCTProcessor* dct;

    if ((input_size < 2) || (spectrum_size < 2)
          || (spectrum_size > input_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

    dct = (DCTProcessor*) malloc(sizeof(DCTProcessor));
    if (NULL == dct)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL;
    }
    dct->input_size = input_size;
    dct->spectrum_size = spectrum_size;

    dct->cos_functions = (double**) malloc(sizeof(double*) * spectrum_size);
    if (NULL == dct->cos_functions)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        dctDestroy(dct);
        return NULL;
    }

    for (i = 0; i < spectrum_size; i++)
    {
        dct->cos_functions[i] = (double*) malloc(sizeof(double) * input_size);
        if (NULL == dct->cos_functions[i])
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            dctDestroy(dct);
            return NULL;
        }
    }

    for (i = 0; i < spectrum_size; i++)
    {
        for (j = 0; j < input_size; j++)
        {
            dct->cos_functions[i][j] = 
              cos(i * (j + 0.5) * (M_PI/input_size));
        }
    }

    return dct;
}


void dctDestroy(DCTProcessor* dct)
{
    int i;

    if (NULL == dct)
    {
        return;
    }

    if (NULL == dct->cos_functions)
    {
        free(dct);
        return;
    }

    for (i = 0; i < dct->spectrum_size; i++)
    {
        free(dct->cos_functions[i]);
    }

    free(dct->cos_functions);

    free(dct);

    return;
}


int dctGetInputSize(DCTProcessor* dct)
{
    if (NULL == dct)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return dct->input_size;
}


int dctGetSpectrumSize(DCTProcessor* dct)
{
    if (NULL == dct)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return dct->spectrum_size;
}


int dctCompute(DCTProcessor* dct, double* input, double* spectrum)
{
    int i, j;

    if ((NULL == dct) || (NULL == input) || (NULL == spectrum))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if ((NULL == dct->cos_functions)
          || (dct->input_size < 2) || (dct->spectrum_size < 2)
          || (dct->spectrum_size > dct->input_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    for (i = 0; i < dct->spectrum_size; i++)
    {
        spectrum[i] = 0;
        for (j = 0; j < dct->input_size; j++)
        {
            spectrum[i] += dct->cos_functions[i][j] * input[j] ;
        }
    }

    return ESR_SUCCEED;
}


