/*!
\file melfilters.c
\brief Mel filter bank implementation

Implementation of mel-filterbank creation.\n
$Id: melfilters.c,v 1.14 2004/01/02 16:09:27 juhani Exp $
\date 2003
\author Evgeny Karpov
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include <errorapi.h>


#include "melfilters.h"


/*-----------------------------------------------------------------*/


double Hz2Mel(double Hz_frequency) ;
double Mel2Hz(double mel_frequency) ;
int    Hz2Index(double frequency, int sampling_rate, int spectrum_size) ;


/*======================== Functions ==============================*/


/*!
\brief Create mel filter bank

Create a filter bank with set of triangular
band pass filters.
\param filters Number of filter in the created filter bank
\param sampling_rate Signal sampling rate
\param spectrum_size Size of magnitude spectra
that shall be processed with the filter bank
\return Pointer to the created filter bank, a \p HFBANK
\todo Use real_filter_begin, real_filter_center, real_filter_end
instead of rounded values when computing the coefficients
in both the floating point and fixed point code
\sa FilterBank
\author Evgeny Karpov
*/
HFBANK * melfbCreate(
	int filters,
	int sampling_rate,
	int fourier_spectrum_size) 
{
    HFBANK fbank ;
    double * coeff ;
    int curr_filter,filter_center,filter_begin,filter_end ;
    int i ;

    double mel_max, mel_freq_Hz, fmel;

    if ((sampling_rate <= 0) || (fourier_spectrum_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }

    /* number of filters must be set inadvance */
    if (filters<=0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }

    /* create filterbank (with averaging) */
    fbank = fbCreate(filters, 1);
    if (NULL == fbank)
    {
        ReportError(ESR_FILTER_BANK_CREATE, NULL);
        return NULL ;
    }

	/* allocate temporary coefficient table */
    coeff = malloc(sizeof(double) * fourier_spectrum_size / 2 + 1);
    if (NULL == coeff)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        fbDestroy(fbank) ;
        return NULL ;
    }

    mel_max = Hz2Mel(sampling_rate/2.0f) ;
    if (mel_max < 0.0)
    {
        ReportError(ESR_HZ2MEL, NULL);
        fbDestroy(fbank) ;
        return NULL ;
    }

    curr_filter = 1;
    filter_begin = -1;
    filter_center = 0;
    fmel = (mel_max * curr_filter) / (filters - 1);
    mel_freq_Hz = Mel2Hz(fmel);
    filter_end = Hz2Index(mel_freq_Hz,
                          sampling_rate,
                          fourier_spectrum_size) ;
    if (filter_end < 0)
    {
        ReportError(ESR_HZ2INDEX, NULL);
        free(coeff);
        fbDestroy(fbank) ;
        return NULL ;
    }
    /*
    printf("fmel = %f, ", fmel);
    printf("mel_freq_Hz = %f, ", mel_freq_Hz);
    printf("filter_end = %d\n", filter_end);
    */

    if (0 == filter_end)
    {
        filter_end = 1;
    }

    for (i = filter_center; i < filter_end ; i++)
    {
        coeff[i - filter_center] = 
          (filter_end - i) / (double) (filter_end - filter_center);
    }
    /*
    fbPrintFilter(stderr, filter_begin + 1, filter_end - 1, coeff);
    */
    i = fbAddFilter(fbank, filter_begin + 1, filter_end - 1, coeff);
    if (ESR_FAILED == i)
    {
        ReportError(ESR_FILTER_BANK_ADD_FILTER, NULL);
        free(coeff) ;
        fbDestroy(fbank) ;
        return NULL;
    }

    for (curr_filter = 2; curr_filter < filters; curr_filter++)
    {
        filter_begin = filter_center;
        filter_center = filter_end;

        fmel = (mel_max * curr_filter) / (filters - 1);
        mel_freq_Hz = Mel2Hz(fmel);
        filter_end    = Hz2Index(mel_freq_Hz,
                                 sampling_rate,
                                 fourier_spectrum_size) ;
        if (filter_end < 0)
        {
            ReportError(ESR_HZ2INDEX, NULL);
            free(coeff);
            fbDestroy(fbank) ;
            return NULL ;
        }
        /*
        printf("fmel = %f, ", fmel);
        printf("mel_freq_Hz = %f, ", mel_freq_Hz);
        printf("filter_end = %d\n", filter_end);
        */

        /* fill array of coefficients */
		if (filter_begin < filter_center)
		{
            for (i = filter_begin + 1; i <= filter_center; i++)
            {
                coeff[i - filter_begin] =
                  (i - filter_begin)
                    / (double) (filter_center - filter_begin);
            }
		}

		if (filter_center < filter_end)
		{
            for (i = filter_center; i < filter_end ; i++)
            {
                coeff[i - filter_begin] = 
                  (filter_end - i)
                    / (double) (filter_end - filter_center);
            }
		}

        /*
        fbPrintFilter(stderr, filter_begin + 1, filter_end - 1, coeff + 1);
        */
        i = fbAddFilter(fbank,
              filter_begin + 1, filter_end - 1, coeff + 1);
        if (ESR_FAILED == i)
        {
            ReportError(ESR_FILTER_BANK_ADD_FILTER, NULL);
            free(coeff) ;
            fbDestroy(fbank) ;
            return NULL;
        }
    }

    filter_begin = filter_center;
    filter_center = filter_end;
    filter_end = fourier_spectrum_size / 2 + 1;
    /*
    printf("filter_begin = %d, ", filter_begin);
    printf("filter_center = %d, ", filter_center);
    printf("filter_end = %d\n", filter_end);
    */

    for (i = filter_begin + 1; i <= filter_center; i++)
    {
        coeff[i - filter_begin] =
          (i - filter_begin) / (double) (filter_center - filter_begin);
    }
    /*
    fbPrintFilter(stderr, filter_begin + 1, filter_end - 1, coeff + 1);
    */
    i = fbAddFilter(fbank, filter_begin + 1, filter_end - 1, coeff + 1);
    if (ESR_FAILED == i)
    {
        ReportError(ESR_FILTER_BANK_ADD_FILTER, NULL);
        free(coeff) ;
        fbDestroy(fbank) ;
        return NULL;
    }

    free(coeff) ;

    return fbank ;
}

/*-----------------------------------------------------------------*/


/*!
\brief Map Hertz frequency to corresponding melody scale frequency

\param Hz_frequency Frequency in Hertz
\return Frequency value on the mel scale (melody scale)
or negative value in case of failure
\author Evgeny Karpov
*/
double Hz2Mel(double Hz_frequency)
{
    if (Hz_frequency < 0.0)
    {
      ReportError(ESR_INVALID_PARAMETER, NULL);
      return -1.0;
    }

    return 2595.0f * (double)log10(1 + Hz_frequency/700.0);
} 

/*!
\brief Map melody scale frequency to corresponding Hertz frequency

\param mel_frequency Frequency value on the mel scale (melody scale)
\return Frequency in Hertz
or negative value in case of failure
\author Evgeny Karpov
*/
double Mel2Hz(double mel_frequency)
{
    return (700.0 * (pow(10, (mel_frequency/2595.0)) - 1));
}


/*!
\brief Index of a given frequency in specturm

Determine the index of a harmonic frequency
in the frequency domain, or spectrum for given
time domain frequency, sampling rate, and
spectrum length.
\param frequency Frequency in Hertz
\param sampling_rate Signal sampling rate
\param spectrum_size Length of spectrum
\return Index of the signal frequency in a spectrum
or negative value in case of failure
\pre frequency must be at most half the sampling_frequency
\bug A comment in code says "round to nearest integer"
but the code assigns \p double to \p int which
means truncation ???
\author Evgeny Karpov
*/
int Hz2Index(
	double frequency,
	int sampling_rate,
	int fourier_spectrum_size)
{
    double Nyquist_freq; /* Nyquist-frequency of the signal */
    double freq_step;    /* Frequency step */

    if ((frequency < 0.0)
          || (sampling_rate <= 0) || (fourier_spectrum_size <= 0))
    {
      ReportError(ESR_INVALID_PARAMETER, NULL);
      return -1;
    }

    Nyquist_freq = sampling_rate/2.0f;  

    if (frequency > Nyquist_freq)
    {
      ReportError(ESR_INVALID_PARAMETER, NULL);
      return -1;
    }

    freq_step = Nyquist_freq/(fourier_spectrum_size / 2.0f);

    /*
    printf("floating point index = %16.12f -> %d\n",
      (frequency / freq_step),
      (int)(frequency / freq_step));
    */

    /* return (int)(frequency / freq_step); */
    return (int)(ceil(frequency / freq_step));
}

/*======================== Functions ==============================*/

/*

  int k, prev_center = 0, next_center, curr_center;
  double mel_max, mel_freq_Hz, fmel;
  Filterbank* filter_bank;

  mel_max = Hz2Mel(sampling_rate/2.0f);   

  filter_bank = allocate(sizeof(Filterbank), NO_INIT);

  filter_bank->num_mel_filters = count_of_filters;
  filter_bank->filters = allocate(sizeof(Filter)*count_of_filters, NO_INIT);

  fmel = 1/((double)count_of_filters) * mel_max;
  mel_freq_Hz = (double)(700 * (pow(10,(fmel/2595.0)) - 1));
  curr_center = Hz2Index(mel_freq_Hz, sampling_rate, length);

  for (k = 1; k <= count_of_filters; k++) {
    fmel = (k + 1)/((double)count_of_filters) * mel_max;
    mel_freq_Hz = (double)(700 * (pow(10,(fmel/2595.0)) - 1));
		
    if (k != count_of_filters)
      next_center = Hz2Index(mel_freq_Hz, sampling_rate, length);
    else
      next_center = curr_center;

    //    printf("%d %d\n", curr_center, next_center-curr_center);

    filter_bank->filters[k - 1].filter_begin = prev_center;
    filter_bank->filters[k - 1].filter_end   = next_center;
    filter_bank->filters[k - 1].coeffs       = 
      FillCoeffTable(prev_center, curr_center, next_center);

    // Fill coefficient array 
    prev_center = curr_center;
    curr_center = next_center;
  }
	
  filter_bank->filters[count_of_filters - 1].filter_end = 
    (int)(length/2.0 - 1);

  return filter_bank;

static double *FillCoeffTable(int begin, int center, int end)
{
  int i;
  double *coeffs;
	
  //  printf("%d %d Size: %d\n", end, begin,(end - begin + 1));
  coeffs = (double *) allocate(sizeof(double)*(end - begin + 1), NO_INIT);
		
  for (i = 1; i <= center - begin; i++)
    coeffs[i - 1] = i/(double)(center - begin);
	
  for (i = 0; i <= end - center; i++)
    coeffs[center - begin + i] = (end - center - i)/(double)(end - center + 1);

  return coeffs;
} 

int NumberOfMelFilters(int sampling_rate)
{
  return (int)ceil((double)(3 * log(sampling_rate)));
} 
*/


