/*!
\file preprocess.c
\brief Signal pre-processing implementation

Implementation of routines for signal
preprocessing before feature extraction\n
$Id: preprocess.c,v 1.5 2004/01/05 14:42:59 juhani Exp $
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>


#include <errorapi.h>


#include "preprocess.h"

/*-----------------------------------------------------------------*/


/*======================== Functions ==============================*/


/*!
\brief Create Pre-Processor

Create a new pre-processor.
\return Pointer to a new Pre-Processor
\sa PreProc
\author Evgeny Karpov
*/
PreProc * ppCreate() 
{
    PreProc * pproc ;

    pproc = malloc(sizeof(PreProc));
    if (NULL == pproc)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
        return NULL ;
    }

    pproc->size = 0 ;
    pproc->funcs = NULL ;

    return pproc ;
}


/*!
\brief Destroy Pre-Processor

Destroy a previously allocated pre-processor.
\param pproc The Pre-Processor that needs to be destroyed
\author Evgeny Karpov
*/
void ppDestroy(PreProc * pproc) 
{
    if (pproc==NULL) return ;

    free(pproc->funcs) ;
    free(pproc) ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Pre-Processor Setup

Add new processing step to a signal pre-processor.
\param pproc The Pre-Processor, where new processing step is added
\param pproc_func Pointer to the processing function
\param func_params Pointer to the place that holds the
parameters for the processing function.
\sa PreProc
\sa PREPROCFUN
\author Evgeny Karpov
*/
int ppAddFunction(PreProc * pproc,PREPROCFUN pproc_func,void * func_params) 
{
    void * tmp ;

    if (pproc==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return ESR_FAILED ;
    }

    tmp = realloc(pproc->funcs,sizeof(PreProcFunc)*(pproc->size+1));
    if (NULL == tmp)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
        return ESR_FAILED ;
    }

    pproc->funcs = tmp ;

    pproc->funcs[pproc->size].func = pproc_func ;
    pproc->funcs[pproc->size].func_params = func_params ;

    pproc->size++ ;

    return pproc->size ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Process the Data

Use all the signal processing steps that have been
set up for the pre-processor \p pproc.
\param pproc Pre-Processor that shall process the input signal
\param input Input signal
\param output Output signal
\param size Size of both signals
\bug The function either needs a temporary buffer
or then the documentation must notify developer
that input and output pointers are always the same
for functions used in processing steps 2 to N 
For example, the difference filtering function dspHighEmphFilter
should not be used with the current implementation!
\sa PreProc
\author Evgeny Karpov
*/
int ppProcess(
    PreProc * pproc,
    double * input,
    double * output,
    int size) 
{
    int ret_val ;
    int i ;

    ret_val = 0 ;

    for (i=0;i<pproc->size;i++)
        if (pproc->funcs[i].func!=NULL)
        {
            ret_val = pproc->funcs[i].func(
                        input, output, size,
                        pproc->funcs[i].func_params) ;
            input = output ;
        }

    return ret_val ;
}

/*======================== Functions ==============================*/
