/*!
\file preprocess.h
\brief Signal pre-processing interface

Interface for working with 
signal preprocessing before feature extraction\n
$Id: preprocess.h,v 1.1 2003/10/31 15:36:53 juhani Exp $
\author Evgeny Karpov
*/

#ifndef PREPROCESS_H
#define PREPROCESS_H

#if defined(__cplusplus)
extern "C" {
#endif

/*-----------------------------------------------------------------*/

/*!
\brief Signal Processing Function Pointer.

All signal pre-processing functions use the same
parameters and return an integer. The parameters
are input signal \p input, output signal \p output,
and the length of both signals \p size.
\sa PreProcFunc
\sa PreProc
*/
typedef int (*PREPROCFUN) (
	double * input,
	double * output,
	int size,
	void * params) ;

/*======================== Structures =============================*/

/*!
\brief Pre-processing Function and Parameters.

A Helper structure that holds a signal processing
function pointer and pointer to the parameters
for that function.
\sa PREPROCFUN
\sa PreProc
*/
typedef struct {
    PREPROCFUN   func ;        /*!< The signal processing function */
    void       * func_params ; /*!< Parameters of the function */
} PreProcFunc ;

/*!
\brief A Signal Pre-Processor

A signal pre-processor consists of one or more
signal processing stages. The number of stages
is defined by \p size. 
\sa PREPROCFUN
\sa PreProcFunc
*/
typedef struct {
    int size ;            /*!< Number of pre-processing functions */
    PreProcFunc * funcs ; /*!< Pre-processing function pointer table */
} PreProc ;

/*======================== Structures =============================*/

/*======================== Functions ==============================*/


/*
Create Pre-Processor
*/
PreProc * ppCreate() ;


/*
Destroy Pre-Processor
*/
void ppDestroy(
	PreProc * pproc) ;


/*
Pre-Processor Setup
*/
int ppAddFunction(
	PreProc * pproc,
	PREPROCFUN pproc_func,
	void * func_params) ;


/*
Process the Data
*/
int ppProcess(
	PreProc * pproc,
	double * input,
	double * output,
	int size) ;

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* PREPROCESS_H */
