/*!
\file winfunc.c
\brief Signal windowing functions implementation

Implementation of several standard signal window functions\n
$Id: winfunc.c,v 1.6 2003/12/01 12:24:30 juhani Exp $
\date 2003
\author Evgeny Karpov
*/
#include <stdlib.h>
#include <math.h>


#include <errorapi.h>


#include "winfunc.h"


#if !defined(M_PI)
/*! Define pi if not already defined */
#define M_PI 3.14159265358979323846
#endif

/*======================== Functions ==============================*/


/*!
\brief Hamming window

Compute symmetric Hamming window function in \f$ n \f$ points.
The Hamming window coefficients are
\f[
        w[k] = 0.54 - 0.46 \cos\left(2\pi\frac{k}{n-1}\right),
	\qquad k \in 0,\dots,n-1.
\f]
\param coeff Pointer to the output vector data
\param win_size The output window function vector length \f$ n \f$
\note There must be at least \p size
elements available for writing in the output pointer.
\author Evgeny Karpov
 */
int wfHamming(double * coeff,int win_size) 
{
    int i ;

    for (i=0;i<win_size;i++)
        coeff[i] = (double)(0.54 - 0.46 *cos(2 * M_PI * i /win_size)) ;

    return win_size ;
}


/*!
\brief Bartlett window

Compute Bartlett window function in \f$ n \f$ points.
The Bartlett window coefficients are
\li for odd \f$ n \f$
\f[
        w[k] =
	\left\{
	\begin{array}{ll}
	\frac{2k}{n-1}, & 0 \leq k \leq \frac{n-1}{2} \\
	2 - \frac{2k}{n-1}, & \frac{n-1}{2} \leq k \leq n-1 \\
	\end{array}
	\right.
	\qquad k=0,\dots,n-1,
\f]
\li for even \f$ n \f$
\f[
        w[k] =
	\left\{
	\begin{array}{ll}
	\frac{2k}{n-1}, & 0 \leq k \leq \frac{n}{2}-1 \\
	\frac{2(n-k-1)}{n-1}, & \frac{n}{2} \leq k \leq n-1 \\
	\end{array}
	\right.
	\qquad k=0,\dots,n-1.
\f]
\param coeff Pointer to the output vector data
\param win_size The output window function vector length \f$ n \f$
\note There must be at least \p size
elements available for writing in the output pointer.
\author Evgeny Karpov
 */
int wfBarlett(double * coeff,int win_size) 
{
    int i ;

    for (i=0;i<win_size;i++)
        if (i <= (win_size / 2.0))
            coeff[i] = (double)(2.0f * i) / win_size;
        else 
            coeff[i] = (double) 2.0f - (2.0f * i)/win_size;

    return win_size ;
}


/*!
\brief Hanning window

Compute symmetric Hanning window function in \f$ n \f$ points.
The Hanning window is
\f[
        w[k] =
	  \frac{1}{2}
	  \left(1 - \cos \left(2\pi\frac{k}{n-1}\right)\right),
	\qquad k \in 0,\dots,n-1.
\f]
\param coeff Pointer to the output vector data
\param win_size The output window function vector length \f$ n \f$
\note There must be at least \p size
elements available for writing in the output pointer.
\author Evgeny Karpov
 */
int wfHanning(double * coeff,int win_size) 
{
    int i ;

    for (i=0;i<win_size;i++)
        coeff[i] = (double)(0.5 - 0.5 * cos(2 * M_PI * i / win_size));

    return win_size ;
}


/*!
\brief Blackman window

Compute symmetric Blackman window function in \f$ n \f$ points.
The Blackman window is
\f[
        w[k] =
	  0.42 - 0.5\cos \left(2\pi\frac{k}{n-1}\right)
	    + 0.08\cos \left(4\pi\frac{k}{n-1}\right),
	\qquad k \in 0,\dots,n-1.
\f]
\param coeff Pointer to the output vector data
\param win_size The output window function vector length \f$ n \f$
\note There must be at least \p size
elements available for writing in the output pointer.
\author Evgeny Karpov
 */
int wfBlackman(double * coeff,int win_size) 
{
    int i ;

    for (i=0;i<win_size;i++)
        coeff[i] = (double)(0.42 - 0.5 * cos(2 * M_PI * i / win_size)
                    + 0.08 * cos(4 * M_PI * i / win_size));

    return win_size ;
}


/*!
\brief Gaussian window

Compute Gaussian in \f$ n \f$ points.
The Gauss window is
\f[
        w[k] =
	  \mathrm{e}^{\frac{1}{2}\left(\alpha\frac{k-N/2}{N/2}\right)^2},
	\qquad k \in 0,\dots,n,\ \alpha \geq 2.
\f]
\param coeff Pointer to the output vector data
\param win_size The output window function vector length \f$ n \f$
\todo Implement it.
\note There must be at least \p size
elements available for writing in the output pointer.
\author Evgeny Karpov
 */
int wfGaussian(double * coeff, int win_size) 
{
    ReportError(ESR_NOT_IMPLEMENTED,
                "wfGaussian (Gaussian window)");
    coeff = NULL;
    win_size = 0;
    return 0 ;
}


/*======================== Functions ==============================*/
