/*! 
\file
\brief Error message framework implementation

Error and logging processing mechanism. All libraries must use this
module to report errors or to log events. Provides also various 
error codes. Default error and logging handling may be overriden.

$Id: errorapi.c,v 1.24 2003/12/31 12:59:03 juhani Exp $

\date 03 November 2003
\author Juhani Saastamoinen
*/

/*-----------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*-----------------------------------------------------------------*/

#include "errorapi.h"

/*-----------------------------------------------------------------*/

/*!
\brief Error description that cannot be found

When the function GetErrorString cannot
find error message string for given error code
then this string is returned instead.
 */
#define ESR_UNKNOWN_ERROR       "Unknown error"

/*======================== Structures =============================*/

/*!
\brief Error message
 */
typedef struct {

	ErrorCode code; /*!< Error code */
	char * descr;   /*!< Error description in words */

} ErrMsg;

/*!
\brief Error description array

Global array of ErrMsg structs for storing
descriptions of all possible errors. If new error
code is needed and added then description
of the new error code should be added here.
\sa ErrorCode
\sa ErrMsg
 */
static ErrMsg ErrMsgList[] =
{

    /* basic error codes for all modules */

    { ESR_FATAL_ERROR,
	  "Fatal error" },

	{ ESR_INVALID_PARAMETER,
	  "Function parameter has invalid value" },

	{ ESR_NOT_ENOUGH_MEMORY,
	  "Not enough memory to continue" },

	{ ESR_NOT_IMPLEMENTED,
	  "The called function exists but implementation is not ready" },

	{ ESR_CANT_OPEN_FILE,
	  "Can't open file" },

	{ ESR_CANT_LOAD_FILE,
	  "Can't load file" },

	{ ESR_INVALID_FILE_FORMAT,
	  "Invalid file format" },     

    /* modeling specific errors  */

    { ESR_VECTOR_CREATE,
    "Feature vector creating failed" },

    { ESR_FFT_CREATE,
    "Creating Fast Fourier transformation processor failed" },

    { ESR_FFT_COMPUTE,
    "FFT processor failed in FFT computation" },
   
    { ESR_FFT,
    "FFT algorithm used by the FFT processor failed" },

    { ESR_DCT_CREATE,
    "Creating Discrete Cosine Tranformation processor failed" },

    { ESR_DCT_COMPUTE,
    "DCT processor failed in DCT computation" },
   
    { ESR_DCT,
    "DCT algorithm used by the DCT processor failed" },

    { ESR_LOG2,
    "Base-2 logarithm function failed" },

    { ESR_SQRT,
    "Square root function failed" },

    { ESR_DISTANCE_FUNCTION,
    "Vector distance function failed" },

    { ESR_VECTOR_SET_CREATE,
    "Feature vector set creating failed" },

    { ESR_VECTOR_SET_ADD,
    "Adding a vector to a feature vector set failed" },

    { ESR_VECTOR_SET_GET,
    "Getting pointer to given vector in a vector set failed" },

    { ESR_MSINDEX_CREATE_FUNCTION,
    "Metric space index creating function has failed" },

    { ESR_MSINDEX_SEARCH_FUNCTION,
    "Metric space indexing search function has failed" },

    { ESR_MSINDEX_NN_SEARCH_FUNCTION,
    "Metric space indexing nearest neighbour search function has failed" },

    { ESR_SILENCE_REMOVAL_TRAIN,
    "Silence removal training function has failed" },

    { ESR_SILENCE_REMOVAL_DETECT,
    "Silence removal detection function has failed" },

    { ESR_FILTER_BANK_CREATE,
    "Filter bank creating failed" },

    { ESR_FILTER_BANK_ADD_FILTER,
    "Adding a new filter in a filter bank has failed" },

    { ESR_FILTER_BANK_CHECK,
    "Filter bank checking failed" },

    { ESR_FILTER_BANK_SIZE,
    "Filter bank size query failed" },

	{ ESR_FILTER_BANK_APPLY,
	"Applying filter bank to a spectrum failed" },

    { ESR_FILTER_BANK_APPLY_ONE,
    "Applying a single filter of filter bank to a spectrum failed" },

	{ ESR_FILTER_BANK_PRINT,
	"Filter bank printing failed" },

	{ ESR_FILTER_BANK_PRINT_FILTER,
	"Filter bank filter printing failed" },

    { ESR_FILTER_CREATE,
    "Creating a new filter failed" },

    { ESR_HZ2MEL,
    "Hertz to mel frequency conversion failed" },

    { ESR_HZ2INDEX,
    "Hertz to frequency index conversion failed" },

    { ESR_MEL_FILTER_BANK_CREATE,
    "Creating mel filter bank failed" },

	{ ESR_MODEL_CONSTR_FAIL,
    "Model construction failed" },

    { ESR_VECTOR_QUANTIZER_CREATE,
    "Creating vector quantizer failed" },

    { ESR_VECTOR_QUANTIZER_TRAIN,
    "Training vector quantizer failed" },

    { ESR_GLA_COPY_RANDOM_VECTORS,
    "Computation of random vectors for initial solution has failed" },

    { ESR_GLA_RUN,
    "GLA clustering algorithm has failed" },

	{ ESR_GLA_CENTROID_COMPUTATION,
    "Computation of new centroids has failed during GLA iteration" }

};

/*======================== Structures =============================*/

int ReportErrorDefault(int error_code,char* error_description,char* file,int line) ;
int LogEventDefault(char* event_description) ;

/*-----------------------------------------------------------------*/

/*!
\brief Pointer to error handler function

Developers may set this variable to point
to their preferred error handler function.
 */
ERRFN ReportError = ReportErrorDefault;

/*!
\brief Pointer to event logging function

Developers may set this variable to point
to their preferred event logging function.
 */
LOGFN LogEvent = LogEventDefault;

/*======================== Functions ==============================*/
int SetReportError(ERRFN err_func) 
{
    ReportError = err_func ;
    
    return 0 ;
}

/*!
\brief Default error handler

Provides default error handler realization. By default
just prints error message to standard error output

\param error_code Error code
\param error_description Caller description of error
\param file File name where error handler was called
\param line Line number where error handler was called

\return Always returns zero

\note To override default error handler mechanism assign
new error handler pointer to variable ReportError. New handler
must be of type ERRFN
\sa ERRFN ReportError
\author Juhani Saastamoinen
*/
int ReportErrorDefault(int error_code,char* error_description,char* file,int line)
{
	fprintf(stderr, "%s (%d): ", file, line);
	fprintf(stderr, "error %d (%s)\n",error_code,GetErrorString(error_code));

	if (NULL != error_description)
	{
		fprintf(stderr, ":\n%s\n", error_description);
	}

	return 0;
}

/*!
\brief Default logging handler

Provides default logging handler realization. By default
just prints event to standard output.

\param event_description Event description supplied by the caller
\return Always returns zero

\note To override default logging handler mechanism assign
new logging handler pointer to variable LogEvent. New handler
must be of type LOGFN

\sa LOGFN LogEvent
\author Juhani Saastamoinen
*/
int LogEventDefault(char* event_description)
{
	if (NULL == event_description)
	{
		return 0;
	}

	fprintf(stdout, "%s\n", event_description);

	return 0;
}

/*!
\brief Default error descriptions

Provides default description for error codes defined in 
errorapi module

\param error_code Error code

\return Textual description of error code
\todo Add mechanism to override default descriptions
\author Juhani Saastamoinen
*/
char* GetErrorString(int error_code)
{
	int i;

	for (i = 0; i < sizeof(ErrMsgList)/sizeof(ErrMsg); i++)
	{
		if (error_code == ErrMsgList[i].code)
		{
			return ErrMsgList[i].descr;
		}
	}

	return ESR_UNKNOWN_ERROR ;
}


/*======================== Functions ==============================*/

