/*! 
\file
\brief Error message framework implementation

Error and logging processing mechanism. All libraries must use this
module to report errors or to log events. Provides also various 
error codes. Default error and logging handling may be overriden.

$Id: errorapi.h,v 1.25 2003/12/31 12:59:03 juhani Exp $

\date 03 November 2003
\author Juhani Saastamoinen
*/

#ifndef __ERRORAPI_H__
#define __ERRORAPI_H__

/*-----------------------------------------------------------------*/

#if defined(__cplusplus)
extern "C" {
#endif

/*-----------------------------------------------------------------*/

/*! \brief Error handling function type */
typedef int (*ERRFN) (
	int error_code,
	char* descr,
	char* file,
	int line);

/* variable to keep error handler */
extern ERRFN ReportError;

/*!
\brief Macro to report errors easily

Macro for reporting errors without having to specify
source code file and line every time.
 */
#define ReportError(code, descr) \
	ReportError((code), (descr), __FILE__, __LINE__)

int SetReportError(ERRFN err_func) ;


/*! \brief Logging function type */
typedef int (*LOGFN) (
	char* descr);

/* variable to keep logging handler */
extern LOGFN LogEvent;

/*-----------------------------------------------------------------*/

/*!
\brief Function return status
 */
typedef enum {

    ESR_FAILED,  /*!< Processing failed */
    ESR_SUCCEED  /*!< Processing was successfully completed */

} RetStatus ;

/*!
\brief Error code
 */
typedef enum {

    /* basic error codes for all modules */
	ESR_FATAL_ERROR,       /*!< Fatal error */
    ESR_INVALID_PARAMETER, /*!< Invalid parameter has been
                                given to a function */
	ESR_NOT_ENOUGH_MEMORY, /*!< Memory allocation error */
    ESR_NOT_IMPLEMENTED,   /*!< Called function is not implemented */
	ESR_CANT_OPEN_FILE,    /*!< File cannot be opened */
	ESR_CANT_LOAD_FILE,    /*!< File cannot be loaded */

	ESR_INVALID_FILE_FORMAT, /*!< File format is invalid */

    /* modeling specific errors  */
    ESR_VECTOR_CREATE,      /*!< Feature vector creating failed */
    ESR_FFT_CREATE,      /*!< Creating FFT processor failed  */
    ESR_FFT_COMPUTE,      /*!< FFT processor failed in computation */
    ESR_FFT,             /*!< Internally used FFT algorithm failed */
    ESR_DCT_CREATE,      /*!< Creating DCT processor failed  */
    ESR_DCT_COMPUTE,      /*!< DCT processor failed in computation */
    ESR_DCT,             /*!< Internally used DCT algorithm failed */
    ESR_LOG2,            /*!< Base-2 logarithm function failed */
    ESR_SQRT,            /*!< Square root function failed */
    ESR_DISTANCE_FUNCTION,  /*!< Vector distance function failed */
    ESR_VECTOR_SET_CREATE,      /*!< Feature vector set
                                     creating failed */
    ESR_VECTOR_SET_ADD, /*!< Adding a vector to
                             a feature vector set failed */
    ESR_VECTOR_SET_GET, /*!< Getting pointer to given vector
                             in a vector set failed */
    ESR_MSINDEX_CREATE_FUNCTION, /*!< Metrix space index creating
                                  function has failed */
    ESR_MSINDEX_SEARCH_FUNCTION, /*!< Metrix space indexing search
                                  function has failed */
    ESR_MSINDEX_NN_SEARCH_FUNCTION, /*!< Metrix space indexing nearest
                                     neighbour search function
                                     has failed */
    ESR_SILENCE_REMOVAL_TRAIN, /*!< Silence removal training
                                    function has failed */
    ESR_SILENCE_REMOVAL_DETECT, /*!< Silence removal detection
                                     function has failed */
    ESR_FILTER_BANK_CREATE, /*!< Filter bank creating failed */
    ESR_FILTER_BANK_ADD_FILTER, /*!< Adding a new filter in
                                     a filter bank has failed */
    ESR_FILTER_BANK_CHECK,  /*!< Filter bank checking failed */
    ESR_FILTER_BANK_SIZE,   /*!< Filter bank size query failed */
    ESR_FILTER_BANK_APPLY, /*!< Applying filter bank to a spectrum failed */
    ESR_FILTER_BANK_APPLY_ONE, /*!< Applying a single filter
                                    of filter bank to a spectrum failed */
    ESR_FILTER_BANK_PRINT, /*!< Filter bank printing failed */
    ESR_FILTER_BANK_PRINT_FILTER, /*!< Filter bank filter printing failed */
    ESR_FILTER_CREATE,      /*!< Creating a new filter failed */
    ESR_HZ2MEL, /*!< Hertz to mel frequency conversion failed */
    ESR_HZ2INDEX, /*!< Hertz to frequency index conversion failed */
    ESR_MEL_FILTER_BANK_CREATE, /*!< Creating mel filter bank failed */
	ESR_MODEL_CONSTR_FAIL,  /*!< Model cannot be constructed */
    ESR_VECTOR_QUANTIZER_CREATE, /*!< Creating vector quantizer
                                      failed */
    ESR_VECTOR_QUANTIZER_TRAIN, /*!< Training vector quantizer
                                      failed */
    ESR_GLA_COPY_RANDOM_VECTORS,  /*!< Computation of random vectors for
                                       initial solution has failed */
    ESR_GLA_RUN, /*!< GLA clustering algorithm has failed */
    ESR_GLA_CENTROID_COMPUTATION  /*!< Computation of new centroids
                                       has failed during GLA iteration */
} ErrorCode;


/*======================== Functions ==============================*/

/* returns readable description of error */
char* GetErrorString(int error_code);

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* __ERRORAPI_H__ */

