/*! 
\file fixmath.c
\brief Implementation of math library functions in fixed point arithmetic.

This modules implements some functions most commonly seen
in any standard C-library. The functions are implemented
using fixed point arithmetic only scaling the variables
appropriately to minimize relative error.\n
$Id: fixmath.c,v 1.31 2004/01/05 13:24:14 juhani Exp $
\date 2003/11
\author Juhani Saastamoinen
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>


#include "fixmath.h"


/*
   Define FRAC_METHOD to be either
     FRAC_LOOKUP: evaluate log2(x), 1 <= x < 2,
	              with help of a lookup table and linear interpolation
     FRAC_FACTOR: evaluate log2(x), 1 <= x < 2,
	              using factoring, no multiple factors
 */
#define FRAC_LOOKUP  0
#define FRAC_FACTOR  1
#ifndef FRAC_METHOD
#define FRAC_METHOD FRAC_LOOKUP
#endif

/*
   When interpolating the logarithm value use also
   derivative interpolation
 */
/* #define FRAC_D1LOOKUP */


#define DBL_FRAC_ONE ((FLOAT) FRAC_ONE)


double frac_dbl(INT32 x, UINT32 f)
{
	return (x + (f / DBL_FRAC_ONE));
}

const char* frac_str(INT32 x, UINT32 f)
{
	static char buf[64];

	sprintf(buf, "%ld.%7.7lu",
	  x, (unsigned long int)(10000000.0 * (f / DBL_FRAC_ONE)));

	return buf;
}


/* compute y.f = x1.f1 + x2.f2 */
INT16 frac_sum(
	INT32 x1, UINT32 f1,
	INT32 x2, UINT32 f2,
	INT32* y, UINT32* f)
{
	UINT32 tmp;

#if defined(DEBUG_PRINT)
	printf("frac_sum(%ld, %lu, %ld, %lu, y, f):\n", x1, f1, x2, f2);
#endif

	if ((NULL == y) || (NULL == f))
	{
		fprintf(stderr,	"frac_sum: fraction output pointer in NULL\n");
		return -1;
	}
	if (FRAC_ONE <= f1 || FRAC_ONE <= f2)
	{
		fprintf(stderr,	"frac_sum: fraction input >= 1\n");
		return -1;
	}

	tmp = x1 + x2;
	if (tmp >= LONG_MAX)
	{
		fprintf(stderr,	"frac_sum: overflow\n");
		return -1;
	}

	*f = f1 + f2;
	*y = tmp;
	if (FRAC_ONE <= *f)
	{
		*f -= FRAC_ONE;
		if (tmp >= LONG_MAX - 1)
		{
			fprintf(stderr,	"frac_sum: overflow\n");
			return -1;
		}
		*y += 1;
	}

	return 0;
} /* frac_sum */


/* compute y.f = x1.f1 * x2.f2 */
INT16 frac_mul(
	INT32 x1, UINT32 f1,
	INT32 x2, UINT32 f2,
	INT32* y, UINT32* f)
{
#if defined(MUL_DEBUG_PRINT)
	printf("frac_mul(%ld, %lu, %ld, %lu, y, f):\n", x1, f1, x2, f2);
	printf("  %16.10f * %16.10f should be %16.10f\n",
	  frac_dbl(x1, f1),
	  frac_dbl(x2, f2),
	  frac_dbl(x1, f1) * frac_dbl(x2, f2));
#endif

	if ((NULL == y) || (NULL == f))
	{
		fprintf(stderr,	"frac_mul: fraction output pointer in NULL\n");
		return -1;
	}
	if (FRAC_ONE <= f1 || FRAC_ONE <= f2)
	{
		fprintf(stderr,	"frac_mul: fraction input >= 1\n");
		fprintf(stderr,	"  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n",
			x1, f1, x2, f2);
		return -1;
	}

	if ((FRAC_ONE <= x1) || (FRAC_ONE <= x2))
	{
		fprintf(stderr,	"frac_mul: overflow\n");
		return -1;
	}
	
#if defined(MUL_DEBUG_PRINT)
	printf("  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n", x1, f1, x2, f2);
#endif

	*y = (x1 * x2) + ((x1 * f2 + x2 * f1) / FRAC_ONE);

#if defined(MUL_DEBUG_PRINT)
	printf("  x1 * x2 = %ld, (x1 * f2 + x2 * f1) / FRAC_ONE = %ld, *y = %ld\n",
		x1 * x2, (x1 * f2 + x2 * f1) / FRAC_ONE, *y);
#endif

	*f = ((x1 * f2 + x2 * f1) % FRAC_ONE) + ((f1 * f2) / FRAC_ONE);
	if (*f >= FRAC_ONE)
	{
		*f -= FRAC_ONE;
		*y += 1;
	}

#if defined(MUL_DEBUG_PRINT)
	printf("  ((x1 * f2 + x2 * f1) %% FRAC_ONE) = %ld\n",
		((x1 * f2 + x2 * f1) % FRAC_ONE));
	printf("  ((f1 * f2) / FRAC_ONE) >> 1 = %ld\n",
		((f1 * f2) / FRAC_ONE));
#endif

	return 0;
} /* frac_mul */


/* compute y.f = x1.f1 / x2.f2 for x2.f2 >= 1 */
INT16 frac_div(
	INT32 x1, UINT32 f1,
	INT32 x2, UINT32 f2,
	INT32* y, UINT32* f)
{
	INT32 n, d;
#if 0
	UINT32 fn, fd;
#endif

#if defined(DIV_DEBUG_PRINT)
	double real_n, real_d;

	printf("frac_div(%ld, %lu, %ld, %lu, y, f):\n", x1, f1, x2, f2);
	printf("  %16.10f / %16.10f should be %16.10f\n",
	  frac_dbl(x1, f1),
	  frac_dbl(x2, f2),
	  frac_dbl(x1, f1) / frac_dbl(x2, f2));
#endif

	if ((NULL == y) || (NULL == f))
	{
		fprintf(stderr,	"frac_div: fraction output pointer in NULL\n");
		return -1;
	}
	if ((FRAC_ONE <= f1) || (FRAC_ONE <= f2))
	{
		fprintf(stderr,	"frac_div: fraction input >= 1\n");
		fprintf(stderr,	"  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n",
			x1, f1, x2, f2);
		return -1;
	}

	if ((0 == x2) && (0 == f2))
	{
		fprintf(stderr,	"frac_div: division by zero\n");
		return -1;
	}

#if defined(DIV_DEBUG_PRINT)
	printf("  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n", x1, f1, x2, f2);
#endif


	/* brute force approach */
	if (((x1 > (2*FRAC_ONE - 1)) && (x2 == 0))
	      || ((x1 == 0) && (x2 > (2*FRAC_ONE - 1))))
	{
		fprintf(stderr,	"frac_div: overflow\n");
		return -1;
	}
	while ((x1 > (2*FRAC_ONE - 1)) || (x2 > (2*FRAC_ONE - 1)))
	{
		f1 = f1 / 2 + (x1 & 1) * (FRAC_ONE >> 1);
		x1 = x1 / 2;
		f2 = f2 / 2 + (x2 & 1) * (FRAC_ONE >> 1);
		x2 = x2 / 2;
#if defined(DIV_DEBUG_PRINT)
		printf("  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n", x1, f1, x2, f2);
#endif
	}

	n = (x1 * FRAC_ONE) + f1;
	d = (x2 * FRAC_ONE) + f2;
#if defined(DIV_DEBUG_PRINT)
	printf("  n = %ld, d = %ld\n", n, d);
#endif
	while ((n % d) > ((1 << (31 - FRAC_NBITS)) - 1))
	{
		n = n / 2;
		d = d / 2;
#if defined(DIV_DEBUG_PRINT)
		printf("    n = %ld, d = %ld (avoiding overflow)\n", n, d);
#endif
	}
	*y = n / d;
	*f = ((n % d) * FRAC_ONE) / d;
#if defined(DIV_DEBUG_PRINT)
	printf("  final result is %16.10f\n",
	  frac_dbl(*y, *f));
#endif
	return 0;


#if 0
	while (x2 < 2)
	{
		if (x1 > (LONG_MAX >> 1))
		{
			fprintf(stderr, "frac_div: overflow\n");
			return -1;
		}
		x1 = (x1 << 1) + (((f1 << 1) & FRAC_ONE) != 0);
		f1 = (f1 << 1) & (FRAC_ONE - 1);
		x2 = (x2 << 1) + (((f2 << 1) & FRAC_ONE) != 0);
		f2 = (f2 << 1) & (FRAC_ONE - 1);

#if defined(DIV_DEBUG_PRINT)
		printf("  made x2 at least 2; ");
		printf("  x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n", x1, f1, x2, f2);
#endif
	}

	if (x2 > (FRAC_ONE >> 1))
	{
		/* round fraction */
		if (f2 >= (FRAC_ONE >> 1))
		{
			x2 += 1;
		}
		f2 = 0;
	}

	if (f2 == 0)
	{
		/* the case where demoninator is an integer */

#if defined(DIV_DEBUG_PRINT)
		printf("  fractional / integer\n");
#endif

		while ((x1 % x2) > FRAC_ONE)
		{
#if defined(DIV_DEBUG_PRINT)
			printf("  x1 = %ld, f1 = %lu, x2 = %ld; ", x1, f1, x2);
			printf(" avoiding multiplication overflow ... ");
#endif
			/* multiplication with FRAC_ONE may overflow;
			 * loose some precision before the operation */
			f1 = (f1 / 2) + (x1 & 1) * (FRAC_ONE >> 1);
			x1 = x1 / 2;
			x2 = x2 / 2;
#if defined(DIV_DEBUG_PRINT)
			printf("x1 = %ld, f1 = %lu, x2 = %ld\n", x1, f1, x2);
#endif
		}

#if defined(DIV_DEBUG_PRINT)
		printf("  x1 = %ld, f1 = %lu, x2 = %ld\n", x1, f1, x2);
#endif

		*y = x1 / x2;
		*f = ((x1 % x2) * FRAC_ONE) / x2 + (f1 / x2);

		if (*f >= FRAC_ONE)
		{
			*f -= FRAC_ONE;
			*y += 1;
		}

#if defined(DIV_DEBUG_PRINT)
		printf("  final result is %16.10f\n",
		  frac_dbl(*y, *f));
#endif

		return 0;
	}


	/* general case */

#if defined(DIV_DEBUG_PRINT)
		printf("  general case;\n");
#endif

	if (x2 > 2*f2)
	{
		/* fractional part can be ignored, assumes 16 bits for both */

#if defined(DIV_DEBUG_PRINT)
		printf("  x2 = %ld; rounding fractional part %lu\n", x2, f2);
#endif

		if (f2 >= (FRAC_ONE >> 1))
		{
			x2 += 1;
		}

		n = frac_div(x1, f1, x2, 0, y, f);
		if (n != 0)
		{
			fprintf(stderr,	"frac_div: recursive call failed\n");
			return -1;
		}

#if defined(DIV_DEBUG_PRINT)
		printf("  final result is %16.10f\n",
		  frac_dbl(*y, *f));
#endif
		return 0;
	}


	/* avoid overflow */
	while ((x1 > FRAC_ONE) || (x1 < (2*x2)))
	{
#if defined(DIV_DEBUG_PRINT)
		printf("  x1 = %ld, x2 = %ld; ", x1, x2);
		printf(" avoiding multiplication overflow ...");
#endif
		/* multiplication with FRAC_ONE may overflow;
		 * loose some precision before the operation */
		f1 = (f1 / 2) + (x1 & 1) * (FRAC_ONE >> 1);
		x1 = x1 / 2;

		f2 = (f2 / 2) + (x2 & 1) * (FRAC_ONE >> 1);
		x2 = x2 / 2;

#if defined(DIV_DEBUG_PRINT)
		printf("x1 = %ld, f1 = %lu, x2 = %ld, f2 = %lu\n", x1, f1, x2, f2);
#endif

		if ((0 == x1) || (0 == x2))
		{
			fprintf(stderr,	"frac_div: underflow or overflow\n");
			return -1;
		}
	}

	n = (x1 * x2) + (f1 * x2) / FRAC_ONE - (x1 * f2) / FRAC_ONE;
	fn = (f1 * x2) % FRAC_ONE;
#if defined(DIV_DEBUG_PRINT)
	printf("  n = %ld, fn = %lu; fraction underflow?\n", n, fn);
#endif
	if (fn < (((x1 * f2) % FRAC_ONE) + (f1 * f2) / FRAC_ONE))
	{
#if defined(DIV_DEBUG_PRINT)
		printf("    yes: ((x1 * f2) %% FRAC_ONE) = %lu, ",
		  ((x1 * f2) % FRAC_ONE));
		printf("((f1 * f2) / FRAC_ONE) = %lu\n",
		  ((f1 * f2) / FRAC_ONE));
#endif
		fn += FRAC_ONE;
		n = n - 1;
	}
#if defined(DIV_DEBUG_PRINT)
	else
	{
		printf("    no.\n");
	}
	printf("    => n = %ld, fn = %lu\n", n, fn);
#endif

	fn = fn - ((x1 * f2) % FRAC_ONE);

#if defined(DIV_DEBUG_PRINT)
	printf("  n = %ld, fn = %lu; fraction underflow twice?\n", n, fn);
#endif
	if (fn < ((f1 * f2) / FRAC_ONE))
	{
#if defined(DIV_DEBUG_PRINT)
		printf("    yes: ((f1 * f2) / FRAC_ONE) = %lu\n",
		  ((f1 * f2) / FRAC_ONE));
#endif
		fn += FRAC_ONE;
		n = n - 1;
	}
#if defined(DIV_DEBUG_PRINT)
	else
	{
		printf("    no.\n");
	}
	printf("    => n = %ld, fn = %lu\n", n, fn);
#endif

	fn = fn - ((f1 * f2) / FRAC_ONE);

#if defined(DIV_DEBUG_PRINT)
	printf("  n = %ld, fn = %lu\n", n, fn);
#endif

	d = (x2 * x2);
	fd = (f2 * f2) / FRAC_ONE;
	if (fd != 0)
	{
		fd = FRAC_ONE - fd;
		d = d - 1;
	}

	if (n > 2*FRAC_ONE)
	{
		fn = 0;
	}

	if (d > FRAC_ONE)
	{
		/*
		 * Avoid overflow in recursive call
		 * by falling back to division with integer.
		 * The accuracy is already at least FRAC_NBITS - 1 bits.
		 */
		if (fd >= (FRAC_ONE >> 1))
		{
			d += 1;
		}
		fd = 0;
	}
	else if (d >= fd)
	{
		if (fd < (1 << 1))
		{
			fd = 0;
		}
		else if (FRAC_ONE - fd < (1 << 1))
		{
			d += 1;
			fd = 0;
		}
	}

#if defined(DIV_DEBUG_PRINT)
	real_n = 
	  frac_dbl(x1, f1) * x2
	    - (frac_dbl(x1, f1) * f2) / DBL_FRAC_ONE;
	printf("  x1.f1 * x2.000 - x1.f1 * 0.f2 = %16.10f (%16.10f)\n",
	  frac_dbl(n, fn), real_n);

	real_d = (x2 * x2);
	if ((f2 * f2) / FRAC_ONE > 0)
	{
		real_d = real_d - ((f2 * f2) / FRAC_ONE) / DBL_FRAC_ONE;
	}
	printf("  (x2.000)^2 - (0.f2)^2 = %16.10f (%16.10f)\n",
	  frac_dbl(d, fd), real_d);

	printf("  now it looks like the result would be %16.10f\n",
	  frac_dbl(n, fn) / frac_dbl(d, fd));
#endif

	n = frac_div(n, fn, d, fd, y, f);
	if (n != 0)
	{
		fprintf(stderr,	"frac_div: recursive call failed\n");
		return -1;
	}

#if defined(DIV_DEBUG_PRINT)
	printf("  final result is %16.10f\n",
	  frac_dbl(*y, *f));
#endif

	return 0;
#endif
} /* frac_div */


INT16 factorial(INT16 i, UINT32 *f)
{
	static UINT32 fact_table[13] =
	{
		1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880, 3628800,
		39916800, 479001600
	};

	if (NULL == f)
	{
		fprintf(stderr,	"factorial: NULL output pointer\n");
		return -1;
	}

	if (i > 12)
	{
		fprintf(stderr,	"factorial: overflow\n");
		return -1;
	}

	*f = fact_table[i];

	return 0;
}
INT16 frac_pow(UINT32 y, INT16 i, UINT32* p)
{
	if (NULL == p)
	{
		fprintf(stderr,	"frac_pow: NULL output pointer\n");
		return -1;
	}

	if (i < 0)
	{
		fprintf(stderr,	"frac_pow: power is negative\n");
		return -1;
	}

	if (0 == i)
	{
		*p = FRAC_ONE;
		return 0;
	}
	if (1 == i)
	{
		*p = y;
		return 0;
	}

	frac_pow(y, (INT16) (i - 1), p);
	*p = (y * (*p)) / FRAC_ONE;
	return 0;
}
/* compute (retval).(*frac) = 2 ** x.y */
INT32 frac_pow2(INT16 x, UINT32 y, UINT32* frac)
{
	INT32 result;
	static UINT32 dpow2_table[30] = 
	{
		32768, 22713, 15743, 10912, 7564, 5242, 3634, 2518, 1746, 1210,
		838, 581, 403, 279, 193, 134, 93, 64, 44, 30,
		21, 14, 10, 7, 4, 3, 2, 1, 1, 0
	};
	INT16 k;
	UINT32 fk, fk1, tmp;

#if defined(POW2_DEBUG_PRINT)
	printf("frac_pow2(%d, %lu, frac):\n", x, y);
#endif

	if (NULL == frac)
	{
		fprintf(stderr,	"frac_pow2: fraction output pointer in NULL\n");
		return -1;
	}
	if (y >= FRAC_ONE)
	{
		fprintf(stderr,	"frac_pow2: y > FRAC_ONE\n");
		fprintf(stderr,	"  x = %d, y = %lu\n", x, y);
		fprintf(stderr,	"  x.y = %16.10f\n", frac_dbl(x, y));
		return -1;
	}
	if (x > INT_NBITS - 1)
	{
		fprintf(stderr,	"frac_pow2: cannot compute pow2 for x.y\n");
		fprintf(stderr, "where x > %d\n", INT_NBITS - 1);
		fprintf(stderr,	"  x = %d, y = %lu\n", x, y);
		fprintf(stderr,	"  x.y = %16.10f\n", frac_dbl(x, y));
		return -1;
	}

	*frac = 0;
	result = (1 << x);

	if (0 == y)
	{
		return result;
	}


	/* general case */

	/* Compute 2^(.y) with Taylor series developed in z = 0 */
	for (k = 1, fk1 = 0, fk = 1;
	     (k < sizeof(dpow2_table) / sizeof(UINT32)) && (fk != fk1);
		 k++)
	{
		fk = fk1;
		if (frac_pow(y, k, &tmp) != 0)
		{
			fprintf(stderr,	"frac_pow2: frac_pow failed\n");
			return -1;
		}
		tmp = ((dpow2_table[k] * tmp) / FRAC_ONE);
#if defined(POW2_DEBUG_PRINT)
		printf("  tmp = %lu\n", tmp);
#endif
		if (factorial(k, frac) != 0)
		{
			fprintf(stderr,	"frac_pow2: factorial failed\n");
			return -1;
		}
#if defined(POW2_DEBUG_PRINT)
		printf("  factorial = %lu\n", *frac);
#endif

		fk1 = fk1 + tmp / *frac;

#if defined(POW2_DEBUG_PRINT)
		printf("  fk1 = %lu\n", fk1);
#endif
	}
	k = frac_mul(result, 0, 1, fk1, &result, frac);
	if (k != 0)
	{
		fprintf(stderr,	"frac_pow2: frac_mul failed\n");
		return -1;
	}
	return result;

#if 0
	/*
	if ((1 == x) && (y <= 3))
	{
		switch (y)
		{
		case 1:
			*frac = 1;
			break;
		case 2:
			*frac = 3;
			break;
		case 3:
			*frac = 4;
		}
		return 2;
	}
	*/
	if ((0 == x) && (y <= 16))
	{
		switch (y)
		{
		case 1:
		case 2:
			*frac = 1;
			break;
		case 3:
			*frac = 2;
			break;
		case 4:
		case 5:
			*frac = 3;
			break;
		case 6:
			*frac = 4;
			break;
		case 7:
			*frac = 5;
			break;
		case 8:
			*frac = 6;
			break;
		case 9:
			*frac = 6;
			break;
		case 10:
			*frac = 7;
			break;
		case 11:
			*frac = 8;
			break;
		case 12:
			*frac = 8;
			break;
		case 13:
			*frac = 9;
			break;
		case 14:
			*frac = 10;
			break;
		case 15:
			*frac = 11;
			break;
		case 16:
			*frac = 11;
			break;
		}
		return 1;
	}

#if defined(POW2_DEBUG_PRINT)
	printf("  1. x.y = %16.10f\n", frac_dbl(x, y));
#endif

	if (x & 1)
	{
		y = (y >> 1) + (FRAC_ONE >> 1);
	}
	else
	{
		y = (y >> 1);
	}
	x >>= 1;

#if defined(POW2_DEBUG_PRINT)
	printf("  2. x.y = x.y / 2 = %16.10f (y = %lu)\n", frac_dbl(x, y), y);
#endif

	result = frac_pow2(x, y, &y);
	if (result < 0)
	{
		fprintf(stderr,	"frac_pow2: recursive call failed\n");
		return -1;
	}

	if (FRAC_ONE <= y)
	{
		fprintf(stderr,	"frac_pow2: internal error\n");
		fprintf(stderr, "  fraction = %lu >= '1' (%lu)\n", y, FRAC_ONE);
		return -1;
	}

#if defined(POW2_DEBUG_PRINT)
	printf("  3. x.y = frac_pow2(x.y / 2) = %16.10f (y = %lu)\n",
	  frac_dbl(result, y), y);
#endif

	x = frac_mul(result, y, result, y, &result, frac);
	if (x < 0)
	{
		fprintf(stderr,	"frac_pow2: frac_mul failed\n");
		printf("  frac_mul(x.y, x.y) was tried with x = %ld, y = %lu\n",
			result, y);
		return -1;
	}

#if defined(POW2_DEBUG_PRINT)
	printf("  4. x.y = frac_mul(x.y, x.y) = %16.10f\n",
	  frac_dbl(result, *frac));
#endif
		
	return result;
#endif
} /* frac_pow2 */


/* compute (retval).(*frac) = 2 ** x.y */
INT32 frac_pow2_orig(INT8 x, UINT32 y, UINT32* frac)
{
	INT32 result;
	INT32 factor;
	INT8 shift;

#if defined(POW2_DEBUG_PRINT)
	printf("frac_pow2(%d, %lu, frac):\n", x, y);
#endif

	if (NULL == frac)
	{
		fprintf(stderr,	"frac_pow2: fraction output pointer in NULL\n");
		return -1;
	}
	if (y >= FRAC_ONE)
	{
		fprintf(stderr,	"frac_pow2: y > FRAC_ONE\n");
		fprintf(stderr,	"  x.y = %16.10f\n", frac_dbl(x, y));
		return -1;
	}
	if (x > FRAC_NBITS - 2)
	{
		fprintf(stderr,	"frac_pow2: cannot compute pow2 for x.y\n");
		fprintf(stderr, "where x > FRAC_NBITS - 2\n");
		return -1;
	}

	*frac = 0;
	result = (1 << x);

	if (y == 0)
	{
		return result;
	}

	factor = FRAC_ONE;
	*frac = 0;
	for (shift = 0; shift < FRAC_NBITS; shift++)
	{
#if defined(POW2_DEBUG_PRINT)
		printf("shift = %d, y = 0x%lx ... ", shift, y);
#endif
		if (y & (1 << shift))
		{
			factor += (1 << shift);
			*frac += (1 << shift);
#if defined(POW2_DEBUG_PRINT)
			printf("factor = %ld, *frac = %lu", factor, *frac);
#endif
		}
#if defined(POW2_DEBUG_PRINT)
		printf("\n");
#endif
	}

	result = (result * factor) / FRAC_ONE;

	return result;
} /* frac_pow2_orig */


INT16 frac_sqrt(UINT32 x, UINT32* frac, UINT16 nbits)
{
	INT16 result;
	INT32 xk;
    UINT32 fk;
	INT32 xk1;
    UINT32 fk1;
	INT32 i;
	double real_value;

#if defined(SQRT_DEBUG_PRINT)
	printf("frac_sqrt(%lu, frac, %u):\n", x, nbits);
	printf("  sqrt(%ld) should be %16.10f\n", x, sqrt((double) x));
#endif

	if (NULL == frac)
	{
		fprintf(stderr, "frac_sqrt: fraction pointer is NULL\n");
		return -1;
	}

	if (x > (1 << (INT_NBITS - 2)) - 2)
	{
		fprintf(stderr,
          "frac_sqrt: input is bigger than 2^%d-2\n",
          INT_NBITS);
		return -1;
	}

	*frac = 0;

    if (x == 0)
    {
        return 0;
    }
    if (x == 1)
    {
        return 1;
    }
    if (x == 4)
    {
        return 2;
    }
    if (x == 9)
    {
        return 3;
    }

    if (x == 2)
    {
		*frac = (5741 * FRAC_ONE) / 13860;
        return 1;
    }
    if (x == 3)
    {
		*frac = (7953 * FRAC_ONE) / 10864;
        return 1;
    }
    if (x == 5)
    {
		*frac = (5473 * FRAC_ONE) / 23184;
        return 2;
    }
    if (x == 6)
    {
		*frac = (8721 * FRAC_ONE) / 19402;
        return 2;
    }
    if (x == 7)
    {
		*frac = (7873 * FRAC_ONE) / 12192;
        return 2;
    }
    if (x == 8)
    {
		*frac = (33461 * FRAC_ONE) / 40391;
        return 2;
    }

	/*
	 * construct initial guess of sqrt(x) = 2 ** (log2(x) / 2) in xk1.fk1
	 */

	xk = frac_log2(x, &fk, nbits);
	if (xk < 0)
	{
		fprintf(stderr,	"frac_sqrt: frac_log2 failed\n");
		return -1;
	}
#if defined(SQRT_DEBUG_PRINT)
	printf("  xk1.fk1 = log2(x) = %16.10f (%16.10f)\n",
	  frac_dbl(xk, fk), log((double) x) / log(2.0));
#endif

	fk1 = fk / 2;
	if (xk & 1)
	{
		fk1 += FRAC_ONE >> 1;
	}
	xk1 = xk / 2;
#if defined(SQRT_DEBUG_PRINT)
	printf("  xk1.fk1 = log2(x) / 2 = %16.10f (%16.10f)\n",
	  frac_dbl(xk1, fk1), frac_dbl(xk, fk) / 2.0);
#endif

	xk = xk1;
	fk = fk1;
	xk1 = frac_pow2((INT8) xk, fk, &fk1);
	if (xk1 < 0)
	{
		fprintf(stderr,	"frac_sqrt: frac_pow2 failed\n");
		return -1;
	}

#if defined(SQRT_DEBUG_PRINT)
	printf("  xk1.fk1 = 2^(log2(x) / 2) = xk1.fk1 = %16.10f (%16.10f)\n",
	  frac_dbl(xk1, fk1), pow(2.0, frac_dbl(xk, fk)));
#endif
	/*
	xk1 = 2;
	fk1 = 0;
	*/

	xk = fk = 0;

	i = 0;
    while ((xk1 != xk) || (fk1 != fk))
    {
        xk = xk1;
		fk = fk1;

		real_value = x / frac_dbl(xk, fk);
		result = frac_div(x, 0, xk, fk, &xk1, &fk1);
		if (result != 0)
		{
			fprintf(stderr,	"frac_sqrt: frac_div failed\n");
			return -1;
		}
#if defined(SQRT_DEBUG_PRINT)
		printf("  frac_div: xk1.fk1 = %16.10f (%ld, %lu; %16.10f)\n",
		  frac_dbl(xk1, fk1), xk1, fk1, real_value);
#endif

		real_value = frac_dbl(xk, fk) + frac_dbl(xk1, fk1);
		result = frac_sum(xk, fk, xk1, fk1, &xk1, &fk1);
		if (result != 0)
		{
			fprintf(stderr,	"frac_sqrt: frac_div failed\n");
			return -1;
		}
#if defined(SQRT_DEBUG_PRINT)
		printf("  frac_sum: xk1.fk1 = %16.10f (%ld, %lu; %16.10f)\n",
		  frac_dbl(xk1, fk1), xk1, fk1, real_value);
#endif

		real_value = frac_dbl(xk1, fk1) / 2.0;
		fk1 = fk1 / 2;
		if (xk1 & 1)
		{
			fk1 += (FRAC_ONE >> 1);
		}
		xk1 = xk1 / 2;
#if defined(SQRT_DEBUG_PRINT)
		printf("  divide by 2: xk1.fk1 = %16.10f (%ld, %lu; %16.10f)\n",
		  frac_dbl(xk1, fk1), xk1, fk1, real_value);
#endif
		i++;
		if (((xk1 == xk) && (abs((INT32) fk1 - (INT32) fk) < (xk / 2)))
		      || (i > 10))
		{
#if defined(SQRT_DEBUG_PRINT)
			printf("breaking Newton iteration, x = %lu, i = %ld\n",
			 x, i);
#endif
			break;
		}
    }

#if defined(SQRT_DEBUG_PRINT)
	printf("  result = %16.10f (%ld, %ld iters)\n",
	  frac_dbl(xk1, fk1), x, i);
#endif

	*frac = fk1;

    return (INT16) xk1;
} /* frac_sqrt */


INT32 fix_scale_times_sqrt(UINT16 x, UINT16 scale, INT16 trace)
{
    UINT32 xk;
    UINT32 xk1;
    UINT32 xk_2;
    UINT32 scale2;
	UINT16 i;

	/*
    if (x < 0)
    {
        fprintf(stderr,
            "fix_scale_times_sqrt: negative input value %d\n", x);
        return -1;
    }
	*/

    if (x == 0)
    {
        return 0;
    }
    if (x == 1)
    {
        return scale;
    }

	scale2 = scale * scale;
	i = 0;
    for (xk_2 = 0, xk = 0, xk1 = scale * (x >> 1); xk1 != xk; )
    {
        xk_2 = xk;
        xk = xk1;

		if (scale2 > 0xffff)
		{
			/* lose precision while decrease size first, then multiply */
	        xk1 = (scale2 / xk) * x;
	        xk1 += xk;
        	xk1 >>= 1;
		}
		else
		{
			/* do not lose precision, multiplication scale * x <= 2**32 - 1 */
	        xk1 = (scale2 * x) / xk;
	        xk1 += xk;
        	xk1 >>= 1;
		}

		/*
		if (xk1 < 0)
		{
	        fprintf(stderr,
	            "overflow: xk1 = %ld after computing\n", xk1);
	        fprintf(stderr,
	            "  xk1 = (scale2 * x) / xk; xk1 += xk; xk1 >>= 1\n");
			fprintf(stderr,
				"from scale2 = %ld, x = %d, xk = %ld\n",
				scale2, x, xk);
			return -1;
		}
		*/

		if (trace != 0)
		{
	        fprintf(stderr,
	            "fix_scale_times_sqrt: scale = %d, x = %d,",
				scale, x);
			fprintf(stderr,
				" xk = %ld, xk1 = %ld\n",
			    xk, xk1);
		}

		if (xk_2 == xk1)
		{
			break;
		}

		if (i++ > 30)
		{
        	fprintf(stderr,
	            "fix_scale_times_sqrt: too many iterations\n");
			return -1;
		}
    }

    if (xk1 > 0xffff)
    {
        /* check the calling conditions and algorithm if this happens */
        fprintf(stderr,
            "fix_scale_times_sqrt: output range: xk1 = %ld\n", xk1);
        return -1;
    }

    return xk1;
} /* fix_scale_times_sqrt */



static UINT32* log2array = NULL;
static UINT32* log_table = NULL;
#ifdef FRAC_D1LOOKUP
static UINT32* dlog_table = NULL;
#endif

INT16 frac_log2_init(UINT16 nbits)
{
	UINT32 i, N;
	UINT32 itmp;
	double dtmp;

#if defined(LOG2_DEBUG_PRINT)
	printf("frac_log2_init(%d)\n", nbits);
#endif

	if (nbits > 18)
	{
		fprintf(stderr,
		    "frac_log2_init: nbits, number of bits in lookup table index\n");
		fprintf(stderr,
		    "is not between 0-18.\n");
		return -1;
	}

	N = 1 << nbits;
#if defined(LOG2_DEBUG_PRINT)
	printf("  N = %lu\n", N);
#endif
	log2array = (UINT32*) malloc(N * sizeof(UINT32));
	if (NULL == log2array)
	{
		fprintf(stderr,
		    "frac_log2_init: cannot allocate factor lookup table: %lu bytes\n",
			N * sizeof(UINT32));
		return -1;
	}
	log_table = (UINT32*) malloc((N + 1) * sizeof(UINT32));
	if (NULL == log_table)
	{
		fprintf(stderr,
		    "frac_log2_init: cannot allocate log2 lookup table: %lu bytes\n",
			(N + 1) * sizeof(UINT32));
		return -1;
	}
#ifdef FRAC_D1LOOKUP
	dlog_table = (UINT32*) malloc((N + 1) * sizeof(UINT32));
	if (NULL == dlog_table)
	{
		fprintf(stderr,
		    "frac_log2_init: cannot allocate log2 derivative table: %lu bytes\n",
			(N + 1) * sizeof(UINT32));
		return -1;
	}
#endif

	for (i = 0; i < N; i++)
	{

#if FRAC_METHOD == FRAC_FACTOR
		/* Factoring loop, normal division, no multiple factors */
		dtmp = DBL_FRAC_ONE
			     * (log(1.0 + 1.0 / pow(2.0, i + 1.0)) / log(2.0));
		itmp = (UINT32) dtmp;
		if (itmp + 0.5 < dtmp)
		{
			itmp++;
		}
		log2array[i] = itmp;
#elif FRAC_METHOD == FRAC_LOOKUP
		/* Factoring loop, fast division with bit shifts, multiple factors */
		/* while x > 1, x = x/d, frac = frac + log2array[i] */
		/* This is also used in the ultra fast lookup table log2 */
		dtmp = DBL_FRAC_ONE
			     * (log((1 << (i+1)) / ((1 << (i+1)) - 1.0)) / log(2.0));
		itmp = (UINT32) dtmp;
		if (itmp + 0.5 < dtmp)
		{
			itmp++;
		}
		log2array[i] = itmp;
#else
#error FRAC_METHOD must be FRAC_LOOKUP or FRAC_FACTOR
#endif

#if FRAC_METHOD == FRAC_FACTOR
		/* This block is for the factoring with no multiple factors log2 */
		dtmp = DBL_FRAC_ONE * (log(1.0 + 1.0 / (1 << (i + 1))) / log(2.0));
		itmp = (UINT32) dtmp;
		if (itmp + 0.5 < dtmp)
		{
			itmp++;
		}
		log_table[i] = itmp;
#elif FRAC_METHOD == FRAC_LOOKUP
		/* This block is for the ultra fast lookup table log2 */
		dtmp = DBL_FRAC_ONE * (log(1.0 + i / (double) N) / log(2.0));
		itmp = (UINT32) dtmp;
		if (itmp + 0.5 < dtmp)
		{
			itmp++;
		}
		log_table[i] = itmp;

#ifdef FRAC_D1LOOKUP
		/* This block computes log2 derivative evaluated in the same point
		   as the log2 value for the ultra fast lookup table log2 */
		dtmp = DBL_FRAC_ONE / ((1.0 + i / (double) N) * log(2.0));
		itmp = dtmp;
		if (itmp + 0.5 < dtmp)
		{
			itmp++;
		}
		dlog_table[i] = itmp;
#endif
#else
#error FRAC_METHOD must be FRAC_LOOKUP or FRAC_FACTOR
#endif

	}

#if FRAC_METHOD == FRAC_FACTOR
	/* This line is for the factoring with no multiple factors log2 */
	log_table[N] = FRAC_ONE;
#elif FRAC_METHOD == FRAC_LOOKUP
	/* This line is for the ultra fast lookup table log2 */
	log_table[N] = FRAC_ONE;
#ifdef FRAC_D1LOOKUP
	dlog_table[N] = FRAC_ONE;
#endif
#else
#error FRAC_METHOD must be FRAC_LOOKUP or FRAC_FACTOR
#endif

	return 0;
} /* frac_log2_init */


INT8 frac_log2(UINT32 x, UINT32* frac, UINT16 nbits)
{
	UINT8 g;
	UINT32 j;
#if FRAC_METHOD == FRAC_FACTOR
	UINT32 d;
#endif

#if defined(LOG2_DEBUG_PRINT)
	printf("frac_log2(%lu, frac, %u):\n", x, nbits);
#endif

	if ((NULL == log2array) || (NULL == log_table))
	{
		fprintf(stderr,
		    "frac_log2: lookup tables have not been initialized\n");
		return -1;
	}
#ifdef FRAC_D1LOOKUP
	if (NULL == dlog_table)
	{
		fprintf(stderr,
		    "frac_log2: lookup tables have not been initialized\n");
		return -1;
	}
#endif
	if (0 == x)
	{
		fprintf(stderr,
		    "frac_log2: cannot compute logarithm of 0\n");
		return -1;
	}
	
	/* g = 15; */
	g = FRAC_NBITS - 1;
	g = INT_NBITS - 1;

	/* loop until msb of x occupies bit 15 */
	/*
	while(g && !(x & FRAC_ONE))
	*/
	while(g && !(x & INT_ONE))
	{
		x <<= 1;
		g--;
	}

	/* If g==0, fraction is zero or fraction is not wanted; we're done. */
	/*
	if(!g || (x == FRAC_ONE) || (NULL == frac))
	*/
	if(!g || (x == INT_ONE) || (NULL == frac))
	{
		if (NULL != frac)
		{
			*frac = 0;
		}
		return g;
	}

	x >>= INT_NBITS - FRAC_NBITS;

	/* 
	 * We have the integer portion of the log(x).
	 * Now get the fractional part.
	 */

#if defined(LOG2_DEBUG_PRINT)
	printf("  1. g = %u, x = %lu = %16.10f, log(x) = %16.10f\n",
		   g, x, x / DBL_FRAC_ONE, log(x / DBL_FRAC_ONE) / log(2.0));
#endif

#if FRAC_METHOD == FRAC_FACTOR

	*frac = 0;
	for (j = 1, d = FRAC_ONE >> 1;
         (j < (1 << nbits));
         j++, d >>= 1)
	{
		if ((FRAC_ONE + d) < x)
		{
			x = (x * FRAC_ONE) / (FRAC_ONE + d);
			*frac += log2array[j-1];
		}
	}

#elif FRAC_METHOD == FRAC_LOOKUP

	/* get rid of the MSB: x = x - 1 */
	x &= (FRAC_ONE - 1);

	/* Get the array index */
	j = x >> (FRAC_NBITS - nbits - 1);

#if defined(LOG2_DEBUG_PRINT)
	printf("  1.2 j = %lu, x = %lu = %16.10f, log(1 + x) = %16.10f\n",
		   j, x, x / DBL_FRAC_ONE, log(1.0 + x / DBL_FRAC_ONE) / log(2.0));
#endif

	x = x & ((FRAC_ONE - 1) >> nbits);
	*frac = log_table[j]
		      + ((2*x * (log_table[j+1] - log_table[j])) >> (FRAC_NBITS - nbits));

#if defined(LOG2_DEBUG_PRINT)
	printf("  log_table[j] = %lu = %16.10f\n",
		   log_table[j],
		   log_table[j] / DBL_FRAC_ONE);
	printf("  addition = %lu = %16.10f\n",
		   ((2*x * (log_table[j+1] - log_table[j])) >> (FRAC_NBITS - nbits)),
		   ((2*x * (log_table[j+1] - log_table[j])) >> (FRAC_NBITS - nbits)) / DBL_FRAC_ONE);
	printf("  *frac = %lu = %16.10f\n",
	       *frac, *frac / DBL_FRAC_ONE);
#endif

#else
#error FRAC_METHOD must be FRAC_LOOKUP or FRAC_FACTOR
#endif

	return g;
} /* frac_log2 */

