/*! 
\file 
\brief Vantage-Point Tree

Fast nearest neighbour search using VPT

\date 05 November 2003
\author Evgeny Karpov
*/

/*
    $Id: vptree.c,v 1.3 2003/11/20 13:46:12 ekarpov Exp $
*/

#include <stdio.h>
#include <stdlib.h>

#include <errorapi.h>

#include "vptree.h"

/*======================== Structures =============================*/

/* this typedef is needed to make self-refernce in structure */
typedef struct VPTreeNode VPTreeNode ;

/*! 
\brief Tree node

Contains parameters for Vantage Point Tree node
*/
struct VPTreeNode {

    int element_num ;   /*!< Node vector */
    double median ;     /*!< Median distance to all children */

    VPTreeNode * pleft ;    /*!< Left child */
    VPTreeNode * pright ;   /*!< Right child */

} ;

/*! 
\brief Vantage Point Tree 

Contains parameters for Vantage Point Tree
*/
typedef struct {

    VPTreeNode * root ;         /*!< Tree root */

    DISTFUNC dist_func ;     /*!< Distance function */
    void * dist_func_params ;   /*!< Distance function parameters */

} VPTree ;

/* these parameters are used to pass parameters during 
       recursive search through the tree */
typedef struct {

    int      nn_element ;       /*!< Current nearest vector */
    double   min_dist ;         /*!< Current minimum distance */
    double * search_element ;    /*!< Search vector */

    DISTFUNC dist_func ;     /*!< Distance function */
    void * dist_func_params ;   /*!< Distance function parameters */

} VPTSearchParams ;

/*======================== Structures =============================*/

int generate_vptree(VPTree * ptree, HFSET index_set) ;

int search_vptree(VPTreeNode * pnode, HFSET index_set,
                  VPTSearchParams * search_params) ;

void destroy_node(VPTreeNode * pnode) ;

/*======================== Functions ==============================*/

/*!
\brief Creates Vantage Point Tree

Constructs indexes using Vantage Point Tree algorithm

\param index_set Test set to index
\param source_params Indexed set source parameters
\return Pointer to VPT handle
\sa vptDestroy
\author Evgeny Karpov
*/
HVPTREE vptCreate(HFSET index_set, DISTFUNC dist_func,
                  void * dist_func_params) 
{
    VPTree * ptree ;

    if (index_set==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    if ((ptree = malloc(sizeof(VPTree)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    ptree->root = NULL ;
    ptree->dist_func = dist_func ;
    ptree->dist_func_params = dist_func_params ;

    /* generate tree */
    if (generate_vptree(ptree,index_set))
    {
        return (HVPTREE) ptree ;
    }

    /* error occured */
    vptDestroy(ptree) ;

    return NULL ;
}

/*!
\brief Destroys VP tree

Frees all memory used by tree structure

\param vpt Handle to tree
\return Nothing
\author Evgeny Karpov
*/
void vptDestroy(HVPTREE vpt) 
{
    if (vpt==NULL) return ;

    destroy_node(((VPTree*)vpt)->root) ;
    free(vpt) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Nearest neighbour search

Searches for nearest element using indexes

\param vpt Handle to tree
\param index_set Source set to search trough
\param search_element Search point (element for what nearest neighbour
is searched)
\param ret_dist Pointer to a variable where to put distance 
to found nearest neighbour element. If NULL then no distance
will be returned
\param source_params Source parameters (function pointers used to 
manipulate with source)
\return Number of nearest element in a source set
\sa vptSearch
\author Evgeny Karpov
*/
int vptNNSearch(HVPTREE vpt,HFSET index_set,
                double * search_element, double * ret_dist) 
{
    VPTree * ptree ;
    VPTSearchParams search_params ;

    if (vpt==NULL || index_set==NULL || search_element==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return 0 ;
    }

    ptree = (VPTree*) vpt ;

    search_params.nn_element = -1 ; /* mark that min_dist is not valid */
    search_params.search_element = search_element ;
    search_params.dist_func = ptree->dist_func ;
    search_params.dist_func_params = ptree->dist_func_params ;
    search_params.min_dist = 0.0 ;

    /* start recursive search */
    search_vptree(ptree->root,index_set,&search_params) ;

    if (ret_dist) *ret_dist = search_params.min_dist ;

    return search_params.nn_element ;
}

/*!
\brief Searches through the source with fixed radius

Searches nearest elements through the source with fixed radius

\param vpt Handle to tree
\param index_set Source set to search trough
\param search_element Search point (element for what nearest neighbour
is searched)
\param radius Search radius
\param ret_arr Array of elements from source what are as close
to search point as radius is (in other words, elements what are in 
a sphere with center in search point and radius equals to search 
radius)
\param source_params Source parameters (function pointers used to 
manipulate with source)
\return Number of nearest elements found
\todo To implement
\sa vptNNSearch
\author Evgeny Karpov
*/
int vptSearch(HVPTREE vpt,HFSET index_set,double * search_vector,
              double radius,int * ret_arr) 
{
    /* NOT IMPLEMENTED YET */

    return 0 ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Loads VP tree

Loads indexes from file

\param file File pointer
\return Tree handle
\sa vptSave
\author Evgeny Karpov
*/
HVPTREE vptLoad(FILE * file) 
{
    /* NOT IMPLEMENTED YET */
    return NULL ;
}

/*!
\brief Saves VP tree

Saves indexes to a file

\param vpt Tree handle
\param FILE * file
\return (function return value)
\note (note)
\sa (related items)
\author Evgeny Karpov
*/
int vptSave(HVPTREE vpt,FILE * file) 
{
    /* NOT IMPLEMENTED YET */
    return 0 ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Destroys tree node

Deallocates memory used by tree node

\param pnode Pointer to a node to destroy
\return Nothing
\author Evgeny Karpov
*/
void destroy_node(VPTreeNode * pnode) 
{
    if (pnode==NULL) return ;

    destroy_node(pnode->pleft) ;
    destroy_node(pnode->pright) ;
    free(pnode) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Creates VP tree node

Creates VP tree node by first creating root node and sorting 
all other elements into left and right children. This process
is then repeated recursively for all other nodes

\param ptree Pointer to a tree structure
\param index_set Set to index
\param distances Array which contains distances between root and 
other elements. 
\param elements Elements numbers
\param size Number of elements
\param source_params Indexed source parameters
\return Created node or NULL in case if node is a leaf
\note For internal use only
\author Evgeny Karpov
*/
VPTreeNode * create_node(VPTree * ptree,HFSET index_set,
                         double * distances,int * elements,int size) 
{
    VPTreeNode * pnode ;
    int curr_vect ;
    int left, right ;
    double * root_element ;

    if (size<=0) return NULL ;

    if ((pnode = (VPTreeNode*) malloc(sizeof(VPTreeNode)))==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    /* first vector will be the root */
    pnode->element_num = elements[0] ;
    pnode->median = 0.0 ;

    if (size==1)
    {
        /* the last element in branch */
        pnode->pleft = NULL ;
        pnode->pright = NULL ;
        
        return pnode ;
    }

    root_element = index_set->vectors[pnode->element_num-1] ;

    /* calculate distances */
    for (curr_vect=1;curr_vect<size;curr_vect++)
        distances[curr_vect] = 
            ptree->dist_func(root_element,index_set->vectors[elements[curr_vect]-1],
                             index_set->vector_size, ptree->dist_func_params) ;

    /* calculate median distance */
    for (curr_vect=1;curr_vect<size;curr_vect++)
        pnode->median += distances[curr_vect] / (size-1) ;

    /* put elements into two groups */
    left = 1 ; right = size - 1 ;
    while (left<right)
    {
        /* move left pointer till bigger distance or right pointer */
        while (distances[left] <= pnode->median && left<right) left++ ;

        if (left<right)
        {
            /* move right pointer till smaller distance or left pointer */
            while (distances[right] > pnode->median && left<right) right-- ;

            /* exchange indexes and distances */
            if (left<right)
            {
                distances[right] = distances[left] ;

                /* we don't need "distances[left]" anymore */
                /* then don't waste time to copy it */

                elements[0] = elements[left] ;
                elements[left] = elements[right];
                elements[right] = elements[0] ;

                left++ ;

            } 
        }
    }

    if (size==2 && distances[1]<=pnode->median) left=2 ;

    /* recursive call */
    pnode->pleft = create_node(ptree,index_set,distances,elements+1,left-1) ;
    pnode->pright = create_node(ptree,index_set,distances,elements+left,size-left) ;

    if ((pnode->pleft==NULL && left-1>0) ||
        (pnode->pright==NULL && size-left>0))
    {
        free(pnode) ;
        return NULL ;
    }

    return pnode ;
}

/*!
\brief Generates VP tree

Prepares all parameters and calls create_node to generate tree

\param ptree Pointer to a tree structure
\param index_set Set to index
\param source_params Indexed source parameters
\return Zero in error and one in success
\note For internal use only
\author Evgeny Karpov
*/
int generate_vptree(VPTree * ptree,HFSET index_set) 
{
    double * distances ;
    int * elements ;
    int i, set_size ;

    set_size = fvsGetSize(index_set) ;

    distances = (double*) malloc(sizeof(double)*set_size) ;
    elements = (int*) malloc(sizeof(int)*set_size) ;

    if (distances==NULL || elements==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        free(distances) ;
        free(elements) ;
        return 0 ;
    }

    /* fill indexes */
    for (i=1; i<=set_size; i++) elements[i-1] = i ;

    /* create tree */
    ptree->root = create_node(ptree,index_set,distances,elements,set_size) ;

    free(distances) ;
    free(elements) ;

    if (ptree->root)
        return 1 ;
    else
        return 0 ;
}

/*!
\brief Searches VP tree

Performs searching of VP tree by recursive calls

\param ptree Pointer to tree structure
\param index_set Set to search trough
\param pnode Node to search from
\param source_params Indexed source parameters
\return Number of nearest element in a set
\note For internal use only
\author Evgeny Karpov
*/
int search_vptree(VPTreeNode * pnode,HFSET index_set,VPTSearchParams * search_params) 
{
    double curr_dist ;

    /* distance to current node */
    curr_dist = search_params->dist_func(index_set->vectors[pnode->element_num-1],
        search_params->search_element,index_set->vector_size,search_params->dist_func_params) ;

    if (curr_dist < search_params->min_dist || search_params->nn_element == -1)
    {
        search_params->min_dist = curr_dist ;
        search_params->nn_element = pnode->element_num ;
    }

    if (pnode->pleft!=NULL && curr_dist-search_params->min_dist <= pnode->median) 
    {
        /* enter into the left child */
        search_vptree(pnode->pleft,index_set,search_params) ;
    }

    if (pnode->pright!=NULL && curr_dist+search_params->min_dist > pnode->median)
    {
        /* enter into the right child */
        search_vptree(pnode->pright,index_set,search_params) ;
    }

    return 1 ;
}

/*======================== Functions ==============================*/
