/*!  
\file
\brief Profile matcher with multiple feature sets 

Matches profiles with multiple feature sets performing fusion of 
separate classifier scores. 

\date 4 December 2003
\author Ville Hautamki
*/

/*
 * $Id: profmatch.c,v 1.5 2003/12/04 13:41:48 villeh Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h> 
#include <time.h>
#include <string.h>

#ifdef LINUX
#include <getopt.h>
#else
#include "getopt.h"
#endif

#include <fvector.h>

#include "mycsv.h"
#include "list.h"
#include "util.h"
#include "fusion.h"
#include "decision.h"
#include "weight.h"
#include "profmatch.h"


char *id = "Thu Mar 13 22:50:56 EET 2003";

int compare(char *src, char *dest);

int main(int argc, char **argv)
{
  //  int i;
  Parameters *parap;
  List *profile, *unknown; 
  Nbest *results;
  long watch;
  Csv *header;

  watch = setwatch();
  parap = init_params();
  parse_params(parap, argc, argv); 
  open_files(parap);

  profile = create_list();
  header = read_all_data(parap->infile, profile);

  unknown = create_list();
  read_all_data(parap->unknownfile, unknown);

  read_weights(profile, parap->weights);
  results = fusion(profile, unknown->top, header);
  results = decide(results);

  //  for (i = 0; i < parap->nbest; i++) {
  //    results = results->next;
  // }

  printf("%s %f %f %% ", results->id, 
	 results->match_score, 100 * results->confidence);

  if (compare(results->id, unknown->top->data[0]))
    printf("YES");
  else
    printf("NO");

  prtime(watch);

  free(parap);
  free_list(profile);
  free_list(unknown);
  csv_destroy(header); 

  return 0;
}

void free_params(Parameters *parap) 
{
  //  free(parap->instr);
  //  free(parap->weights);
  free(parap);
}

Parameters *init_params(void)
{
  Parameters *parap;

  parap = (Parameters *) allocate(sizeof(Parameters), NO_INIT);

  parap->f0w = 0;
  parap->ltasw = 0;
  parap->lfccw = 0;
  parap->mfccw = 0;
  parap->nbest = 1;
  parap->infile = NULL;
  parap->unknownfile = NULL;

  return parap;
}

void parse_params(Parameters *parap, int argc, char **argv)
{
  int option_index;
  int c;
  extern char *optarg;
  extern int optind;
  char opt_str[] = "p:u:w:n:f:t:l:m:h0";
  
  static struct option long_options[] = {
    { "prof",      1, NULL, 'p' },
    { "unknown",   1, NULL, 'u' },
    { "weights",   1, NULL, 'w' },
    { "nbest",     1, NULL, 'n' },
    { "f0",        1, NULL, 'f' },
    { "ltas",      1, NULL, 't' },
    { "lfcc",      1, NULL, 'l' },
    { "mfcc",      1, NULL, 'm' },
    { "help",      0, NULL, 'h' },
    { NULL,        0, NULL, 0 }
  };

  option_index = 0;
  while ((c = getopt_long_only(argc, argv, opt_str, long_options,
			  &option_index)) != EOF) {

    if (c == -1)
      return;

    switch(c) {
    case 'p':
      //      parap->instr = (char *) allocate(sizeof(char)*strlen(optarg), NO_INIT);
      strcpy(parap->instr, optarg);
      break;
    case 'u':
      //      parap->unknown = (char *) allocate(sizeof(char)*strlen(optarg), NO_INIT);
      strcpy(parap->unknown, optarg);
      break;
    case 'w':
      //      parap->weights = (char *) allocate(sizeof(char)*strlen(optarg), NO_INIT);
      strcpy(parap->weights, optarg);
      break;
    case 'n':
      parap->nbest = (int) atoi(optarg);      
      break;
    case 'f':
      parap->f0w = (double) atof(optarg);
      break;
    case 't':
      parap->ltasw = (double) atof(optarg);
      break;
    case 'l':
      parap->lfccw = (double) atof(optarg);
      break;
    case 'm':
      parap->mfccw = (double) atof(optarg);
      break;
    case 'h':
      print_usage(); 
      exit(0);
      break;
    default: 
      printf("Unknown param: %c\n", c);
      print_usage();
      break; 
    }
  }

}

void open_files(Parameters *parap)
{
  parap->infile = fopen(parap->instr,"r");
  parap->unknownfile = fopen(parap->unknown,"r");

  if ((parap->infile == NULL) || (parap->unknownfile == NULL)) {
    printf("Either profile file or unknown user feature file is missing!\n");
    print_usage();       
  }

  if ((parap->weights == NULL)) {
    printf("You must use either Weights file!\n");
    print_usage();       
  }
      

}

long setwatch()
{
  time_t tmp;
  return (long) time(&tmp);
}

void prtime(long start)
{
  struct timeval tv;
  double elapsed;

  gettimeofday(&tv, NULL);
  elapsed = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) - start;

  printf(" %6.1f\n", elapsed);
}

void print_usage() 
{
  printf("%s\n", id);
  printf("By Ville Hautamki for SuoPuhe\n");
  printf("Department of Computer Science, University of Joensuu\n\n");
  printf("Usage: profmatch [options]\n\n");
  printf("Matches speaker profiles with multiple feature sets\n"); 
  printf("using data fusion technique. Outputs closest match from profiles\n");
  printf("(in MSE sense) to unknown speaker.Match score and confidence are\n");
  printf("also reported\n\n");

  printf("Mandatory options:\n");
  printf("  --prof [profile file]\n");
  printf("  --weights [weights file]\n");
  printf("  --unknown [unknown user file]\n");
  printf("Optional parameters:\n");
  printf("  --nbest [number of best matches displayed]\n");
  exit(-1);
}

Csv *read_all_data(FILE *infile, List *profile)
{
  int i;
  Csv *header;
  Iterator *itep;
  Element *user;

  header = csv_parse(profile, infile);

  itep = iterate_start();
  while ((user = iterate_element(profile, itep)) != NULL) {
    user->vss = (HFSET *) allocate(sizeof(HFSET) * (user->num_fields -1), 
				     NO_INIT);
    for (i = 0; i < user->num_fields -1; i++) {
      user->vss[i] = fvsLoad(user->data[i+1]);
    }
  }

  return header;  
}

int compare(char *src, char *dest)
{
  int i;
  
  for (i = 0; i < 7; i++) {
    if (src[i] != dest[i]) 
      return 0;
  }

  return 1;
}
