/*!
\file dsp.c
\brief Digital signal processing implementation

Implementation of several well-known 
digital signal processing algorithms\n
$Id: dsp.c,v 1.3 2004/01/02 10:47:23 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32
\date 2003
\author Evgeny Karpov
*/
#include <stdlib.h>


#include <errorapi.h>


#include "dsp.h"


#if !defined(M_PI)
/*! Define pi if not already defined */
#define M_PI 3.14159265358979323846
#endif


/*======================== Functions ==============================*/


/*!
\brief Inner product of two vectors

Return sum of products of elements
of two vectors \p factor1 and \p factor2.
The inner product of vectors \f$ f, g \in \mathbb{R}^n\f$
is defined as
\f[
    \langle f, g \rangle = \sum_{k=1}^n f_k g_k.
\f]
\param factor1 Pointer to the first factor's input vector data
\param factor2 Pointer to the second factor's input vector data
\param size Length of both vectors
\return Inner product of the vectors
\author Juhani Saastamoinen
 */
INT32 dspInnerProduct(
	INT16 * factor1,
	INT16 * factor2,
	int size)
{
    int i;
    INT32 s;

    for (i = 0, s = 0; i < size; i++)
    {
        s += factor1[i] * factor2[i];
    }

    return s;
}


/*!
\brief Pointwise multiplication of vectors

Multiply two vectors \p factor1 and \p factor2.
The multiplication output in the variable \p output.
The pointwise multiplication of vectors \f$ f, g \in \mathbb{R}^n\f$
is defined as
\f[
	h_k = f_k g_k,\qquad k \in 1,\dots,n.
\f]
\param factor1 Pointer to the first factor's input vector data
\param factor2 Pointer to the second factor's input vector data
\param output Pointer to the data resulting from the multiplication.
\param size The input signal length
\note There must be at least \p size
elements available for writing after the output pointer.
\author Juhani Saastamoinen
 */
int dspMultiply(
	INT16 * factor1,
	INT16 * factor2,
	INT32 * output,
	int size)
{
    int i;

    for (i=0; i < size; i++)
    {
        output[i] = factor1[i] * factor2[i];
    }

    return size ;
}


/*!
\brief Convolution

Convolution of vectors \p input and \p kernel,
a signal and convolution kernel.
The convolution result is output in the variable \p output,
the size of the output will be \p input_size + \p kernel_size - 1.
The full convolution of vectors \f$ f \in \mathbb{R}^n\f$ and
\f$ g \in \mathbb{R}^m\f$ (the kernel) is defined as
\f[
	h_l = (f*g)_l= \sum_{k=1}^m f_{l-k} g_k,\qquad l \in 1,\dots,n+m-1,
\f]
where \f$ f_{t-k} = 0 \f$ is used whenever \f$ t-k \notin \{1,\dots,n\} \f$.
\param input Pointer to the input vector data
\param input_size The input signal length
\param kernel Pointer to the convolution kernel vector data
\param kernel_size The size of the convolution kernel
\param output The result of the convolution
\note There must be at least \p input_size + \p kernel_size - 1
elements available for writing after the output pointer.
\note Convolution results of the same size or valid size
can be obtained by taking subsets of the output vector.
\sa FFT
\author Evgeny Karpov
 */
int dspConvolution(
	INT16 * input,
	int input_size,
	INT16 * kernel,
	int kernel_size,
	INT32 * output) 
{
    int i,j;

    for (i = 0; i < input_size + kernel_size - 1; i++)
    {
        output[i] = 0 ;
    }

    output += kernel_size - 1;
    for (j = 0; j < kernel_size; j++)
    {
        for (i = 0; i < input_size; i++)
        {
            output[i] += kernel[kernel_size - j - 1] * input[i];
        }
        output--;
    }

    return input_size + kernel_size - 1;
}


/*!
\brief DC Removal

Remove the average from the vector \p input.
The DC removal from vector \f$ f \in \mathbb{R}^n\f$ is defined as
\f[
	g_k = f_k - \overline f,\qquad
	  \overline f = {1\over n} \sum_{k=1}^n f_k.
\f]
\param input Pointer to the input vector data
\param output Pointer to the output vector data, DC removed
\param size The input and output signal length
\param params Unnecessary parameter, not used.
(In some signal processing applications DC level
is computed once and the same DC level used in the later calls)
\return size, or negative value in case of failure
\note This can also be done in the frequency
domain if frequency domain signal processing is done
in the same signal processing block, e.g. DC removal
before FFT should be removed completely and done
with the FFT output instead.
\note There must be at least \p size
elements available for writing after the output pointer.
\todo Remove the params argument ???
\sa FFT
\author Evgeny Karpov
 */
int dspDCRemove(
	INT16 * input,
	INT16 * output,
	int size,
	void * params) 
{
    INT32 dc ;
    int i ;

	if ((size <= 0) || (size > (1 << 16)))
	{
		/* avoid INT32 overflow when computing dc */
		ReportError(ESR_INVALID_PARAMETER, NULL);
		return -1;
	}

    if (NULL == params)
    {
        /* compute mean */
        dc = 0 ;
        for (i=0;i<size;i++)
        {
            dc += input[i];
        }
        dc /= size;
    }
    else
    {
        dc = *((INT16*)params);
    }

    /* remove mean */
    for (i=0;i<size;i++)
    {
        output[i] = input[i] - (INT16) dc ;
    }

    return size ;
}


/*!
\brief First Order Difference Filtering

Compute first order differences of the vector \p input.
The input argument \p params defines the weighting
of the differencing; the default weight is 0.97
which results in a low frequency cancelling
band-pass filter, which is why the function is named
\e dspHighEmphFilter as in <em>High Frequency Emphasis Filtering</em>.
The first order differencing of a vector \f$ f \in \mathbb{R}^n\f$
is defined as
\f[
	g_k = f_k - \alpha f_{k-1},\qquad k \in 2,\dots,n
\f]
There is no standard definition for \f$ g_1 \f$.
The function dspHighEmphFilter computes it as
\f$ g_1 = f_1 - \alpha f_1\f$.
\param input Pointer to the input vector data
\param output Pointer to the filtered output vector data
\param size The input and output signal length
\param params Pointer to the difference parameter. if \p params is
\p NULL, then the default parameter \f$ \alpha = 0.97 \f$ is used.
Otherwise, the number <tt>*params</tt> will be used.
\note This can also be done in the frequency
domain by pointwise multiplication of the signal
spectrum. This is an advantage if there are
other linear filters applied consecutively.
\note There must be at least \p size
elements available for writing after the output pointer.
\note There is no standard definition for <tt>output[0]</tt>.
This function computes <tt>output[0] = input[0] - a * input[0]</tt>. 
\sa dspConvolution
\author Evgeny Karpov
 */
int dspHighEmphFilter(
	INT16 * input,
	INT16 * output,
	int size,
	void * params) 
{
    INT16 coeff ;
    INT32 tmp;
    int i ;

    if (params)
    {
        coeff = *((INT16*) params) ;
    }
    else 
    {
        /* coeff = (INT16) (0.97 * (1 << 15)); */
        coeff = (INT16) 31785; /* 31784.96 rounded, not truncated */
    }

    tmp = input[0];
    for (i = 0; i < size; i++)
    {
        output[i] = (INT16) (((((INT32) input[i]) << 15)
                                - (coeff * tmp)) >> 15) ;
        tmp = input[i] ; 
    }

    return size ;
}


/*!
\brief Number of Zero Crossings

Compute the number of zero crossings in input vector \p input.
Zero crossing rate is defined as the number of times
that signal changes sign, divided by the vector size.
We define the zero crossing rate for
a vector \f$ f \in \mathbb{R}^n \f$ as
\f[
	\mathop\mathrm{ZCR}(f) =
	  \frac{1}{n}
	  \sum_{k=2}^n c(k),\qquad \textrm{ where }c(k) = 
	\left\{
	\begin{array}{ll}
	0, & \textrm{if } f_{k-1} = 0
         \textrm{ or } \mathop\mathrm{sign}(f_k)
                         = \mathop\mathrm{sign}(f_{k-1}), \\
	1, & \textrm{otherwise.}
	\end{array}
	\right.
\f]
Here \f$x = 0\f$ means for a floating
point number \f$x\f$ that \f$|x|<\varepsilon\f$,
where \f$\varepsilon > 0\f$ is a pre-defined
constant. The algorihm uses the constant DBL_EPSILON.
\param input Pointer to the input vector data
\param size The input vector length
\return Number of times the signal changes sign
\note The signal is interpreted as zero, when
its absolute value is below the floating point
constant DBL_EPSILON. If a digital signal
whose floating point representation is scaled
normally, i.e. dynamic range is \f$(-L,L)\f$,
where \f$ L \geq 1\f$, then the constant DBL_EPSILON
easily accounts for the dynamic range of all
digital signals with resolution up to 24 bits at least.
\author Evgeny Karpov
 */
int dspZCR(INT16 * input, int size) 
{
    int ret_val = 0 ;
    int i ;
    int j;

    for (i = 0; input[i] == 0; i++)
    {
        ;
    }

    if (i == size)
    {
        /* whole signal is zero, no zero crossings */
        return 0;
    }

    for (j = i + 1; j < size; j++)
    {
        if ((((input[j] < 0) && (input[i] > 0))
               || ((input[i] < 0) && (input[j] > 0))))
        {
            ret_val++;
            i = j;
        }
    }

    return ret_val ;
}


/*!
\brief Vector Energy

Compute the energy of a vector.
The energy of a vector \f$ f \in \mathbb{R}^n \f$ is
defined as the squared \f$ L_2 \f$ norm
\f[
	||f||_2^2 = 
	  \sum_{k=1}^n f_k^2.
\f]
\param input Pointer to the input vector data
\param size The input vector length
\return Signal energy
\sa Parseval's Theorem.
\author Evgeny Karpov
 */
INT32 dspEnergy(INT16 * input,int size) 
{
    INT32 ret_val = 0;
    int i ;

    for (i=0;i<size;i++)
    {
        ret_val += input[i] * input[i] ;
    }

    return ret_val ;
}


/*!
\brief Power of a Finite Digital Signal

Compute power estimate for a finite signal
as the relative energy.
\param input Pointer to the input vector data
\param size The input vector length
\return Signal power
\sa dspEnergy
\author Evgeny Karpov
 */
INT32 dspPower(INT16 * input,int size) 
{
    return dspEnergy(input, size) / size ;
}

/*======================== Functions ==============================*/


