/*! 
\file fftdct.c
\brief A generic FFT and DCT library implementation,
used by MFCC module.

This is an implementation of FFT and DCT used by MFCC module.
The current implementation uses the shortfft library for
FFT and a direct, not FFT-based computation of the DCT.\n
$Id: fftdct.c,v 1.4 2003/12/30 21:01:43 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32
\date 2003/11
\author Juhani Saastamoinen
*/
#include <stdlib.h>
#include <math.h>


#include <errorapi.h>
#include <shortfft.h>


#include "fftdct.h"


#if !defined(M_PI)
/*! Define pi if not already defined */
#define M_PI 3.14159265358979323846
#endif


FFTProcessor* fftCreate(int input_size)
{
	FFTProcessor* p;

    if (input_size < 2)
    {
        /* It makes no sense to compute 1 point FFT */
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

    if (input_size > 1024)
    {
        /* shortfft supports FFT length up to 2^10 = 1024 */
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

    p = malloc(sizeof(FFTProcessor));
    if (NULL == p)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL;
    }

    p->input_size = input_size;

    return p;
}


void fftDestroy(FFTProcessor* fft)
{
    free(fft);
}


int fftGetInputSize(FFTProcessor* fft)
{
    if (NULL == fft)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return fft->input_size;
}


int fftCompute(FFTProcessor* fft, INT16* input, INT32* spectrum)
{
    int result;

    if ((NULL == fft) || (NULL == input) || (NULL == spectrum))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if (fft->input_size > 1024)
    {
        /* shortfft supports FFT length up to 2^10 = 1024 */
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /*
       compiler may warn about integer size mismatch;
       fft->input_size is checked above and fits in INT16
    */
    result = shortfft((INT16) fft->input_size, input, spectrum, NULL);
    if (result != 0)
    {
        ReportError(ESR_FFT, NULL);
        return ESR_FAILED;
    }

    return ESR_SUCCEED;
}


/*!
\brief Create Discrete Cosine Transformation data structure

The discrete cosine transformation (DCT) of vector \p input.
The DCT of a vector \f$ f \in \mathbb{R}^n\f$ is defined as
\f[
	g_w =
	  \sum_{k=1}^n
	    f_k\,\cos\left({k \over n} \pi \left(k+{1 \over 2}\right)\right).
\f]
\param input Pointer to the input vector data
\param input_size The input signal length
\param output Pointer to the DCT output vector data
\param output_size The size of the DCT output vector
\note The standard DCT output size is the same as the input size
but in some applications it is useful to map the input vector
to a lower dimensional space with a DCT-like manner.
\note There must be at least \p output_size
elements available for writing after the output pointer.
\todo The implementation is not optimal:
Use temporary variables for part of the cosine arguments
or even better: wrap DCT in a struct/class,
compute the cosines once, and use
convolution-like folded pointwise multiplication
to perform the DCT efficiently.
\sa FFT
\author Evgeny Karpov
 */
DCTProcessor* dctCreate(int input_size, int spectrum_size)
{
    int i,j ;
    DCTProcessor* dct;

    if ((input_size < 2) || (spectrum_size < 2)
          || (spectrum_size > input_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

    dct = (DCTProcessor*) malloc(sizeof(DCTProcessor));
    if (NULL == dct)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL;
    }
    dct->input_size = input_size;
    dct->spectrum_size = spectrum_size;

    dct->cos_functions = (INT16**) malloc(sizeof(INT16*) * spectrum_size);
    if (NULL == dct->cos_functions)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        dctDestroy(dct);
        return NULL;
    }

    for (i = 0; i < spectrum_size; i++)
    {
        dct->cos_functions[i] = (INT16*) malloc(sizeof(INT16) * input_size);
        if (NULL == dct->cos_functions[i])
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            dctDestroy(dct);
            return NULL;
        }
    }

    for (i = 0; i < spectrum_size; i++)
    {
        for (j = 0; j < input_size; j++)
        {
            dct->cos_functions[i][j] = 
              (INT16) (DCT_SCALE * cos(i * (j + 0.5) * (M_PI/input_size)));
        }
    }

    return dct;
}


void dctDestroy(DCTProcessor* dct)
{
    int i;

    if (NULL == dct)
    {
        return;
    }

    if (NULL == dct->cos_functions)
    {
        free(dct);
        return;
    }

    for (i = 0; i < dct->spectrum_size; i++)
    {
        free(dct->cos_functions[i]);
    }

    free(dct->cos_functions);
    free(dct);

    return;
}


int dctGetInputSize(DCTProcessor* dct)
{
    if (NULL == dct)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return dct->input_size;
}


int dctGetSpectrumSize(DCTProcessor* dct)
{
    if (NULL == dct)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    return dct->spectrum_size;
}


int dctCompute(DCTProcessor* dct, INT16* input, INT32* spectrum)
{
    int i, j;

    if ((NULL == dct) || (NULL == input) || (NULL == spectrum))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if ((NULL == dct->cos_functions)
          || (dct->input_size < 2) || (dct->spectrum_size < 2)
          || (dct->spectrum_size > dct->input_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    for (i = 0; i < dct->spectrum_size; i++)
    {
        spectrum[i] = 0;
        for (j = 0; j < dct->input_size; j++)
        {
            spectrum[i] += dct->cos_functions[i][j] * input[j] ;
        }
    }

    return ESR_SUCCEED;
}


