/*!
\file filterbank.c
\brief Filter bank implementation

Implementation of bank of linear filters
operating with magnitude spectra
in the frequency domain.\n
$Id: filterbank.c,v 1.7 2004/01/02 16:09:54 juhani Exp $
\date 2003
\author Evgeny Karpov
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <errorapi.h>


#include "dsp.h"
#include "filterbank.h"


/*======================== Structures =============================*/

/*! 
\brief Linear filter for operating in the frequency domain

Frequency domain filter parameters: lower frequency bound,
upper frequency bound, coefficients. These parameters
together specify the frequency response of the filter.
\sa FilterBank
*/
typedef struct {

    int begin ; /*!< The index of the lowest
	                 harmonic frequency passed */
    int end ;   /*!< The index of the highest
	                 harmonic frequency passed */

    INT16 * coeff ; /*!< The frequency multipliers
					      between low/high bound frequencies */

} Filter ;

/*! 
\brief Collection of linear filters for frequency domain

Collection of <tt>Filter</tt>'s of given maximum size.
It is possible to define filters as such or scaled
so that each filter's output is divided
by the number of passed frequencies (???)
\sa Filter
\sa HFBANK
*/
typedef struct {

    int size ;        /*!< maximum number of filters (???) */
    int num_filters ; /*!< number of used filters */

    int average ;     /*!< divide by the filter length or not
                           \todo remove this (it is never used) */

    Filter ** filters ; /*!< The filter table */

} FilterBank ;


/*======================== Structures =============================*/


/*!
\brief Initial maximum number of filters in a filter bank

When a filter bank is first created with the function
fbCreate, this is the maximum number of filters
in the created filter bank.
\sa fbCreate
 */
#define FILTERBANK_INITIAL_SIZE     10

/*-----------------------------------------------------------------*/

int CheckFilterBank(FilterBank * filter_bank) ;
Filter * CreateFilter(int begin, int end, INT16* coeff) ;

/*======================== Functions ==============================*/

/*!
\brief Create a filter bank

Allocate a set of filters for filterbank processing
\param filters Number of filters to allocate
\param average Are the new filter scaled so that
frequency multipliers sum to one or are they not
\return Pointer to the created filter bank
or NULL in case of failure
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
HFBANK fbCreate(
	int filters,
	int average) 
{
    FilterBank * fbank ;

    if ((fbank = malloc(sizeof(FilterBank)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL ;
    }

    fbank->average = average ;

    if (filters != FILTERBANK_FILTERS_AUTO)
	{
        if (filters < 0)
        {
            ReportError(ESR_INVALID_PARAMETER, NULL);
            free(fbank);
            return NULL;
        }
        fbank->size = filters ;
	}
    else
	{
        fbank->size = FILTERBANK_INITIAL_SIZE ;
	}

    fbank->filters = malloc(sizeof(Filter*)*fbank->size);
    if (NULL == fbank->filters)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        free(fbank) ;
        return NULL ;
    }

    memset(fbank->filters, 0, sizeof(Filter*)*fbank->size) ;

    fbank->num_filters = 0 ;

    return (HFBANK) fbank ;
}

/*!
\brief Destroy a filter bank

\param fbank Pointer to a proper filter bank that is
deallocated or NULL
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
void fbDestroy(
	HFBANK fbank) 
{
    FilterBank * filter_bank ;
    Filter ** filters ;
    int i, size ;

    if ((filter_bank = (FilterBank*) fbank)==NULL)
	{
		return ;
	}

    filters = filter_bank->filters ;
    size = filter_bank->num_filters ;
    for (i=0;i<size;i++)
    {
        if (filters[i]!=NULL)
        {
            free(filters[i]->coeff) ;
            free(filters[i]) ;
        }
    }

    free(filter_bank->filters) ;
    free(filter_bank) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Add a filter to end of a filter bank

Add a new filter to the end of a filter bank.
The lower and upper frequency bounds of the filter
and frequency multipliers are given as input.
\param fbank The filter bank where new filter is inserted
\param begin Lower bound frequency index
\param end High bound frequency index
\param coeff Frequency multipliers
\return The number of filters after addition, or error code
\sa fbInsertFilter
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
int fbAddFilter(
	HFBANK fbank,
	int begin,
	int end,
	INT16 * coeff) 
{
    FilterBank * filter_bank ;
    Filter ** filters ;
    int curr ;

	filter_bank = (FilterBank*) fbank;
    if (NULL == filter_bank) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    if ((begin < 0) || (end < begin) || (NULL == coeff))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* check the size of filter bank */
    if (CheckFilterBank(filter_bank)!=ESR_SUCCEED)
    {
        ReportError(ESR_FILTER_BANK_CHECK, NULL);
        return ESR_FAILED ;
    }

    filters = filter_bank->filters ;
    curr = filter_bank->num_filters ;

    /* allocate memory for filter */
    filters[curr] = CreateFilter(begin, end, coeff);
    if (NULL == filters[curr])
    {
        ReportError(ESR_FILTER_CREATE, NULL);
        return ESR_FAILED ;
    }

    filter_bank->num_filters++ ;

    return filter_bank->num_filters ;
}

/*!
\brief Add a filter to middle of a filter bank

Add a new filter to the middle of a filter bank.
Old filter indices after addition index \p position
are pushed one index further.
The lower and upper frequency bounds of the filter
and frequency multipliers are given as input.
\param fbank Filter bank where new filter is inserted
\param begin Lower bound frequency index
\param end High bound frequency index
\param coeff Frequency multipliers
\param position The index where new filter is added
\return The number of filters after addition, or error code
\sa fbAddFilter
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
int fbInsertFilter(
	HFBANK fbank,
	int begin,
	int end,
	INT16 * coeff,
	int position) 
{
    FilterBank * filter_bank ;
    Filter ** filters ;
    int i,curr ;

    if ((filter_bank = (FilterBank*) fbank)==NULL || position==0) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    if ((begin < 0) || (end < begin) || (NULL == coeff))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if (position >= filter_bank->num_filters + 1)
    {
        filter_bank->num_filters = position - 1;
        i = fbAddFilter(fbank, begin, end, coeff) ;
        if (ESR_FAILED == i)
        {
            ReportError(ESR_FILTER_BANK_ADD_FILTER, NULL);
        }
        return i;
    }

    filters = filter_bank->filters ;
    curr = filter_bank->num_filters ;

    /* check the size of filter bank */
    if (CheckFilterBank(filter_bank)==ESR_FAILED)
    {
        ReportError(ESR_FILTER_BANK_CHECK, NULL);
        return ESR_FAILED ;
    }

    /* move filters */
    for (i=curr;i>=position;i--)
    {
        filters[i] = filters[i-1] ;
    }

    filters[position-1] = CreateFilter(begin, end, coeff);
    if (NULL == filters[position-1])
    {
        ReportError(ESR_FILTER_CREATE, NULL);

        /* move filters back */
        for (i=position-1;i<curr;i++)
        {
            filters[i] = filters[i+1] ;
        }

        filters[curr] = NULL ;

        return ESR_FAILED ;
    }

    filter_bank->num_filters++ ;

    return position ;
}

/*!
\brief Remove a filter from a filterbank

Remove a filter at a given index \p filter
from a filterbank \p fbank
\param fbank The filter bank
\param filter Index of the filter that must be removed
\return Number of filters in the filter bank after removal
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
int fbRemoveFilter(
	HFBANK fbank,
	int filter) 
{
    FilterBank * filter_bank ;
    int i ;

    filter_bank = (FilterBank*) fbank;
    if ((NULL == filter_bank)
          || (filter > filter_bank->num_filters))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    /* free filter */
    if (filter_bank->filters[filter-1]!=NULL)
    {
        free(filter_bank->filters[filter-1]->coeff) ;
        free(filter_bank->filters[filter-1]) ;
        filter_bank->filters[filter-1] = NULL ;
    }

    /* move filters */
    for (i = filter - 1; i < filter_bank->num_filters; i++)
    {
        filter_bank->filters[i] = filter_bank->filters[i+1] ;
    }

    filter_bank->num_filters-- ;

    return filter_bank->num_filters ;
}

/*!
\brief Get first passed frequncy of a filter

Get the index of the first passed frequency
of a given filter in a filter bank.
\param fbank Filter bank
\param filter The index of the filter in the filter bank
\return The index of the first passed frequency
\sa HFBANK (a pointer to FilterBank)
\sa Filter
\author Evgeny Karpov
*/
int fbGetFilterBegin(
	HFBANK fbank,
	int filter) 
{
    FilterBank * filter_bank ;

    if ((filter_bank = (FilterBank*) fbank)==NULL ||
        filter_bank->num_filters<filter) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return filter_bank->filters[filter-1]->begin;
}

/*!
\brief Get last passed frequncy of a filter 

Get the index of the last passed frequency
of a given filter in a filter bank.
\param fbank Filter bank
\param filter The index of the filter in the filter bank
\return The index of the last passed frequency
\sa HFBANK (a pointer to FilterBank)
\sa Filter
\author Evgeny Karpov
*/
int fbGetFilterEnd(
	HFBANK fbank,
	int filter) 
{
    FilterBank * filter_bank ;

    if ((filter_bank = (FilterBank*) fbank)==NULL ||
        filter_bank->num_filters<filter) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return filter_bank->filters[filter-1]->end;
}

/*!
\brief Get number of passed frequencies of a filter

Get the index of the last passed frequency
of a given filter in a filter bank.
\param fbank Filter bank
\param filter The index of the filter in the filter bank
\return Number passed harmonic frequencies
\sa HFBANK (a pointer to FilterBank)
\sa Filter
\author Evgeny Karpov
*/
int fbGetFilterLength(
	HFBANK fbank,
	int filter) 
{
    FilterBank * filter_bank ;

    if ((filter_bank = (FilterBank*) fbank)==NULL ||
        filter_bank->num_filters<filter) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return filter_bank->filters[filter-1]->end - 
           filter_bank->filters[filter-1]->begin + 1;
}

/*!
\brief Get pointer to the frequency multipliers of a filter

\param fbank Filter bank
\param filter The index of the filter in the filter bank
\return Pointer to the frequency multiplier array
\sa HFBANK (a pointer to FilterBank)
\sa Filter
\author Evgeny Karpov
*/
INT16 * fbGetFilterCoeff(
	HFBANK fbank,
	int filter) 
{
    FilterBank * filter_bank ;

    if ((filter_bank = (FilterBank*) fbank)==NULL ||
        filter_bank->num_filters<filter) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }

    return filter_bank->filters[filter-1]->coeff;
}

/*!
\brief Get number of filters currently in a filter bank

\param  fbank Filter bank
\return Number of filters currently in the filter bank
\sa HFBANK (a pointer to FilterBank)
\author Evgeny Karpov
*/
int fbGetSize(
	HFBANK fbank) 
{
    FilterBank * filter_bank ;

    filter_bank = (FilterBank*) fbank;
    if (NULL == filter_bank) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return filter_bank->num_filters ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Apply all filters of a filter bank

Apply filters in a filter bank for a given
spectrum and collect the magnitude
\param fbank Filter bank
\param input Magnitude spectrum
\param input_size Spectrum vector length
\param output Sum of output magnitudes collected
per each filter
\param output_size Number of filters to use,
0 means using the whole filter bank
\return The number of outputs computed
or ESR_FAILED in case of failure
\note This function assumes that the input
is a collection of magnitudes of the harmonic
frequencies of a Fourier analysis of a signal
\note If the averaging is set for the filter bank,
then each filter output sum is divided by the number
of passing frequencies for that filter
\sa HFBANK (a pointer to FilterBank)
\sa FFT
\author Evgeny Karpov
*/
int fbApply(
	HFBANK fbank,
	INT16 * input,
	int input_size,
	INT32 * output,
	int output_size) 
{
    FilterBank * filter_bank ;
    int out ;

    filter_bank = (FilterBank*) fbank;
    if ((NULL == filter_bank) || (NULL == input) || (NULL == output)) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    if (output_size == 0)
    {
        output_size = filter_bank->num_filters ;
    }

    for (out = 0;
         (out < filter_bank->num_filters) && (out < output_size);
         out++)
    {
        output[out] = fbApplyFilter(fbank, input, input_size, out);
        if (output[out] < 0.0)
        {
            ReportError(ESR_FILTER_BANK_APPLY_ONE, NULL);
            return ESR_FAILED;
        }
    }

    return out ;
}

/*!
\brief Apply one filter of a filter bank

Compute inner product of a real magnitude spectrum
input with the frequency response of a filter in a filter bank
\param fbank Filter bank
\param spectrum Magnitude spectrum input
\param spectrum_size Length of the spectrum
\param filter Index of the filter to use
\return Sum of the harmonic frequency magnitudes
after pointwise multiplication with the filter's
frequency response, or negative value in case of failure
\pre spectrum_size must be at least index of last
passing frequency plus one.
\todo Explain why average is computed by dividing with
number of passed frequencies? Why not input_size?
\sa HFBANK (a pointer to FilterBank)
\sa FFT
\author Evgeny Karpov
*/
INT32 fbApplyFilter(
	HFBANK fbank,
	INT16 * spectrum,
	int spectrum_size,
    int filter)
{
    FilterBank * filter_bank ;
    Filter * curr_filter ;
    INT32 ret_val ;

	/*
	fprintf(stderr, "fbApplyFilter(fbank, spectrum, %d, %d):\n",
	  spectrum_size, filter);
	*/

    filter_bank = (FilterBank*) fbank;
    if ((NULL == filter_bank)
          || (NULL == spectrum)
          || (filter > filter_bank->num_filters))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1 ;
    }

    ret_val = 0 ;
    curr_filter = filter_bank->filters[filter] ;

    /* Check that the passed frequency indices of
       the filter are within boundaries set by spectrum size */
    if (spectrum_size < curr_filter->end)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    /* filter output */
    ret_val = dspInnerProduct(
                  spectrum + curr_filter->begin,
                  curr_filter->coeff, 
                  curr_filter->end - curr_filter->begin + 1);

    /* average result */
    return ret_val / (curr_filter->end - curr_filter->begin + 1) ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Check filter bank size, resize if necessary

If necessary, based on current number of filters,
double the size of filter pointer table
until there are enough filter pointers
to accommodate to current number of filters.
\param filter_bank Filter bank
\return ESR_SUCCEED if new filter pointers are
succesfully allocated or ESR_FAILED in case of failure
\author Evgeny Karpov
*/
int CheckFilterBank(
	FilterBank * filter_bank) 
{
    Filter ** filters ;
    int i,size ;

    if (NULL == filter_bank)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if ((filter_bank->size < 0)
          || (filter_bank->size > 65536))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if ((filter_bank->num_filters < 0)
          || (filter_bank->num_filters > 65536))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    while (filter_bank->num_filters >= filter_bank->size)
    {
        /* double the size of filter array */
        filters = realloc(filter_bank->filters,
                            2 * filter_bank->size * sizeof(Filter*));
        if (NULL == filters)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            return ESR_FAILED ;
        }

        size = 2 * filter_bank->size ;
        for (i = filter_bank->size; i < size; i++)
        {
            filters[i] = NULL ;
        }

        filter_bank->filters = filters ;
        filter_bank->size *= 2 ;
    }

    return ESR_SUCCEED ;
}


/*!
\brief Create a new filter

Allocate and return a new Filter 
and set its coefficients and its passed frequency indices
\param begin Index of the first passed harmonic
frequency in the fourier spectrum
\param end Index of the last passed harmonic
frequency in the fourier spectrum
\param coeff Pointer to the filter coefficients
\return Pointer to the created filter
or NULL if the operation fails
\sa Filter
\author Evgeny Karpov
*/
Filter * CreateFilter(int begin, int end, INT16* coeff)
{
    Filter * filter ;
    int i;

    if ((begin < 0) || (end < begin) || (NULL == coeff))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }

    for (i = 0; i < end - begin + 1; i++)
    {
        if (coeff[i] < 0.0)
        {
            ReportError(ESR_INVALID_PARAMETER, NULL);
            return NULL;
        }
    }

    if ((filter = malloc(sizeof(Filter)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL ;
    }

    filter->coeff = malloc(sizeof(INT16)*(end - begin + 1));
    if (NULL == filter->coeff)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        free(filter) ;
        return NULL ;
    }

    filter->begin = begin;
    filter->end = end;

    for (i = 0; i < end - begin + 1; i++)
    {
        filter->coeff[i] = coeff[i];
    }

    return filter ;
}


int fbPrintFilter(FILE* f, int begin, int end, INT16* coeff)
{
	int i;

	if (NULL == coeff)
	{
		ReportError(ESR_INVALID_PARAMETER, NULL);
		return ESR_FAILED;
	}
	if ((begin < 0) || (end < 0) || (end < begin))
	{
		ReportError(ESR_INVALID_PARAMETER, NULL);
		return ESR_FAILED;
	}

	fprintf(f, "    begin: %d, end: %d, coefficients:", begin, end);
	for (i = 0; i < end - begin + 1; i++)
	{
		fprintf(f, " %d", coeff[i]);
	}
	fprintf(f, "\n");

	return ESR_SUCCEED;
}

/* print filter bank to stdout */
int fbPrint(FILE* f, HFBANK fbank)
{
	FilterBank* fb;
	int i;

	fb = (FilterBank*) fbank;

	if (NULL == fb)
	{
		ReportError(ESR_INVALID_PARAMETER, NULL);
		return ESR_FAILED;
	}

	fprintf(f, "Filter Bank:\n");
	fprintf(f, "  size: %d, num_filters: %d, average: %d, filters:\n",
	  fb->size, fb->num_filters, fb->average);
	for (i = 0; i < fb->num_filters; i++)
	{
		if (fbPrintFilter(f,
		      fb->filters[i]->begin,
		      fb->filters[i]->end,
		      fb->filters[i]->coeff) != ESR_SUCCEED)
		{
			ReportError(ESR_FILTER_BANK_PRINT_FILTER, NULL);
			return ESR_FAILED;
		}
	}

	return ESR_SUCCEED;
}

/*======================== Functions ==============================*/
