/*!
\file mfcc.c
\brief Mel-frequency cepstral coefficients implementation


Implementation of routines for extraction
of MFCC coefficients from a magnitude spectrum\n
$Id: mfcc.c,v 1.12 2004/01/08 13:02:29 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32
\date 2003
\author Evgeny Karpov
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <limits.h>


#include <fixmath.h>
#include <errorapi.h>
#include <fvector.h>


#include "fftdct.h"
#include "melfilters.h"
#include "mfcc.h"


#if !defined(M_PI)
/*! Define pi if not already defined */
#define M_PI 3.14159265358979323846
#endif


/*======================== Structures =============================*/

/*!
\brief MFCC Extraction
*/
typedef struct {

    int order ;           /*!< vector order */
    int num_mel_filters ; /*!< number of mel-filters */
    int sampling_freq ;   /*!< sampling frequency */

    int win_size ;  /*!< window size */
    int spec_type ; /*!< spectrum type */

    INT16 * input_data ;   /*!< temporal array to store copy of input */
    INT32 * spectrum ;     /*!< temporal array to store spectrum */
    INT16 * spectrum16 ;   /*!< temporal array to store spectrum */
    INT32 * filtered_spec ;/*!< filtered spectrum */
    INT16 * window_coeff ; /*!< window coefficients */

    FFTProcessor* fft; /*!< FFT processor for computing spectra */
    DCTProcessor* dct; /*!< DCT processing data structures */
    DCTProcessor* dct0; /*!< DCT processing data structures for
                             cosine transforming the mel-spectrum
                             without 0'th coefficient*/

    HFBANK  fbank ;       /*!< filter bank for mel-filtering */

    /* pre-processing */
    PreProc * data_preproc ; /*!< input data pre-processor */
    PreProc * spectrum_preproc ;     /*!< Spectrum processor */
    PreProc * mel_spectrum_preproc ; /*!< Mel-spectrum processor */
    PreProc * cepstrum_preproc ;     /*!< Cepstrum processor */

    PreProc32 * spectrum_preproc32 ;     /*!< Spectrum processor */
    PreProc32 * mel_spectrum_preproc32 ; /*!< Mel-spectrum processor */
    PreProc32 * cepstrum_preproc32 ;     /*!< Cepstrum processor */

    void    * coeff_map ; /*!< map of needed coefficients */

    short flags ; /*!< flags */

#if defined(STORE_SIGNAL)
#define SIGNAL_FILE_NAME "signal_INT16.txt"
    FILE *f_signal;
#endif

#if defined(STORE_PPSIGNAL)
#define PPSIGNAL_FILE_NAME "ppsignal_INT16.txt"
    FILE *f_ppsignal;
#endif

#if defined(STORE_WSIGNAL)
#define WSIGNAL_FILE_NAME "wsignal_INT16.txt"
    FILE *f_wsignal;
#endif

#if defined(STORE_SPECTRA)
#define SQUARED_SPECTRA_FILE_NAME "squared_spectra_INT16.txt"
    FILE *f_spectra;
#endif

} MFCCHandle ;

/*======================== Structures =============================*/

/*======================== Functions ==============================*/


/*!
\brief Create MFCC processor

Create data structures used for MFCC computations.
This includes preparing FFT computation, creating
filterbanks and such.
\param params Adjustable MFCC computation parameters
\return New MFCC processor in a \p HMFCC handle
\bug This may have a fundamental error:
the part of FFT output spectrum that should
be used in the mel-filtering is the first N/2 + 1
elements, not N elements?
\todo CHANGE TO memcpy (coeff_map assignment)
\sa FFT
\sa HMFCC
\sa MFCCParams
\author Evgeny Karpov
*/
HMFCC mfccCreate(MFCCParams * params) 
{
    MFCCHandle * mfcc ;
    int fftsize;

    if (params==NULL
          || params->order <= 0 || params->num_mel_filters <= 0
          || params->win_size <= 0 || params->sampling_freq <= 0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return NULL ;
    }

    mfcc = malloc(sizeof(MFCCHandle));
    if (NULL == mfcc)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
        return NULL ;
    }

    /* initialize internal pointers */
    mfcc->window_coeff = NULL;
    mfcc->filtered_spec = NULL;
    mfcc->input_data = NULL;
    mfcc->spectrum = NULL;
    mfcc->spectrum16 = NULL;

    mfcc->fft = NULL;
    mfcc->dct = NULL;
    mfcc->dct0 = NULL;

    /* copy parameters */
    mfcc->order           = params->order ;
    mfcc->num_mel_filters = params->num_mel_filters ;
    mfcc->sampling_freq   = params->sampling_freq ;
    mfcc->win_size        = params->win_size ;
    mfcc->flags           = params->flags ;

    mfcc->spec_type = params->spec_type;
    if (0 == mfcc->spec_type)
    {
        mfcc->spec_type = MFCC_SPECTRUM_MAGNITUDE ; /* by default */
    }

    /* pre-processing parameters */
    if (params->flags & MFCC_FILTER_DATA) 
    {
        mfcc->data_preproc = params->data_preproc ;
    }
    if (params->flags & MFCC_FILTER_SPECTRUM) 
    {
        mfcc->spectrum_preproc = params->spectrum_preproc ;
        mfcc->spectrum_preproc32 = params->spectrum_preproc32 ;
    }
    if (params->flags & MFCC_FILTER_MELSPECTRUM) 
    {
        mfcc->mel_spectrum_preproc = params->mel_spectrum_preproc ;
        mfcc->mel_spectrum_preproc32 = params->mel_spectrum_preproc32 ;
    }
    if (params->flags & MFCC_FILTER_CEPSTRUM)
    {
        mfcc->cepstrum_preproc = params->cepstrum_preproc ;
        mfcc->cepstrum_preproc32 = params->cepstrum_preproc32 ;
    }
    if (params->flags & MFCC_USE_COEFF_MAP) 
    {
        mfcc->coeff_map = params->coeff_map ; /* CHANGE TO memcpy !!! */
    }

    /* initial */

    /* compute FFT size, it will be
       the next power of two starting from win_size */
    /* win_size is positive so this function will not fail */
    fftsize = mfccEstimateFftwSize(mfcc->win_size) ;

    mfcc->fft = fftCreate(fftsize);
    if (NULL == mfcc->fft)
    {
        ReportError(ESR_FFT_CREATE, NULL);
        mfccDestroy(mfcc) ;
        return NULL;
    }

    mfcc->dct = dctCreate(mfcc->num_mel_filters, mfcc->order);
    if (NULL == mfcc->dct)
    {
        ReportError(ESR_DCT_CREATE, NULL);
        mfccDestroy(mfcc) ;
        return NULL;
    }

    mfcc->dct0 = dctCreate(mfcc->num_mel_filters - 1, mfcc->order);
    if (NULL == mfcc->dct0)
    {
        ReportError(ESR_DCT_CREATE, NULL);
        mfccDestroy(mfcc) ;
        return NULL;
    }

    if (params->flags & MFCC_USE_PROVIDED_FILTER_BANK)
    {
        /* filter bank was provided in the parameters */
        mfcc->fbank = params->fbank ;

        /* get number of filters from provided filter bank */
        mfcc->num_mel_filters = fbGetSize(mfcc->fbank) ;
        if (ESR_FAILED == mfcc->num_mel_filters)
        {
            ReportError(ESR_FILTER_BANK_SIZE, NULL);
            mfccDestroy(mfcc) ;
            return NULL;
        }
    }
    else
    {
        /* filter bank was not provided, create one */
        mfcc->fbank = melfbCreate(
                        mfcc->num_mel_filters,
                        mfcc->sampling_freq,
                        fftsize);
        if (NULL == mfcc->fbank)
        {
            ReportError(ESR_MEL_FILTER_BANK_CREATE, NULL);
            mfccDestroy(mfcc) ;
            return NULL ;
        }

        if (fbPrint(stderr, mfcc->fbank) != ESR_SUCCEED)
        {
            ReportError(ESR_FILTER_BANK_PRINT, NULL);
            mfccDestroy(mfcc) ;
            return NULL;
        }
    }

    mfcc->filtered_spec = malloc(sizeof(INT32)
                                   * mfcc->num_mel_filters);
    if (NULL == mfcc->filtered_spec)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }

    /* fill window coefficients */
    if (params->window_func != NULL)
    {
        mfcc->window_coeff = malloc(sizeof(INT16)*mfcc->win_size);
        if (NULL == mfcc->window_coeff)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            mfccDestroy(mfcc) ;
            return NULL ;
        }

        /* compute the window function values */
        params->window_func(mfcc->window_coeff, mfcc->win_size);
    }
    else
    {
        mfcc->window_coeff = NULL ;
    }

    /* check if spectrum buffer should be shared */
    if (params->flags & MFCC_USE_SHARED_SPECTRUM)
    {
        mfcc->spectrum = params->shared_buffer ;
    }
    else
    {
        mfcc->spectrum = (INT32*) malloc(sizeof(INT32)
                                            * (fftsize / 2 + 1)) ;
        if (mfcc->spectrum==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            mfccDestroy(mfcc) ;
            return NULL ;
        }
    }

    /* temporary array for input vector */
    mfcc->input_data = (INT16*) malloc(sizeof(INT16) * fftsize);
    if (NULL == mfcc->input_data)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }

    /* temporary array for 16-bit resolution spectrum */
    mfcc->spectrum16 = (INT16*) malloc(sizeof(INT16)
                                         * (fftsize / 2 + 1)) ;
    if (mfcc->spectrum16==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }

#if defined(STORE_SIGNAL)
	mfcc->f_signal = fopen(SIGNAL_FILE_NAME, "w");
    if (NULL == mfcc->f_signal)
    {
        ReportError(ESR_CANT_OPEN_FILE, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }
#endif

#if defined(STORE_PPSIGNAL)
	mfcc->f_ppsignal = fopen(PPSIGNAL_FILE_NAME, "w");
    if (NULL == mfcc->f_ppsignal)
    {
        ReportError(ESR_CANT_OPEN_FILE, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }
#endif

#if defined(STORE_WSIGNAL)
	mfcc->f_wsignal = fopen(WSIGNAL_FILE_NAME, "w");
    if (NULL == mfcc->f_wsignal)
    {
        ReportError(ESR_CANT_OPEN_FILE, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }
#endif

#if defined(STORE_SPECTRA)
	mfcc->f_spectra = fopen(SQUARED_SPECTRA_FILE_NAME, "w");
    if (NULL == mfcc->f_spectra)
    {
        ReportError(ESR_CANT_OPEN_FILE, NULL);
        mfccDestroy(mfcc) ;
        return NULL ;
    }
#endif

    return (HMFCC)mfcc ;
}


/*!
\brief Destroy MFCC processor

Free all resources used by the MFCC processor
\param hMFCC Handle to the MFCC processor that is destroyed
\sa HMFCC
\author Evgeny Karpov
*/
void mfccDestroy(HMFCC hMFCC) 
{
    MFCCHandle * mfcc ;

    mfcc = (MFCCHandle*) hMFCC;
    if (NULL == mfcc)
    {
        return ;
    }

#if defined(STORE_SIGNAL)
	fclose(mfcc->f_signal);
#endif
#if defined(STORE_PPSIGNAL)
	fclose(mfcc->f_ppsignal);
#endif
#if defined(STORE_WSIGNAL)
	fclose(mfcc->f_wsignal);
#endif
#if defined(STORE_SPECTRA)
	fclose(mfcc->f_spectra);
#endif

    free(mfcc->spectrum16);
    free(mfcc->input_data);

    dctDestroy(mfcc->dct0);
    dctDestroy(mfcc->dct);
    fftDestroy(mfcc->fft);

    free(mfcc->window_coeff) ;
    free(mfcc->filtered_spec) ;

    if (!(mfcc->flags & MFCC_USE_SHARED_SPECTRUM))
    {
        free(mfcc->spectrum) ;
    }
    if (!(mfcc->flags & MFCC_USE_PROVIDED_FILTER_BANK))
    {
        fbDestroy(mfcc->fbank) ;
    }

    free(mfcc) ;
}


/*-----------------------------------------------------------------*/



/*!
\brief Extract MFCC coefficient from a signal frame

This computes all stages involved in MFCC feature
extraction.
\param hMFCC MFCC processor
\param frame The signal vector data
\param size The signal data size
\return Pointer to the MFCC vector data
\bug (1) comment says "pre-process mel-spectrum
but it is the spectrum that is pre-processed
\todo (2) why is the filtered spectrum depending
on the magnitude spectrum flag? What does it mean
to process filterbank spectrum or filterbank magnitude
spectrum in practise? References? 
\bug hard coded nbits for fractional sqrt/log2
\bug scaling of filter bank output spectrum is not optimal
\note The calling function must know the length
of the output vector in advance or deduce
it from the MFCC processor parameters.
\todo Return the MFCC vector in some other
format where the vector length is explicitly shown.
\todo ADD WORKING WITHOUT FILTERBANK,
SET UP POINTERS CORRECTLY DURING INITIALIZATION
\todo WHAT ABOUT FILTERING BEFORE DCT ?
\todo ADD WORKING WITH COEFFICIENT MAP,
MAP - INTEGER ARRAY, INDEXES WHAT COEFF. TO INCLUDE
\todo Explain what standoff means and why used
\note Tries to protect input from overwriting but
may fail because of a bug in ppProcess
\sa HMFCC
\sa ppProcess
\author Evgeny Karpov
*/
INT32 * mfccExtractVector(HMFCC hMFCC, INT16 * frame, int size) 
/*
INT32 * mfccExtractVector(
    HMFCC hMFCC,
    INT16 * frame,
    int size,
    INT32* ret_vector)
*/
{
    int result;
    MFCCHandle * mfcc ;
    INT32 * ret_vector ;
    int standoff ;
    int i ;
    int fft_size;
    INT32 isr; /* integer part of square root of spectrum */
    UINT32 fsr; /* fractional part of square root of spectrum */

    mfcc = (MFCCHandle*) hMFCC;
    if (NULL == mfcc) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;    
    }

    if (NULL == mfcc->fft) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;    
    }
    fft_size = fftGetInputSize(mfcc->fft);

    if (fft_size > mfcc->win_size)
    {
        standoff = (fft_size - mfcc->win_size) / 2 ;
    }
    else
    {
        standoff = 0 ;
    }

    memset(mfcc->input_data,
        0, sizeof(INT16) * standoff) ;
    memcpy(mfcc->input_data + standoff,
        frame, sizeof(INT16) * size);
    memset(mfcc->input_data + standoff + size,
        0, sizeof(INT16) * (fft_size - standoff - size)) ;

#if defined(STORE_SIGNAL)
	for (i = 0; i < fft_size; i++)
	{
		fprintf(mfcc->f_signal, " %d", mfcc->input_data[i]);
	}
	fprintf(mfcc->f_signal, "\n");
#endif

    /* data pre-processing */
    if ((mfcc->flags & MFCC_FILTER_DATA)
          && (mfcc->data_preproc->size > 0) )
    {
        ppProcess(mfcc->data_preproc,
                  mfcc->input_data + standoff,
                  mfcc->input_data + standoff, size) ;
    }

#if defined(STORE_PPSIGNAL)
	for (i = 0; i < fft_size; i++)
	{
		fprintf(mfcc->f_ppsignal, " %d", mfcc->input_data[i]);
	}
	fprintf(mfcc->f_ppsignal, "\n");
#endif

    /* copy to FFTW and apply window coefficients */
    if (mfcc->window_coeff)
    {
        for (i = 0; (i < mfcc->win_size) && (i < size); i++) 
        {
            mfcc->input_data[i] =
              (INT16) ((mfcc->window_coeff[i]*mfcc->input_data[i])
                         / WINDOWING_SCALE);
        }
    }

#if defined(STORE_WSIGNAL)
	for (i = 0; i < fft_size; i++)
	{
		fprintf(mfcc->f_wsignal, " %d", mfcc->input_data[i]);
	}
	fprintf(mfcc->f_wsignal, "\n");
#endif

    /* compute spectrum */
    result = fftCompute(mfcc->fft, mfcc->input_data, mfcc->spectrum);
    if (result != ESR_SUCCEED)
    {
        ReportError(ESR_FFT_COMPUTE, NULL);
        return NULL;
    }

#if defined(STORE_SPECTRA)
	for (i = 0; i < fft_size / 2 + 1; i++)
	{
		fprintf(mfcc->f_spectra, " %ld", mfcc->spectrum[i]);
	}
	fprintf(mfcc->f_spectra, "\n");
#endif

    /* compute spectrum (if shared is not used) */
    if (!(mfcc->flags & (MFCC_SHARE_DATA | MFCC_USE_SHARED_SPECTRUM)))
    {
        if (mfcc->spec_type==MFCC_SPECTRUM_MAGNITUDE)
        {
            for (i = 0; i < fft_size / 2 + 1; i++)
            {
                /* FIXME hard coded "8 bits"
                   (needed in sqrt for log2) */
                isr = frac_sqrt(mfcc->spectrum[i], &fsr, 8);
                if (isr < 0)
                {
                    ReportError(ESR_SQRT, NULL);
                    return NULL;
                }
                mfcc->spectrum[i] = isr * FRAC_ONE + fsr;
            }
        }

        /*
		printf("magnitude spectrum:");
		for (i = 0; i < fft_size / 2 + 1; i++)
		{
			printf(" %ld", mfcc->spectrum[i]);
		}
		printf("\n");
        */
    }

    /* pre-process spectrum */
    if (mfcc->flags & MFCC_FILTER_SPECTRUM)
    {
        ppProcess32(mfcc->spectrum_preproc32, mfcc->spectrum,
                  mfcc->spectrum, fft_size / 2 + 1) ;
    }

    /* reduce spectrum resolution to 16 bits */
    for (i = 0; i < fft_size / 2 + 1; i++)
    {
        /* mfcc->spectrum16[i] = (INT16) (mfcc->spectrum[i] >> 16); */
        mfcc->spectrum16[i] = (INT16) (mfcc->spectrum[i] >> 15);
    }

    /*
	printf("16-bit resolution spectrum:");
	for (i = 0; i < fft_size / 2 + 1; i++)
	{
		printf(" %d", mfcc->spectrum16[i]);
	}
	printf("\n");
    */

    /* apply filter bank */
    if (mfcc->fbank)
    {
        result = fbApply(mfcc->fbank,
                   mfcc->spectrum16, fft_size / 2 + 1,
                   mfcc->filtered_spec, mfcc->num_mel_filters) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_FILTER_BANK_APPLY, NULL);
            return NULL;
        }
    }

    /*
	printf("filter bank spectrum:");
	for (i = 0; i < mfcc->num_mel_filters; i++)
	{
		printf(" %ld", mfcc->filtered_spec[i]);
	}
	printf("\n");
    */

    /* ADD WORKING WITHOUT FILTERBANK !!! */
    /* SET UP POINTERS CORRECTLY DURING INITIALIZATION */

    /* pre-process mel-spectrum */
    /* (1) */
    if (mfcc->flags & MFCC_FILTER_MELSPECTRUM)
    {
        ppProcess32(mfcc->mel_spectrum_preproc32,
          mfcc->spectrum, mfcc->spectrum, fft_size / 2 + 1) ;
    }

    /* compute log-filtered spectrum */
    /* (2) */
    for (i = 0; i < mfcc->num_mel_filters; i++)
    {
        isr = frac_log2(mfcc->filtered_spec[i] + 1, &fsr, 8);
        if (isr < 0)
        {
            ReportError(ESR_LOG2, NULL);
            return NULL;
        }
        mfcc->filtered_spec[i] = (isr * FRAC_ONE) + fsr;
    }
    if (mfcc->spec_type != MFCC_SPECTRUM_MAGNITUDE)
    {
        for (i = 0; i < mfcc->num_mel_filters; i++)
        {
            mfcc->filtered_spec[i] *= 2;
        }
    }

    /* At this point all of filtered_spec is less than
       2 * 32 * FRAC_ONE * (maximal filter coefficient sum) */

    /* scale down and copy to spectrum16 */
    for (i = 0; i < mfcc->num_mel_filters; i++)
    {
        /* FIXME the numerator must be computed 
           with help of a new filter bank routine MaxCoeffSum */
        mfcc->spectrum16[i] = (INT16) (mfcc->filtered_spec[i]
                                / (1 << 15));
    }

    /* WHAT ABOUT FILTERING BEFORE DCT ? */

    ret_vector = fvCreate(mfcc->order);
    if (NULL == ret_vector)
    {
        ReportError(ESR_VECTOR_CREATE, NULL);
        return NULL ;
    }

    /* ADD WORKING WITH COEFFICIENT MAP */
    /* MAP - INTEGER ARRAY, INDEXES WHAT COEFF. TO INCLUDE */

    /* discrete cosine transform, exclude coefficient zero or not */
    if (mfcc->flags & MFCC_INCLUDE_ZERO_COEFF)
    {
        result = dctCompute(mfcc->dct,
                   mfcc->spectrum16, ret_vector);
        if (result != ESR_SUCCEED)
        {
            ReportError(ESR_DCT_COMPUTE, NULL);
            return NULL;
        }
    }
    else
    {
        result = dctCompute(mfcc->dct0,
                   mfcc->spectrum16 + 1, ret_vector);
        if (result != ESR_SUCCEED)
        {
            ReportError(ESR_DCT_COMPUTE, NULL);
            return NULL;
        }
    }

    /* pre-process cepstrum (post-process?) */
    if (mfcc->flags & MFCC_FILTER_CEPSTRUM)
    {
        ppProcess32(mfcc->cepstrum_preproc32, ret_vector,
                    ret_vector, mfcc->order) ;
    }

    return ret_vector ;
}

/*-----------------------------------------------------------------*/


/*!
\brief Estimate the necessary power of two for radix-2 FFT

Compute the power of two needed for the spectrum
computation with radix-2 fast Fourier transformation.
\param win_length The signal frame length
\return The power of two needed for radix-2 FFT
\sa FFT
\author Evgeny Karpov
*/
int mfccEstimateFftwSize(int win_length) 
{
    int size ;

    if (win_length <= 0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1;
    }

    for (size = 1; size < win_length; size *= 2)
    {
        if (size > (INT_MAX >> 1))
        {
            ReportError(ESR_INVALID_PARAMETER, NULL);
            return -1;
        }
    }

    return size ;
}

/*-----------------------------------------------------------------*/


/*!
\brief Get pointer to MFCC processor data buffer

Get pointer to the FFT output buffer or the spectrum
data or NULL in case of failure
\param hMFCC MFCC processor
\param buff_id Which buffer pointer is needed,
MFCC_BUFF_FFTW returns pointer to FFT output buffer,
MFCC_BUFF_SPECTRUM returns pointer to the magnitude spectrum
\return Pointer to the buffer data
\sa HMFCC (a pointer to MFCCHandle)
\author Evgeny Karpov
*/
void * mfccGetBuffer(HMFCC hMFCC, int buff_id) 
{
    MFCCHandle * mfcc ;

    mfcc = (MFCCHandle*) hMFCC;
    if (NULL == mfcc) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;    
    }

    if (buff_id == MFCC_BUFF_FFTW)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }
    else if (buff_id == MFCC_BUFF_SPECTRUM)
    {
        if (NULL == mfcc->spectrum)
        {
            ReportError(ESR_INVALID_PARAMETER, NULL);
        }
        return (void*) mfcc->spectrum ;
    }

    ReportError(ESR_INVALID_PARAMETER, NULL);
    return NULL ;
}


/*======================== Functions ==============================*/
