/*!
\file mfcc.h
\brief Mel-frequency cepstral coefficients interface

Interface of routines for extraction
of MFCC coefficients from a magnitude spectrum\n
$Id: mfcc.h,v 1.7 2004/01/08 13:02:29 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32;
this may also change the interface, if the scaling will
be adjustable during run time.
\date 2003
\author Evgeny Karpov
*/

#ifndef MFCC_H
#define MFCC_H


#include <fixmath.h>


#include "dsp.h"
#include "winfunc.h"
#include "filterbank.h"
#include "preprocess.h"

/*-----------------------------------------------------------------*/

#if defined(__cplusplus)
extern "C" {
#endif


/*======================== Structures =============================*/

/*!
\brief Parameters for MFCC Algorithm
*/
typedef struct {
    int order ;           /*!< vector order */
    int num_mel_filters ; /*!< number of mel-filters */
    int win_size ;        /*!< window size (in samples) */
    int spec_type ;       /*!< spectrum type */
    int sampling_freq ;   /*!< sampling frequency */

    WINFUNC window_func ; /*!< window coefficients function */    
    HFBANK  fbank ;       /*!< filter bank */

    /* data pre-processing */
    PreProc * data_preproc ; /*!< input data pre-processor */

    /* additional processing */
    PreProc * spectrum_preproc ;     /*!< Spectrum processor */
    PreProc * mel_spectrum_preproc ; /*!< Mel-spectrum processor */
    PreProc * cepstrum_preproc ;     /*!< Cepstrum processor */

    PreProc32 * spectrum_preproc32 ;     /*!< Spectrum processor */
    PreProc32 * mel_spectrum_preproc32 ; /*!< Mel-spectrum processor */
    PreProc32 * cepstrum_preproc32 ;     /*!< Cepstrum processor */

    void * coeff_map ;    /*!< map of needed coefficients */
    short flags ;         /*!< flags */

    /* support for sharing extracted data */
    void * shared_buffer ;     /*!< buffer for shared data */

} MFCCParams ;

/*======================== Structures =============================*/

/*! 
\brief Abstract MFCC processor handle

In practise this will be a pointer to
a MFCCParams or MFCCHandle ??? MFCCParams ???
\todo Why there are two copies of
almost the same big struct: MFCCHandle, MFCCParams ???
\sa MFCCParams
\sa MFCCHandle
*/
typedef void * HMFCC ;

/*-----------------------------------------------------------------*/

#define MFCC_USE_DEFAULT                    0

/* Spectrum types */
#define MFCC_SPECTRUM_MAGNITUDE             1

/*-------------------------- Flags --------------------------------*/

/* first byte */
#define MFCC_INCLUDE_ZERO_COEFF             0x0001 
#define MFCC_USE_PROVIDED_FILTER_BANK       0x0002 
#define MFCC_FILTER_SPECTRUM                0x0004 
#define MFCC_FILTER_MELSPECTRUM             0x0008 
#define MFCC_FILTER_CEPSTRUM                0x0010 
#define MFCC_USE_COEFF_MAP                  0x0020 
#define MFCC_FILTER_DATA                    0x0040 
#define MFCC_PROTECT_INPUT                  0x0080 

/* second byte */
#define MFCC_SHARE_DATA                     0x0100 

#define MFCC_USE_SHARED_FFTW                0x0400 
#define MFCC_USE_SHARED_SPECTRUM            0x0800 

/*-------------------------- Flags --------------------------------*/

/* internal buffers */
#define MFCC_BUFF_FFTW                      1
#define MFCC_BUFF_SPECTRUM                  2

/*======================== Functions ==============================*/

/* creation/deletion */
HMFCC mfccCreate(MFCCParams * params) ;
void mfccDestroy(HMFCC hMFCC) ;

/* vector extraction */
INT32 * mfccExtractVector(HMFCC hMFCC, INT16 * frame,int size) ;
/*
INT32 * mfccExtractVector(
    HMFCC hMFCC,
    INT16* frame,
    int size,
    INT32* ret_vector) ;
*/

/* parameter estimation */
int mfccEstimateFftwSize(int win_length) ;

/* working with internal buffers */
void * mfccGetBuffer(HMFCC hMFCC,int buff_id) ;

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* MFCC_H */

