/*!
\file preprocess.h
\brief Signal pre-processing interface

Interface for working with 
signal preprocessing before feature extraction\n
$Id: preprocess.h,v 1.3 2003/12/29 14:27:26 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32;
this may also change the interface, if the scaling will
be adjustable during run time.
\date 2003
\author Evgeny Karpov
*/

#ifndef PREPROCESS_H
#define PREPROCESS_H


#include <fixmath.h>

    
#if defined(__cplusplus)
extern "C" {
#endif


/*-----------------------------------------------------------------*/

/*!
\brief Signal Processing Function Pointer.

All signal pre-processing functions use the same
parameters and return an integer. The parameters
are input signal \p input, output signal \p output,
and the length of both signals \p size.
\sa PreProcFunc
\sa PreProc
*/
typedef int (*PREPROCFUN) (
	INT16 * input,
	INT16 * output,
	int size,
	void * params) ;
typedef int (*PREPROCFUN32) (
	INT32 * input,
	INT32 * output,
	int size,
	void * params) ;

/*======================== Structures =============================*/

/*!
\brief Pre-processing Function and Parameters.

A Helper structure that holds a signal processing
function pointer and pointer to the parameters
for that function.
\sa PREPROCFUN
\sa PreProc
*/
typedef struct {
    PREPROCFUN   func ;        /*!< The signal processing function */
    void       * func_params ; /*!< Parameters of the function */
} PreProcFunc ;
typedef struct {
    PREPROCFUN32   func ;        /*!< The signal processing function */
    void       * func_params ; /*!< Parameters of the function */
} PreProcFunc32 ;

/*!
\brief A Signal Pre-Processor

A signal pre-processor consists of one or more
signal processing stages. The number of stages
is defined by \p size. 
\sa PREPROCFUN
\sa PreProcFunc
*/
typedef struct {
    int size ;            /*!< Number of pre-processing functions */
    PreProcFunc * funcs ; /*!< Pre-processing function pointer table */
} PreProc ;
typedef struct {
    int size ;            /*!< Number of pre-processing functions */
    PreProcFunc32 * funcs ; /*!< Pre-processing function pointer table */
} PreProc32 ;

/*======================== Structures =============================*/

/*======================== Functions ==============================*/


/*
Create Pre-Processor
*/
PreProc * ppCreate() ;
PreProc32 * ppCreate32() ;


/*
Destroy Pre-Processor
*/
void ppDestroy(
	PreProc * pproc) ;
void ppDestroy32(
	PreProc32 * pproc) ;


/*
Pre-Processor Setup
*/
int ppAddFunction(
	PreProc * pproc,
	PREPROCFUN pproc_func,
	void * func_params) ;
int ppAddFunction32(
	PreProc32 * pproc,
	PREPROCFUN32 pproc_func,
	void * func_params) ;


/*
Process the Data
*/
int ppProcess(
	PreProc * pproc,
	INT16 * input,
	INT16 * output,
	int size) ;
int ppProcess32(
	PreProc32 * pproc,
	INT32 * input,
	INT32 * output,
	int size) ;

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* PREPROCESS_H */
