/*!
\file sremoval.c
\brief Silence removal implementation

Implementation of several algorithms
for silence removal.\n
$Id: sremoval.c,v 1.2 2003/12/29 14:26:34 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
/*
#include <math.h>
#include <float.h>
*/
#include <fixmath.h>


#include <errorapi.h>


#include "dsp.h"
#include "sremoval.h"

/*-----------------------------------------------------------------*/

#ifndef max
/*! \brief Resolve maximum of two items */
#define max(a,b)    (((a) > (b)) ? (a) : (b))
/*! \brief Resolve minimum of two items */
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

/*-----------------------------------------------------------------*/



/*!
\brief Silence removal training/silence detection function pointer type

All silence remover functions take a SRHandle
handle to silence remover, a signal frame vector,
and size of the frame. This includes functions
used to train silence removal as well as silence detection.
\param handle Pointer to silence remover 
\param frame Pointer to input vector, the signal
\param size Length of the input vector
\return 1 for successful training or detected silence from signal,
0 for detecting non-silence, ESR_FAILED for failure in all cases.
\sa SRHandle
\author Evgeny Karpov
*/
typedef int (*SRFUNC) (
	void * handle,
	INT16 * frame,
	int size) ;


/*======================== Structures =============================*/


/*!
\brief Silence Remover Handle

Handle to a silence removal technique structure.
\sa SRTypes
\sa SRHandle
\sa SRFUNC
\author Evgeny Karpov
*/
typedef struct {

    SRTypes type ; /*!< The type of silence removal used */

    void * alg_handle ; /*!< The silence remover, either
	                         a SRHandle_BPBP or SRHandle_KINN */

    SRFUNC train_func ; /*!< silence check training function */
    SRFUNC check_func ; /*!< silence checking function */

} SRHandle ;


/*!
\brief BPBP Silence Remover Parameters

Parameters used in SRTYPE_BPBP type of silence removal
\sa SRTypes
\sa SRTYPE_BPBP
\sa SRHandle
\author Evgeny Karpov
*/
typedef struct {

    INT32 en_max ; /*!< \todo Describe this */
    INT32 en_min ; /*!< \todo Describe this */

    INT32 SL ; /*!< \todo Describe this */
    int sl_count ; /*!< \todo Describe this */

    INT32 thr1 ; /*!< \todo Describe this */
    INT32 thr2 ; /*!< \todo Describe this */

    int initial ; /*!< \todo Describe this */

} SRHandle_BPBP ;


/*!
\brief Kinnunen Silence Remover Parameters

Parameters used in SRTYPE_KINN type of silence removal
\sa SRTypes
\sa SRTYPE_KINN
\sa SRHandle
\author Evgeny Karpov
*/
typedef struct {

    INT32 en_max ; /*!< \todo Describe this */
    INT32 en_min ; /*!< \todo Describe this */

    INT32 eta ; /*!< \todo Describe this */
    INT32 thr ; /*!< \todo Describe this */

    int initial ; /*!< \todo Describe this */

} SRHandle_KINN ;


/*======================== Structures =============================*/


/* BPBP */
int sr_train_bpbp(
	void * handle,
	INT16 * frame,
	int size) ;
int sr_check_bpbp(
	void * handle,
	INT16 * frame,
	int size) ;

/* KINN */
int sr_train_kinn(
	void * handle,
	INT16 * frame,
	int size) ;
int sr_check_kinn(
	void * handle,
	INT16 * frame,
	int size) ;

/*======================== Functions ==============================*/


/*!
\brief Create a New Silence Remover

Create a new silence remover of the specified
type and with the specified parameters.
\param type Type of the silence remover
\param params Silence remover parameters
\return Handle to a new silence remover or
NULL in case of failure
\sa HSREM
\sa SRTypes
\author Evgeny Karpov
*/
HSREM srCreate(
	SRTypes type,
	void * params) 
{
    SRHandle * srp ;

    if ((srp = malloc(sizeof(SRHandle)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
        return NULL ;
    }

    srp->type = type ;

    switch(type)
    {
    case SRTYPE_BPBP:

        if ((srp->alg_handle = malloc(sizeof(SRHandle_BPBP)))==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
            free(srp) ;
            return NULL ;
        }

        /* initial parameters */
        ((SRHandle_BPBP*)srp->alg_handle)->en_max = -1 ;
        ((SRHandle_BPBP*)srp->alg_handle)->en_min = -1 ;
        ((SRHandle_BPBP*)srp->alg_handle)->SL = 0 ;
        ((SRHandle_BPBP*)srp->alg_handle)->sl_count = 0 ;
        ((SRHandle_BPBP*)srp->alg_handle)->initial = 1 ;

        if (params) 
        {
            ((SRHandle_BPBP*)srp->alg_handle)->thr1 = ((SRParams_BPBP*)params)->thr1 ;
            ((SRHandle_BPBP*)srp->alg_handle)->thr2 = ((SRParams_BPBP*)params)->thr2 ;
        
        } else {

            ((SRHandle_BPBP*)srp->alg_handle)->thr1 = -1 ;
            ((SRHandle_BPBP*)srp->alg_handle)->thr2 = -1 ;
        }

        srp->train_func = sr_train_bpbp ;
        srp->check_func = sr_check_bpbp ;

        break ;

    case SRTYPE_KINN:

        if ((srp->alg_handle = malloc(sizeof(SRHandle_KINN)))==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
            free(srp) ;
            return NULL ;
        }

        /* initial parameters */
        ((SRHandle_KINN*)srp->alg_handle)->en_max = -1 ;
        ((SRHandle_KINN*)srp->alg_handle)->en_min = -1 ;
        ((SRHandle_KINN*)srp->alg_handle)->thr = -1 ;
        ((SRHandle_KINN*)srp->alg_handle)->eta = 1 ;

        ((SRHandle_KINN*)srp->alg_handle)->initial = 1;

        if (params)
        {
            ((SRHandle_KINN*)srp->alg_handle)->eta = *((INT32*)params) ;
        }

        srp->train_func = sr_train_kinn ;
        srp->check_func = sr_check_kinn ;

        break ;

    default:
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        free(srp) ;
        return NULL ;
    }

    return (HSREM) srp ;
}


/*!
\brief Destroy a Silence Remover

Destroy the silence remover \p sr
\param sr Handle to the silence remover that needs to be destroyed
\sa HSREM
\author Evgeny Karpov
*/
void srDestroy(
	HSREM sr) 
{
    SRHandle * srp ;

    if ((srp = (SRHandle*)sr)==NULL)
    {
        return ;
    }

    /* add special freeing stuff if needed */

    free(srp->alg_handle) ;
    free(sr) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Silence Remover Training

Training adjusts the silence remover parameters.
\param sr Handle to the silence remover to be trained
\param frame Input signal frame to training
\param size Size of the input frame
\sa HSREM
\author Evgeny Karpov
 */
int srTrain(
	HSREM sr,
	INT16 * frame,
	int size) 
{
    int result;

    if (sr==NULL || frame==NULL || size <= 0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return 0 ;
    }

    result = ((SRHandle*)sr)->train_func(
               ((SRHandle*)sr)->alg_handle, frame, size);
    if (ESR_FAILED == result)
    {
        ReportError(ESR_SILENCE_REMOVAL_TRAIN, NULL);
    }
    return result;
}


/*-----------------------------------------------------------------*/


/*!
\brief Silence Checking

Check if the input frame is silence or not according
to the given silence remover \p sr.
\param sr Handle to a trained silence remover technique
\param frame Input signal frame to check for silence
\param size Size of the input frame
\return 1 for silence, 0 for non-silence,
ESR_FAILED in case of failure
\sa SRFUNC, HSREM
\author Evgeny Karpov
*/
int srIsSilent(
	HSREM sr,
	INT16 * frame,
	int size) 
{
    int result;

    if (sr==NULL || frame==NULL || size <= 0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return ESR_FAILED ;
    }

    result = ((SRHandle*)sr)->check_func(
               ((SRHandle*)sr)->alg_handle, frame, size);
    if (ESR_FAILED == result)
    {
        ReportError(ESR_SILENCE_REMOVAL_DETECT, NULL);
    }
    return result;
}


/*!
\brief Start the next silence removal with the initial state

This function tells a silence remover \p sr
to drop to initial state, i.e.
the next time silence is checked the algorithm
starts with the initial state.
\param sr Handle to a silence remover
\return 1 for success, ESR_FAILED for failure
\todo MAKE "initial" A MEMBER OF SRHandle !!!
\author Evgeny Karpov
*/
int srSetInitial(
	HSREM sr) 
{
    SRHandle * srp ;

    srp = (SRHandle*) sr;
    if (NULL == srp)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return ESR_FAILED ;
    }

    /* MAKE "initial" A MEMBER OF SRHandle !!! */

    if (srp->type == SRTYPE_BPBP)
        ((SRHandle_BPBP*)srp->alg_handle)->initial = 1 ;
    else if (srp->type == SRTYPE_KINN)
        ((SRHandle_KINN*)srp->alg_handle)->initial = 1;

    return 1 ;
}


/*-----------------------------------------------------------------*/



/*!
\brief BPBP Silence Remover Training

Train a silence remover of the type SRTYPE_BPBP
\param handle Handle to the silence remover
\param frame Signal frame to train
\param size Length of the frame data
\return 1 for success, ESR_FAILED in case of failure
\bug hard coded nbits for fractional log2
\sa SRTypes
\sa SRTYPE_BPBP
\author Evgeny Karpov
*/
int sr_train_bpbp(
	void * handle,
	INT16 * frame,
	int size) 
{
    SRHandle_BPBP * bpbp ;
    INT32 frame_pow ;

    INT32 energy;
    UINT32 frac;


    bpbp = (SRHandle_BPBP*) handle ;
    if ((NULL == bpbp) || (NULL == frame) || (size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* compute frame energy */
    frame_pow = dspEnergy(frame,size) ;

    if (bpbp->initial)
    {
        bpbp->en_max = frame_pow ;
        bpbp->en_min = frame_pow ;

        bpbp->initial = 0 ;

    } else {

        bpbp->en_max = max(bpbp->en_max, frame_pow) ;
        bpbp->en_min = min(bpbp->en_min, frame_pow) ;
    }

    /* compute thresholds */
    /*
    bpbp->thr1 = bpbp->en_min
                   * (1 + 2 * log10(bpbp->en_max / bpbp->en_min)) ;
    */
    /*
    bpbp->thr1 = bpbp->en_min
                   *(1 + 2 * fix_log10(bpbp->en_max / bpbp->en_min)) ;
    */

    /* FIXME hard coded nbits */
    energy = frac_log2(bpbp->en_max / bpbp->en_min, &frac, 8);
    if (energy < 0)
    {
        ReportError(ESR_LOG2, NULL);
        return -1;
    }
    bpbp->thr1 = bpbp->en_min * (FRAC_ONE + 2 * (energy * FRAC_ONE + frac));

    if (frame_pow>bpbp->thr1)
    {
        bpbp->sl_count++ ;
        bpbp->SL += frame_pow ;
    }

    if (bpbp->sl_count)
    {
        bpbp->thr2 = (4 * bpbp->thr1
                       + (bpbp->SL / bpbp->sl_count
                                   - bpbp->thr1)) / 4 ;
    }
    else
    {
        bpbp->thr2 = (4 * bpbp->thr1 - bpbp->thr1) / 4 ;
    }

    return 1 ;
}


/*!
\brief BPBP Silence Checking

Check if input signal is silence or not according to
the given silence \p SRTYPE_BPBP type of remover \p handle.
\param handle Handle to the silence remover
\param frame Input signal
\param size Length of the input signal
\return 1 if signal is silence, 0 if signal is not silent,
ESR_FAILED in case of failure
\todo ADD CALCULATION OF ZCR !!!
\todo SIMPLE ASSUMPTION => REFINE ALGORITHM !!!
\sa SRTypes
\sa SRTYPE_BPBP
\author Evgeny Karpov
*/
int sr_check_bpbp(
	void * handle,
	INT16 * frame,
	int size) 
{
    SRHandle_BPBP * bpbp ;
    INT32 frame_pow ;

    bpbp = (SRHandle_BPBP*) handle ;
    if ((NULL == bpbp) || (NULL == frame) || (size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if (bpbp->initial)
    {
        return 0 ;
    }

    /* compute frame energy */
    frame_pow = dspEnergy(frame,size) ;

    /* ADD CALCULATION OF ZCR !!! */

    /* SIMPLE ASSUMPTION => REFINE ALGORITHM !!! */
    if (frame_pow < bpbp->thr1)
    {
        return 1 ;
    }

    return 0 ;
}


/*-----------------------------------------------------------------*/



/*!
\brief Signal Energy for Kinnunen Silence Removal

Measure energy of a signal frame in a manner suitable for
silence remover of type \p SRTYPE_KINN.
The returned energy is presented in the decibel scale.
\param frame Input signal frame
\param size Input signal length
\return The measured energy, negative value
is returned in case of failure
\bug hard coded nbits for fractional log2
\sa SRTypes
\sa SRTYPE_KINN
\author Evgeny Karpov
*/
INT32 sr_energy_kinn(
	INT16 * frame,
	int size)
{
    INT32 energy;
    UINT32 frac;

    if ((NULL == frame) || (size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1;
    }

    /*
    return 10 * fix_log10(dspEnergy(frame,size)) ;
    */
    /* FIXME hard coded nbits */
    energy = frac_log2(dspEnergy(frame, size), &frac, 8);
    if (energy < 0)
    {
        ReportError(ESR_LOG2, NULL);
        return -1;
    }

    return 2 * (energy * FRAC_ONE + frac);
}


/*!
\brief Kinnunen Silence Remover Training

Train a silence remover of the type SRTYPE_KINN
\param handle Handle to the silence remover
\param frame Signal frame to train
\param size Length of the frame data
\return 1 means successful training, ESR_FAILED means failure
\sa SRTypes
\sa SRTYPE_KINN
\author Evgeny Karpov
*/
int sr_train_kinn(
	void * handle,
	INT16 * frame,
	int size) 
{
    SRHandle_KINN * kinn ;
    INT32 frame_pow ;

    kinn = (SRHandle_KINN*) handle ;
    if ((NULL == kinn) || (NULL == frame) || (size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* compute frame energy */
    frame_pow = sr_energy_kinn(frame,size) ;

    if (kinn->initial)
    {
        kinn->en_max = frame_pow ;
        kinn->en_min = frame_pow ;

        kinn->initial = 0 ;

    } else {

        kinn->en_max = max(kinn->en_max,frame_pow) ;
        kinn->en_min = min(kinn->en_min,frame_pow) ;
    }

    /* compute thresholds */
    kinn->thr = kinn->en_min + kinn->eta*(kinn->en_max - kinn->en_min) ;

    return 1 ;
}


/*!
\brief Kinnunen Silence Checking

Check if input signal is silence or not according to
the given silence \p SRTYPE_KINN type of remover \p handle.
\param handle Handle to the silence remover
\param frame Input signal
\param size Length of the input signal
\return 1 if signal is silence, 0 if signal is not silent,
ESR_FAILED in case of failure
\sa SRTypes
\sa SRTYPE_KINN
\author Evgeny Karpov
*/
int sr_check_kinn(
	void * handle,
	INT16 * frame,
	int size) 
{
    SRHandle_KINN * kinn ;
    INT32 frame_pow ;

    kinn = (SRHandle_KINN*) handle ;
    if ((NULL == kinn) || (NULL == frame) || (size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    if (kinn->initial)
    {
        return 0 ; 
    }

    /* compute frame energy */
    frame_pow = sr_energy_kinn(frame,size) ;

    if (frame_pow < kinn->thr)
    {
        return 1 ;
    }

    return 0 ;
}

/*======================== Functions ==============================*/
