/*!
\file sremoval.h
\brief Silence removal interface

Interface for working with several
algorithms for silence removal.\n
$Id: sremoval.h,v 1.2 2003/12/23 09:53:46 juhani Exp $
\todo The logarithms and square roots and all math functions
values must be scaled later after a proper analysis is done
about how they should be scaled when using INT16 or INT32;
this may also change the interface, if the scaling will
be adjustable during run time.
\date 2003
\author Evgeny Karpov
*/

#ifndef SREMOVAL_H
#define SREMOVAL_H


#include <fixmath.h>


#if defined(__cplusplus)
extern "C" {
#endif


/*======================== Structures =============================*/

/*!
\brief Silence Removal Algorithm Types

Classified silence removal algorithms.
*/
typedef enum {

    SRTYPE_BPBP, /*!< Burileanu, Pascalin, Burileanu and Puchiu */
    SRTYPE_KINN  /*!< Kinnunen algorithm */

} SRTypes ;

/*!
\brief BPBP Algorithm Params

Parameter type for silence removal algorithms of the type SRTYPE_BPBP.
\sa SRTypes
*/
typedef struct {

    INT32 thr1 ; /*!< first energy threshold */
    INT32 thr2 ; /*!< second energy threshold */

} SRParams_BPBP ;

/*======================== Structures =============================*/

/*!
\brief Silence Remover Handle

A handle is defined as a void pointer.
In reality it is a pointer to a silence remover
struct/object.
*/
typedef void * HSREM ;

/*======================== Functions ==============================*/

/*
Create a New Silence Remover
*/
HSREM srCreate(
	SRTypes type,
	void * params) ;

/*
Destroy a Silence Remover
*/
void srDestroy(
	HSREM sr) ;

/*
Silence Remover Training
 */
int srTrain(
	HSREM sr,
	INT16 * frame,
	int size) ;

/*
Silence Checking
*/
int srIsSilent(
	HSREM sr,
	INT16 * frame,
	int size) ;

/*
Start the next silence removal with the initial state
*/
int srSetInitial(
	HSREM sr) ;

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* SREMOVAL_H */
