/*! 
\file shortfft.h
\brief 16-bit int fixed point FFT interface.

This module uses a fixed point aritchmetic
Fast Fourier transformation implementation
to compute the Fourier power spectrum of a vector.
The code is created with the fftgen public domain C-code.
The old documentation coming with the fftgen package says
that the offcial fftgen archive is in
ftp://csc.canberra.edu.au/pub/ise/eyal/fft
but the code package was actually only available in
<a href="http://www.jjj.de/fft/fftgen.tgz">
http://www.jjj.de/fft/fftgen.tgz</a>.
\n
$Id: shortfft.h,v 1.4 2003/11/30 20:30:33 juhani Exp $
\sa Link to fftgen software and other FFT implementations
in www pages
<a href="http://www.jjj.de/fft/fftpage.html">
http://www.jjj.de/fft/fftpage.html</a>
\date 2003/11
\author Juhani Saastamoinen
*/
#ifndef SHORTFFT_H
#define SHORTFFT_H


/*!
\brief Maximum length of input array for shortfft
is 2 to the power 10, this macro defines the exponent 2
\note This value cannot be adjusted!
\sa FFT_N
*/
#define FFT_M 10
/*!
\brief Maximum length of input array for shortfft
is 2 to the power FFT_M, i.e. FFT_N = 1024
\sa FFT_M
*/
#define FFT_N (1 << FFT_M)


/*!
\brief Compute Fourier power spectrum of a short int input array

This routine uses the optimized functions generated
with the fftgen software to compute a radix-2 FFT
and the power spectrum of the input array.
\param n Number of short int values in the input array
and the optional xoutput array;
the value of n must be a power of two within the interval 2-1024
\param input Input array of n short int values that is FFT'd
\param output Output Fourier power spectrum of input;
only n/2+1 long int values are used to store the spectrum
\param xoutput Input array after the FFT computations;
the array elements are not in any specific order
when the function returns, the array is returned
just for reference. This parameter can be set to NULL
to avoid copying the output at all.
\return 0 for success, anything else in case of failure
\author Juhani Saastamoinen
*/
int shortfft(
	short n,
	short* input,
	long* output,
	short* xoutput);


#endif /* #ifndef SHORTFFT_H */
