/*! 
\file distfunc.c
\brief Distance function implementations

Implementations of several different
metrics for real vector spaces.
Currently only the Euclidean metric is implemented.\n
$Id: distfunc.c,v 1.2 2003/12/18 10:15:25 juhani Exp $
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <fixmath.h>
#include <errorapi.h>


#include "distfunc.h"


#ifndef max
/*! \brief Resolve maximum of two items */
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#endif


/*======================== Functions ==============================*/


/*!
\brief Euclidean distance of two vectors

Compute the distance of two vectors induced
by the 2-norm, or <i>euclidean norm</i>
\f$ ||\cdot||_2\f$, i.e.
the distance of vectors \f$x, y\in\mathbb{R}^n\f$
is
\f[
\mathrm{d}(x,y) = ||x-y||_2 = \sqrt{\sum_{k=1}^n (x_k-y_k)^2}.
\f]
\param fv1 Vector 1
\param fv2 Vector 2
\param dimension Length of both vectors
\param params Not used in this function, can be any
\return The computed distance
\note This implementation uses integer precision square root,
everything after decimal point is dropped
\author Evgeny Karpov
*/
INT32 dfEuclidean(
	INT32 * fv1,
	INT32 * fv2,
    int dimension,
	void * params) 
{
    INT32 tmp, sum ;
    UINT32 frac;
    int i ;

    if ((NULL == fv1) || (NULL == fv2) || (dimension <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1;
    }

    sum = 0 ;

    for (i = 0 ;i < dimension; i++) 
    {
        tmp  = fv1[i] - fv2[i] ;
        sum += tmp * tmp ; 
    }

    tmp = frac_sqrt(sum, &frac, 8) ;
    if (tmp < 0)
    {
        ReportError(ESR_SQRT, NULL);
        return -1;
    }

    return tmp;
}


/*!
\brief Manhattan distance of two vectors

Compute the distance of two vectors induced by
the 1-norm, or <i>sum norm</i> \f$||\cdot||_1\f$, i.e. 
the distance of vectors \f$x,y\in\mathbb{R}^n\f$
is
\f[
\mathop\mathrm{d}(x,y) = \sum_{k=1}^n |x_k-y_k|.
\f]
\param fv1 Vector 1
\param fv2 Vector 2
\param dimension Length of both vectors
\param params Not used in this function, can be any
\return The computed distance
\author Evgeny Karpov
*/
INT32 dfManhattan(
	INT32 * fv1,
	INT32 * fv2,
    int dimension,
	void* params) 
{
    INT32 d;
    int i ;

    if ((NULL == fv1) || (NULL == fv2) || (dimension <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1;
    }

    for (i = 0, d = 0; i < dimension; i++)
    {
        d += abs(fv1[i] - fv2[i]);
    }

    return d ;
}


/*!
\brief Maximum distance of vector elements

Compute the distance of two vectors induced
by the infinity-norm, or <i>maximum norm</i>
\f$||\cdot||_\infty\f$, i.e. the distance
of vectors \f$x, y\in\mathbb{R}^n\f$ is
\f[
\mathop\mathrm{d}(x,y) = \max_{k=1,\dots,n} |x_k-y_k|.
\f]
\param fv1 Vector 1
\param fv2 Vector 2
\param dimension Length of both vectors
\param params Not used in this function, can be any
\return The computed distance
\author Evgeny Karpov
*/
INT32 dfMaximum(
	INT32 * fv1,
	INT32 * fv2,
    int dimension,
	void * params) 
{
    INT32 d;
    int i ;

    if ((NULL == fv1) || (NULL == fv2) || (dimension <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1;
    }

    for (i = 0, d = 0; i < dimension; i++)
    {
        d += max(d, abs(fv1[i] - fv2[i]));
    }

    return d ;
}


/*======================== Functions ==============================*/
