/*! 
\file fvector.c
\brief Feature vector set implementation

Implementation of all functionality
needed for working with sets of feature
vectors.\n
$Id: fvector.c,v 1.3 2004/01/02 16:08:05 juhani Exp $
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>


#include <errorapi.h>


#include "fvector.h"

/*-----------------------------------------------------------------*/

/*!
\brief The size of vector set initially
when created with the function fvsCreate
\sa fvsCreate
*/
#define FVSET_INITIAL_SIZE      50

/*======================== Functions ==============================*/

/*!
\brief Create a feature vector

Allocate an array of \p size 16-bit integers
\param size The length of the array allocated array
\return Pointer to the allocated array
or NULL in case of failure
\pre size must be positive
\author Evgeny Karpov
*/
INT32 * fvCreate(
	int size) 
{
    INT32 * ret_vector ;

    if (size<=0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ;
    }

    if ((ret_vector = malloc(sizeof(INT32)*size))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL ;
    }

    return ret_vector ;
}


/*!
\brief Deallocate a featuve vector

Free memory allocated for a feature vector
\param fv Pointer to a vector allocated with fvCreate or NULL
\sa fvCreate
\author Evgeny Karpov
*/
void fvDestroy(
	INT32 * fv) 
{
    free(fv) ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Create a vector set

Allocates a vector set structure and
its vector pointer table and returns pointer
to the new structure.
\param size Maximum size of the vector set or
non-positive number to use the default set size
\param vector_size Vector size in this set
\return Pointer to the new vector set
or NULL in case of failure
\pre vector_size must be positive
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET fvsCreate(
	int size,
	int vector_size) 
{
    FVSet * vector_set ;

    if ((vector_set = (FVSet*) malloc(sizeof(FVSet)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL ; /* not enough memory */
    }

    if (vector_size <= 0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ; /* invalid vector size */
    } 

    vector_set->vector_number = 0 ;
    vector_set->vector_size = vector_size ;

    /* Initial set size */
    if (size<=0)
    {
        size = FVSET_INITIAL_SIZE ;
    }

    vector_set->array_size = size ; 

    vector_set->vectors =
      (INT32**) malloc(sizeof(INT32*)*vector_set->array_size);
    if (NULL == vector_set->vectors)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return NULL ; /* Not enough memory */
    }

    memset(vector_set->vectors,0,sizeof(FVSet*)*vector_set->array_size) ;

    return (HFSET) vector_set ;
}


/*!
\brief Deallocate a featuve vector set

Free memory allocated for a feature vector
set structure and its vectors.
\param fvs Pointer to a feature vector set
allocated by fvsCreate or NULL
\sa HFSET (a pointer to FVSet)
\sa fvsCreate
\author Evgeny Karpov
*/
void fvsDestroy(
	HFSET fvs) 
{
    FVSet * vector_set ;
    int i ;

    if ((vector_set = (FVSet*)fvs)==NULL)
	{
		return ;
	}

    for (i=0;i<vector_set->vector_number;i++)
	{
        fvDestroy(vector_set->vectors[i]) ;
	}

    free(vector_set->vectors) ;
    free(vector_set) ;

}


/*-----------------------------------------------------------------*/


/*!
\brief Add vector to end of a feature vector set

Add a vector to end of a feature vector set
as pointer, i.e. without copying the vector data.
\param fvs Feature vector set where new vector is added
\param fv Pointer to the new vector data
\return Number of the added vector in
the vector set or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
the added vector fv must not be NULL and it must
have the same length as the vectors in the vector set fvs;
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsAddVector(
	HFSET fvs,
	INT32 * fv) 
{
    FVSet * vector_set ;
    INT32 ** new_buffer ;

    vector_set = (FVSet*)fvs;
    if ((NULL == vector_set) || (NULL == fv))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ; 
    }

    if (vector_set->vector_number+1 > vector_set->array_size)
    {
        /* if there is not enough empty space in array then double the size */
        new_buffer = realloc(
                       vector_set->vectors,
                       sizeof(INT32*) * 2 * vector_set->array_size);
        if (NULL == new_buffer)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            return ESR_FAILED ; 
        }

        memset(new_buffer+vector_set->vector_number,0,sizeof(INT32*)*vector_set->array_size) ;
        vector_set->vectors = new_buffer ; 
        vector_set->array_size *= 2 ;

    }

    vector_set->vectors[vector_set->vector_number] = fv ;
    vector_set->vector_number++ ;

    return vector_set->vector_number ;
}

/*!
\brief Copy vector to end of a feature vector set

Add a copy of given vector data
to the end of a feature vector set. This function
allocates the new memory needed for the copy
using fvCreate and then uses fvsAddVector
to add the new vector in the set.
\param fvs Feature vector set where new vector is added
\param fv Pointer to the new vector data
\return Number of the added vector
in the vector set or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
the added vector fv must not be NULL and it must
have the same length as the vectors in the vector set fvs;
\sa HFSET (a pointer to FVSet)
\sa fvCreate, fvsAddVector
\author Evgeny Karpov
*/
int fvsAddVectorCopy(
    HFSET fvs,
    INT32 * fv) 
{
    INT32 * ptmp ;
    int ret_val ;

    ptmp = fvCreate(((FVSet*)fvs)->vector_size);
    if (NULL == ptmp)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    memcpy(ptmp,fv,sizeof(INT32)*((FVSet*)fvs)->vector_size) ;

	ret_val = fvsAddVector(fvs,ptmp);
    if (ESR_FAILED == ret_val)
	{
        ReportError(ESR_VECTOR_SET_ADD, NULL);
		free(ptmp) ;
	}

    return ret_val ;

}

/*!
\brief Remove a vector from vector set

Removes a vector with a given number from
a vector set and deallocates it using \p fvDestroy
\param fvs Vector set
\param num Number of vector that must be removed
\return Number of last vector in the set after the removal
or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
num must be within the limits set by the size of the vector set:
num must be at least one (first vector) and no more
than fvsGetSize(fvs) would return.
\sa HFSET (a pointer to FVSet)
\sa fvDestroy
\author Evgeny Karpov
*/
int fvsRemoveVector(
	HFSET fvs,
	int num) 
{
    FVSet * vector_set ;

    if ((vector_set = (FVSet*)fvs)==NULL ||
        vector_set->vector_number<num)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ; 
    }

    fvDestroy(vector_set->vectors[num-1]) ;

    memcpy(
      vector_set->vectors+num-1,
      vector_set->vectors+num,
      sizeof(INT32*)*(vector_set->vector_number-num)) ;

    vector_set->vector_number-- ;
    vector_set->vectors[vector_set->vector_number] = NULL ;

    return vector_set->vector_number ;
}

/*!
\brief Insert vector in any position in a vector set

Insert a given vector to any position
in a vector set. Existing vectors
in given position and after that will
be pushed one index further.
\param fvs The modified vector set
\param fv Vector that will be inserted
\param num Number of inserted vector in the vector set
\return Number of last vector after insertion
or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
the inserted vector fv must not be NULL and it must
have the same length as the vectors in the vector set fvs;
num must be within the limits set by the size of the vector set:
num must be at least one (first vector) and no more
than fvsGetSize(fvs) would return.
\bug There is code missing in the end of the function:
the old elements are push further but the input
vector is not inserted.
\bug The return value is the number of
last vector but this is not modified
even though new vector is inserted
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsInsertVector(
	HFSET fvs,
	INT32 * fv,
	int num) 
{
    FVSet * vector_set ;
    INT32 ** new_buffer ;

    if ((vector_set = (FVSet*)fvs)==NULL ||
        vector_set->vector_number<num || fv==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ; 
    }

    if (vector_set->vector_number+1>vector_set->array_size)
    {
        new_buffer = realloc(
                       vector_set->vectors,
                       sizeof(INT32*)*2*vector_set->array_size);
        if (NULL == new_buffer)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            return ESR_FAILED ; 
        }

        memset(
          new_buffer + vector_set->vector_number,
          0,
          sizeof(INT32*)*vector_set->array_size) ;
        vector_set->vectors = new_buffer ; 
        vector_set->array_size *= 2 ;

    }

    memmove(
      vector_set->vectors+num,
      vector_set->vectors+num-1,
      sizeof(INT32*)*(vector_set->vector_number-num+1)) ;

    return vector_set->vector_number ;
}

/*!
\brief Replace a vector in a feature vector set

Destroy an old vector with given number
(using fvDestroy) and replace that
pointer with the given new pointer.
\param fvs Feature vector set operated upon
\param fv Pointer of the new vector data
\param num Number of the vector that should be replaced
\return Number of the last vector in the set
or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
the replacement vector fv must not be NULL and it must
have the same length as the vectors in the vector set fvs;
num must be within the limits set by the size of the vector set:
num must be at least one (first vector) and no more
than fvsGetSize(fvs) would return.
\sa HFSET (a pointer to FVSet)
\sa fvDestroy
\author Evgeny Karpov
*/
int fvsReplaceVector(
	HFSET fvs,
	INT32 * fv,
	int num) 
{
    FVSet * vector_set ;

    if ((vector_set = (FVSet*)fvs)==NULL ||
        vector_set->vector_number<num || fv==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ; 
    }

    fvDestroy(vector_set->vectors[num-1]) ;
    vector_set->vectors[num-1] = fv ;

    return vector_set->vector_number ;
}

/*!
\brief Get number of vectors in a vector set

\param fvs Vector set whose size is requested
\return The number of vectors in the vector set
or ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsGetSize(
	HFSET fvs) 
{
    if (fvs==NULL) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return ((FVSet*)fvs)->vector_number ;
}

/*!
\brief Get size of vectors in a given feature vector set

Returns the vector size assigned to a vector
set when the vector set is created.
\param fvs Vector set
\return Size of vectors in the vector set or
ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL
\note The algorithm cannot check validity of a non-NULL pointer,
i.e. the vector may or may not point to a properly constructed
vector set.
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsGetVectorSize(
	HFSET fvs) 
{
    if (fvs==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    return ((FVSet*)fvs)->vector_size ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Get vector data from a vector set

Returns pointer to the data of a given
vector in a vector set
\param fvs Vector set
\param num Number of the vector whose data pointer
is requested
\return Pointer to the requested vector's data
or NULL in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
num must be within the limits set by the size of the vector set:
num must be at least one (first vector) and no more
than fvsGetSize(fvs) would return.
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
INT32 * fvsGetVector(
	HFSET fvs,
	int num) 
{
    FVSet * vector_set ;

    if ((vector_set = (FVSet*)fvs)==NULL ||
        vector_set->vector_number<num)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ; 
    }

    return vector_set->vectors[num-1] ;
}

/*!
\brief Copy a vector from vector set

Allocate new vector using fvCreate and copy given vector
from a vector set to the new vector.
\param fvs Vector set
\param num Number of vector whose copy is requested
\return Pointer to the new vector or NULL in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
num must be within the limits set by the size of the vector set:
num must be at least one (first vector) and no more
than fvsGetSize(fvs) would return.
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
INT32 * fvsCopyVector(
	HFSET fvs,
	int num) 
{
    FVSet * vector_set ;
    INT32 * copy_vector ;

    if (((vector_set = (FVSet*)fvs)==NULL)
          || (num <= 0) || (vector_set->vector_number < num))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ; 
    }

    copy_vector = fvCreate(vector_set->vector_size);
    if (NULL == copy_vector)
	{
        ReportError(ESR_VECTOR_CREATE, NULL);
		return NULL ;
	}

    memcpy(copy_vector,
           vector_set->vectors[num-1],
           sizeof(INT32)*(vector_set->vector_size)) ;

    return copy_vector ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Load vector set from a named file

Read a vector set from a file stored by
fvsSave and reconstruct the vector set that
was stored in the file
\param file_name file name where a vector set was stored
\return Pointer to the loaded vector set
or NULL in case of failure
\pre file_name must be a proper file name and the
file should contain a vector set previously stored with fvsSave
\todo Too long a function and too complicated code.
\sa HFSET (a pointer to FVSet)
\sa fvsSave
\author Evgeny Karpov
*/
HFSET fvsLoad(
	char * file_name) 
{
    FVSet * vector_set ;
    FILE * filep ;

    INT32 * tmp_vector ;
    char * str_buff ;
    int str_buff_size, vector_size ;

    int char_count,vector_pos,stop ;
    char ch ;

    if (file_name==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ; 
    }

    if ((filep = fopen(file_name,"r"))==NULL)
    {
        ReportError(ESR_CANT_OPEN_FILE, file_name);
        return NULL ; 
    }

    /* 
        For reading vectors from file temporary array str_buff 
        (array of char) is used for storing one line from file. 
        If size str_buff is not enough its size is doubled.
        Vector size is assumed to be the size of first vector
        stored in set.
    */

    /* create set with default size and dummy vector size */
    if ((vector_set = fvsCreate(0,1))==NULL) 
    {
        ReportError(ESR_VECTOR_SET_CREATE, NULL);
        fclose(filep) ;
        return NULL ;
    }

    /* set vector size to zero */
    vector_set->vector_size = 0 ;

    vector_size = 1 ; /* initial vector size */
    tmp_vector = (INT32*)malloc(sizeof(INT32)*vector_size) ;

    str_buff_size = 50 ; /* initial string size */
    str_buff = (char*) malloc(sizeof(char)*str_buff_size) ;

    if (tmp_vector==NULL || str_buff==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        fvsDestroy(vector_set) ;
        fclose(filep) ;
        free(tmp_vector) ;
        free(str_buff) ;
        return NULL ; 
    }

    char_count = 0 ; vector_pos = 0 ; stop = 0 ;

    while (stop==0)
    {
        ch=getc(filep) ;

        if (isdigit((int)ch) || ch=='-' || ch=='+' || ch=='.')
        {
            if (char_count>=str_buff_size) /* temporary buffer is not enough */
            {
                str_buff_size*=2 ;
                if ((str_buff = (char*)realloc(str_buff,sizeof(char)*str_buff_size))==NULL)
                {
                    ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
                    fvsDestroy(vector_set) ;
                    fclose(filep) ;
                    free(tmp_vector) ;
                    free(str_buff) ;
                    return NULL ; 
                }
            }

            str_buff[char_count] = ch ;
            char_count++ ;
        }

        if (!isdigit((int)ch) && ch!='-' && ch!='+' && ch!='.')
        {
            if (char_count>0)
            {
                if (vector_pos>=vector_size) /* temporary vector is not big enough */
                {
                    if (vector_set->vector_size == 0)
                    {
                        /* first vector => estimate size */
                        vector_size++ ;
                        tmp_vector = (INT32*) realloc(
                                                tmp_vector,
                                                sizeof(INT32)*vector_size);
                        if (NULL == tmp_vector)
                        {
                            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
                            fvsDestroy(vector_set) ;
                            fclose(filep) ;
                            free(tmp_vector) ;
                            free(str_buff) ;
                            return NULL ; 
                        }

                    } else {
                        /* vector size != 0; invalid parameter? */
                        ReportError(ESR_INVALID_PARAMETER, NULL);
                        fvsDestroy(vector_set) ;
                        fclose(filep) ;
                        free(tmp_vector) ;
                        free(str_buff) ;
                        return NULL ; 
                    }
                }

                str_buff[char_count] = 0 ;
                tmp_vector[vector_pos] = atoi(str_buff) ;
                vector_pos++ ;

                char_count=0 ;
            }
        }

        if (ch=='\r' || ch=='\n' || ch==EOF)
        {
            /* Remove all '\r' and '\n' */
            while ((ch=='\r' || ch=='\n') && ch!=EOF)
			{
				ch=getc(filep) ;
			}
            if (ch!=EOF)
			{
				ungetc(ch,filep) ; /* Return back one character */
			}

            /* if it is the first vector set vector size */
            if (vector_set->vector_size == 0)
            {
                vector_set->vector_size = vector_size ;
            }

            /* add vector to set */
            if (fvsAddVector(vector_set,tmp_vector)==ESR_FAILED)
            {
                ReportError(ESR_VECTOR_SET_ADD, NULL);
                fvsDestroy(vector_set) ;
                fclose(filep) ;
                free(tmp_vector) ;
                free(str_buff) ;
                return NULL ; 
            }

            /* create new temporary vector */
            tmp_vector = (INT32*) malloc(sizeof(INT32)*vector_size);
            if (NULL == tmp_vector)
            {
                ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
                fvsDestroy(vector_set) ;
                fclose(filep) ;
                free(tmp_vector) ;
                free(str_buff) ;
                return NULL ; 
            }

            vector_pos=0 ;

            if (ch==EOF)
			{
				stop = 1 ;
			}
        }

    }

    free(str_buff) ;
    free(tmp_vector) ;
    fclose(filep) ;

    return (FVSet*) vector_set ;
}

/*!
\brief Store a vector set in a file

\param fvs Vector set
\param file_name Name of the file where vector set is stored
\return ESR_SUCCEED indicating success or
ESR_FAILED in case of failure
\pre fvs must be a previously allocated vector set, not NULL;
file_name must be a proper file name
\sa HFSET (a pointer to FVSet)
\sa fvsLoad
\author Evgeny Karpov
*/
int fvsSave(
	HFSET fvs,
	char * file_name) 
{
    FVSet * vector_set ;
    int vector_count,pos ;
    INT32 * fv ;
    FILE * filep ;

    if (fvs==NULL || file_name==NULL) 
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ; 
    }

    vector_set = (FVSet*) fvs ;

    filep=fopen(file_name,"w");
    if (NULL == filep)
    {
        ReportError(ESR_CANT_OPEN_FILE, file_name);
        return ESR_FAILED;
    }

    for (vector_count = 0;
         vector_count < vector_set->vector_number;
         vector_count++)
    {
        if ((fv=vector_set->vectors[vector_count])!=NULL)
        {
            for (pos=0;pos<vector_set->vector_size;pos++)
                fprintf(filep,"%ld ",fv[pos]) ;
            fprintf(filep,"\n") ;
        }
    }

    fclose(filep) ;
    return ESR_SUCCEED ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Merge two vector sets

Append vectors of the set 2 in set 1.
If either of the sets is NULL, the other
one is the merged set. If either of them
is empty, it will be deallocated with fvsDestroy.
\param fvs1 Vector set 1
\param fvs2 Vector set 2
\return Pointer to the merged vector set
or NULL in case of failure
\pre At least one of fvs1 and fvs2 must be
previously allocated vector sets, not NULL
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET fvsMerge(
	HFSET fvs1,
	HFSET fvs2) 
{
    FVSet * master_set, * slave_set ;
    INT32 ** tmp_buff ;

    if (fvs1==NULL && fvs2==NULL)
	{
        ReportError(ESR_INVALID_PARAMETER, NULL);
		return NULL ;
	}

    if (fvs1==NULL)
	{
		return fvs2 ;
	}
    if (fvs2==NULL)
	{
		return fvs1 ;
	}

    if (fvsGetSize(fvs2)==0)
    {
        fvsDestroy(fvs2) ;
        return fvs1 ;
    }

    if (fvsGetSize(fvs1)==0)
    {
        fvsDestroy(fvs1) ;
        return fvs2 ;
    }

    master_set = (FVSet*)fvs1 ;
    slave_set = (FVSet*)fvs2 ;        

    if (master_set->array_size
		  < master_set->vector_number + slave_set->vector_number)
    {
        while (master_set->array_size
			     < master_set->vector_number + slave_set->vector_number) 
		{
            master_set->array_size*= 2 ;
		}

		tmp_buff = realloc(master_set->vectors,
			               sizeof(INT32*)*master_set->array_size);
        if (NULL == tmp_buff)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
            return NULL ; 
        }

        master_set->vectors = tmp_buff ;
        memset(master_set->vectors + master_set->vector_number,
			     0,
                 sizeof(INT32*)*(master_set->array_size
				                    - master_set->vector_number)) ;

    }

    memcpy(master_set->vectors + master_set->vector_number,
		   slave_set->vectors,
           sizeof(INT32*)*slave_set->vector_number) ;

    master_set->vector_number += slave_set->vector_number ;

    free(slave_set->vectors) ;
    free(slave_set) ;

    return fvs1 ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Change size of a vector set

Change the number of vectors in a vector set.
\param fvs Vector set
\param new_size New size for the vecotr set
\return 1 for success, 0 for failure, in which case
the variable errno will be set
\pre fvs must be a previously allocated vector set, not NULL;
new_size must be positive
\bug Should we check that we do not increase the
size of set, so that there appears empty
vectors?
\bug Memory leak: should we remove non-empty
vectors if the set size is decreased?
Do the other functions set the vector pointers
without checking old pointer value?
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsSetSize(
	HFSET fvs,
	int new_size) 
{
    FVSet * vector_set ;

    if (((vector_set = (FVSet*)fvs)==NULL)
          || (new_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return 0 ; 
    }

    vector_set->vector_number = new_size ;

    return 1 ;
}


/*!
\brief Change size of vectors in a vector set

\param fvs Vector set
\param new_size New vector size in the set
\return 1 for success, o for failure, in which
case the errno will be set
\pre fvs must be a previously allocated vector set, not NULL;
new_size must be positive
\todo Explain where is this used (where
and why is such resizing needed)
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
int fvsSetVectorSize(
	HFSET fvs,
	int new_size) 
{
    FVSet * vector_set ;

    if (((vector_set = (FVSet*)fvs)==NULL)
          || (new_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return 0 ; 
    }

    vector_set->vector_size = new_size ;

    return 1 ;
}


/*!
\brief Get pointer to the vector table

\param fvs Vector set
\return Pointer to the vector table or
NULL in case of failure
\pre fvs must be a previously allocated vector set, not NULL
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
void * fvsGetVectorBuff(
	HFSET fvs) 
{
    FVSet * vector_set ;

	vector_set = (FVSet*) fvs;
    if (NULL == vector_set)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL ; 
    }

    return vector_set->vectors ;
}


/* print a vector of a vector set */
int fvsPrintVector(FILE* f, HFSET fvs, int num)
{
    FVSet * vector_set ;
	int i;
	INT32* vector;

	vector_set = (FVSet*) fvs;
    if (NULL == vector_set)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED; 
    }

	vector = fvsGetVector(vector_set, num);
	if (vector == NULL)
	{
		ReportError(ESR_VECTOR_SET_GET, NULL);
		return ESR_FAILED;
	}

	fprintf(f, "vector %d:", num);
	for (i = 0; i < fvsGetVectorSize(vector_set); i++)
	{
		fprintf(f, " %ld", vector[i]);
	}
	fprintf(f, "\n");

	return ESR_SUCCEED;
}


/*======================== Functions ==============================*/

