/*! 
\file 
\brief Metric space indexing

Common framework to work with indexes. Can be used to create 
indexes and search trough the data. Supposed to work with 
vector sets only. Should be used instead of direct usage of 
indexing algorithms.

\date 05 November 2003
\author Evgeny Karpov
*/
/*
    $Id: msindex.c,v 1.1 2003/11/30 20:39:30 juhani Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errorapi.h>

/* indexing algorithms */
#include "vptree.h"

#include "msindex.h"

/*-----------------------------------------------------------------*/

/*!
\brief Create index function type

This function type is used to hold pointer to indexing algorithm 
specific function for index creation

\param index_set Indexed vector set
\param dist_func Distance function
\param dfunc_params Parameters for distance function
\return Created indexes
\author Evgeny Karpov
*/
typedef void * (*CREATEIDX)  (HFSET index_set, 
                              DISTFUNC dist_func, void * dfunc_params) ;

/*!
\brief Search index function type

This function type is used to hold pointer to indexing algorithm 
specific function for index creation

\param indexes Indexes used for search
\param search_set Set what was indexed
\param element Element for which nearest vectors are searched
\param radius Searching radius
\param ret_arr Returned array filled by numbers of nearest
elements in a source set
\param source_params Indexed source params
\return Amount of nearest elements found
\author Evgeny Karpov
*/
typedef int    (*SEARCHIDX)  (void * indexes, HFSET search_set,
                              double * element, double radius, int * ret_arr) ;

/*!
\brief Nearest neighbour search function type

This function type is used to hold pointer to indexing algorithm 
specific function for searching for nearest element using indexes

\param indexes Indexes
\param search_set Search source
\param element Search point (element for what nearest neighbour
is searched)
\param ret_dist Pointer to a variable where to put distance 
to found nearest neighbour element. If NULL then no distance
will be returned
\return Number of nearest element in a source set
\author Evgeny Karpov
*/
typedef int    (*NNSEARCHIDX)(void * indexes, HFSET search_set,
                              double * element, double * ret_dist) ;

/*!
\brief Destroy index funtion type

This function type is used to hold pointer to indexing algorithm 
specific function for destroying indexes

\param indexes Indexes to destroy
\return Nothing
\author Evgeny Karpov
*/
typedef void   (*DESTROYIDX) (void * indexes) ;

/*======================== Structures =============================*/

/*! 
\brief Indexer parameters

Set of parameters and function pointers which represents
indexing algorithm in use
*/
typedef struct {

    MSIndexType type ;      /*!< indexing algorithm */

    /* working functions pointers */
    CREATEIDX   create_idx_func ;    /*!< create index function */
    SEARCHIDX   search_idx_func ;    /*!< search index function */
    NNSEARCHIDX nn_search_idx_func ; /*!< nn-serach index function */
    DESTROYIDX  destroy_idx_func ;   /*!< destroy index function */

} MSIndex ;

/*======================== Structures =============================*/

/*======================== Functions ==============================*/

/*!
\brief Creates indexer

Creates indexer using requested algorithm 

\param type Type of indexing algorithm
\return Created indexer
\sa MSIndexType msidxDestroy msidxTrain
\author Evgeny Karpov
*/
HMSIDX msidxCreate(MSIndexType type) 
{
    MSIndex * pmsidx ;

    /* allocate memory */
    if ((pmsidx = malloc(sizeof(MSIndex)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* initialize to zero */
    memset(pmsidx,0,sizeof(MSIndex)) ;

    pmsidx->type = type ;

    /* set up indexing algorithm */
    if (type == MSINDEX_ALGTYPE_VPT)
    {
        pmsidx->create_idx_func = (CREATEIDX)vptCreate ;
        pmsidx->search_idx_func = (SEARCHIDX)vptSearch ;
        pmsidx->nn_search_idx_func = (NNSEARCHIDX)vptNNSearch ; 
        pmsidx->destroy_idx_func = (DESTROYIDX)vptDestroy ;

    } else {

        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    return (HMSIDX) pmsidx ;
}

/*!
\brief Destroys indexer

Deallocate all memory used for indexer

\param msidx Indexer to destroy
\return Nothing
\note If indexer is "NULL" then no error will be arisen
\sa msidxDestroyIndexes
\author Evgeny Karpov
*/
void msidxDestroy(HMSIDX msidx) 
{
    if (msidx != NULL) free(msidx) ;    
}

/*!
\brief Destroy indexes

Deallocate all memory used for indexes

\param msidx Indexer used to create indexes
\param indexes Indexes to destroy
\return Nothing
\sa msidxDestroy
\author Evgeny Karpov
*/
void msidxDestroyIndexes(HMSIDX msidx,void * indexes) 
{
    if (msidx != NULL) ((MSIndex*)msidx)->destroy_idx_func(indexes) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Creates indexes

Creates indexes for concrete source using indexer

\param msidx Indexer what will be used for creation of indexes
\param test_set Source to index
\param dist_func Distance function
\param dfunc_params Parameters for distance function
\return Created indexes
\note Use indexer and msidxDestroyIndexes to destroy created indexes
\sa msidxDestroyIndexes
\author Evgeny Karpov
*/
HMSINDEX msidxTrain(HMSIDX msidx,HFSET test_set,
                    DISTFUNC dist_func, void * dfunc_params) 
{
    MSIndex * pmsidx ;
    HMSINDEX result;

    if ((pmsidx = (MSIndex*)msidx)==NULL || test_set==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    result = pmsidx->create_idx_func(test_set,
                dist_func, dfunc_params) ;

    if (NULL == result)
    {
        ReportError(ESR_MSINDEX_CREATE_FUNCTION, NULL);
        return NULL ;
    }

    return result;
}

/*-----------------------------------------------------------------*/

/*!
\brief Searches through the source with fixed radius

Searches nearest elements through the source with fixed radius

\param msidx Indexer used to create indexes
\param indexes Indexes
\param search_set Source set to search trough  
\param element Search point (center)
\param radius Search radius
\param ret_arr Array of elements from source what are as close
to search point as radius is (in other words, elements what are in 
a sphere with center in search point and radius equals to search 
radius)
\return Number of nearest elements found
\sa msidxNNSearch
\author Evgeny Karpov
*/
int msidxSearch(HMSIDX msidx, HMSINDEX indexes, HFSET search_set,
                double * element, double radius, int * ret_arr) 
{
    MSIndex * pmsidx ;
    RetStatus result;

    pmsidx = (MSIndex*)msidx;

    if ((NULL == pmsidx) || (indexes==NULL)
          || (NULL == search_set) || (element==NULL)
          || (ret_arr==NULL))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return ESR_FAILED ;
    }

    result = pmsidx->search_idx_func(indexes, search_set,
                    element, radius,ret_arr) ;

    if (ESR_FAILED == result)
    {
        ReportError(ESR_MSINDEX_SEARCH_FUNCTION, NULL);
        return ESR_FAILED ;
    }

    return result;
}

/*!
\brief Nearest neighbour search

Searches for nearest element using indexes

\param msidx Indexer used to create indexes
\param indexes Indexes
\param search_set Search source
\param element Search point (element for what nearest neighbour
is searched)
\param ret_dist Pointer to a variable where to put distance 
to found nearest neighbour element. If NULL then no distance
will be returned
\return Number of nearest element in a source set
\author Evgeny Karpov
*/
int msidxNNSearch(HMSIDX msidx, HMSINDEX indexes, HFSET search_set,
                  double * element, double * ret_dist) 
{
    MSIndex * pmsidx ;
    RetStatus result;

    pmsidx = (MSIndex*)msidx;

    if ((NULL == pmsidx) || (indexes==NULL)
          || (NULL == search_set) || (element==NULL)
          || (NULL == ret_dist))
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        if (ret_dist)
        {
            *ret_dist = -1.0 ;
        }
        return ESR_FAILED ;
    }

    result = pmsidx->nn_search_idx_func(indexes, search_set,
                                        element, ret_dist) ;

    if (ESR_FAILED == result)
    {
        ReportError(ESR_MSINDEX_NN_SEARCH_FUNCTION, NULL);
        return ESR_FAILED ;
    }

    return result;
}

/*-----------------------------------------------------------------*/

/*!
\brief Gets type of indexing

Returns the type of indexing algorithm used in indexer

\param msidx Indexer
\return Type of indexing algorithm
\author Evgeny Karpov
*/
MSIndexType msidxGetType(HMSIDX msidx) 
{
    if (NULL == msidx)
    {
        ReportError(ESR_NOT_IMPLEMENTED, NULL);
        return ESR_FAILED;
    }

    return ((MSIndex*)msidx)->type ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Loads indexer parameters

Loads indexer parameters

\param file_name File name
\return Indexer
\note Not implemented
\todo To implement
\sa msidxSave
\author Evgeny Karpov
*/
HMSINDEX msidxLoad(MSIndexType type, char * file_name)
{
    /* NOT IMPLEMENTED YET */

    ReportError(ESR_NOT_IMPLEMENTED, NULL);

    return NULL ;
}

/*!
\brief Saves indexer parameters

Saves indexer parameters

\param msidx Indexer to save
\param file_name File name
\return Whether saving failed or succeeded
\note Not implemented
\todo To implement
\todo Add RetStatus from errorapi as a return value
\sa msidxLoad
\author Evgeny Karpov
*/
int msidxSave(MSIndexType type, HMSINDEX indexes,char * file_name) 
{
    /* NOT IMPLEMENTED YET */

    ReportError(ESR_NOT_IMPLEMENTED, NULL);

    return 0 ;
}

/*======================== Functions ==============================*/
