/*! 
\file 
\brief SProfiler configuration parser

Parses configuration files to work with SProfiler
\date 10 November 2003
\author Evgeny Karpov
*/

/*
    $Id: sprcfg.h,v 1.2 2003/12/04 13:41:12 villeh Exp $
*/

#ifndef __SPRCFG_H__
#define __SPRCFG_H__

/*-----------------------------------------------------------------*/

#include <errorapi.h>

#include <sremoval.h>
#include <preprocess.h>

#include <fvector.h>

#include <speakerdb.h>
#include <modelset.h>

/*-----------------------------------------------------------------*/

typedef double * (*FEXTRACTOR) (void * handle, double * frame,int size) ;
typedef void     (*DESTROYFUN) (void * handle) ;

typedef void *   (*MODELTRAIN) (void * handle,HFSET training_set,int model_size) ;
typedef double   (*MODELMATCH) (void * handle,void * model,HFSET test_set) ;
typedef void *   (*MODELLOAD)  (void * handle,char * file_name) ;
typedef void     (*MODELDESTR) (void * handle,void * model) ;

typedef int      (*SPRUNING)   (void * handle,HMS models,HFSET test_set,double * ret_score,void * params) ;

/*======================== Structures =============================*/

/* feature extraction parameters */
typedef struct {

    void * alg_params ;
    void * alg_handle ;

    int win_size_ms ;
    int win_shift_ms ;

    FEXTRACTOR extractor_func ;
    DESTROYFUN destroy_func ;

    PreProc * pp ;  /* preprocessing functions */
    HSREM sr ;      /* silence removal */

    int mel_filters_auto ; /* MFCC specific, TODO: to get rid of this */

} FE_params ;

/* speaker modeling parameters */
typedef struct {

    void * alg_handle ; 

    MODELTRAIN train_func ;
    MODELMATCH match_func ;
    MODELDESTR destroy_func ;
    MODELLOAD  load_func ;

    SPRUNING spruning_func ;
    void *   spruning_params ;

    /* HPQ pq *//* pre-quantization handle */

} MODEL_params ;


/* Classifier fusion parameters, temporarily here ... Let's see what is
   the best way to handle this. */

typedef struct {
  int nbest ;
  char **weight_tags ;
  double *weights ;

} FUSION_params ;


/* sprofiler action types */
typedef enum {

    SPR_ACTION_MATCH,
    SPR_ACTION_MODEL,
    SPR_ACTION_VERIFY,
    SPR_ACTION_FEXTR

} SPAction ;

/* sprofiler modeling types */
typedef enum {

    SPR_MODEL_VQ

} SPModelType ;

/* sprofiler features types */
typedef enum {

    SPR_FEAT_LOAD,
    SPR_FEAT_MFCC

} SPFeatureType ;

/* sprofiler logging levels */
typedef enum {

    SPR_LOG_NOTHING     = 0,
    SPR_LOG_LOW         = 1,
    SPR_LOG_MODERATE    = 2,
    SPR_LOG_FULL        = 3

} SPLogLevel ;

/* speaker pruning types */
typedef enum {

    SPR_PRUNING_REORD ,
    SPR_PRUNING_PQBASED ,
    SPR_PRUNING_HIERARCH ,
    SPR_PRUNING_CFBASED ,
    SPR_PRUNING_STATIC ,
    SPR_PRUNING_ADAPTIVE 

} SPuningType ;

/* main type to handle application states and parameters */
typedef struct {

    SPAction      action ;
    SPModelType   model_type ;
    SPFeatureType feat_type ;
    SPuningType   pruning_type ;

    short flags ;

    int * speaker_ids ;
    char ** input_files ;
    char ** output_files ;
    int     input_list_size ;

    char * database_file ;

    HSDB sdb ;
    HMS  models ;
    HFSET test_set ;
    double curr_score ;

    FE_params    * fe_params ;
    MODEL_params * model_params ;

    FUSION_params  * fusion_params ;

    SPLogLevel log_level ;
    FILE * logfile ; /* log file */

} SProfiler ;

/*======================== Structures =============================*/

/* flags */
#define SPR_REMOVE_SILENCE      0x0001
#define SPR_PREPROCESS          0x0002
#define SPR_PREQUANTIZE         0x0004
#define SPR_SPRUNING            0x0008
#define SPR_INPUT_LIST          0x0010

/* state flags */
#define SPR_STATE_FINISHED      0x4000

/*======================== Functions ==============================*/

/* load configuration */
SProfiler * load_config(int argc,char * argv[]) ;
void close_profiler(SProfiler * sprofiler) ;

/*======================== Functions ==============================*/

#endif /* __SPRCFG_H__ */
