/*! 
\file 
\brief SProfiler configuration parser

Parses configuration files to work with SProfiler
\date 10 November 2003
\author Evgeny Karpov
*/

/*
    $Id: sprcfg.c,v 1.13 2003/12/18 15:32:53 ekarpov Exp $
*/

/*-----------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <math.h> /* needed to compute number of mel-filters */

#include <mfcc.h>
#include <smetric.h>
#include <vq.h>
#include <spruning.h>
#include <errorapi.h>

#include "sprcfg.h"
#include "cfgfile.h"

/*-----------------------------------------------------------------*/

#define PROFILER_MAX_ERROR_LEVEL            20

/*-----------------------------------------------------------------*/

char * copy_text_param(HCFGFILE cfgfile) ;
char * copy_text(char * copystr) ;

/*-----------------------------------------------------------------*/

/* functions to load parts of configuration file */
int cfload_general(SProfiler * profiler,HCFGFILE cfgfile, 
                   int argc,char * argv[]) ;
int cfload_logging(SProfiler * profiler,HCFGFILE cfgfile,
                   int argc,char * argv[]) ;
int cfload_modeling(SProfiler * profiler,HCFGFILE cfgfile,
                    int argc,char * argv[]) ;
int cfload_fextr(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) ;
int cfload_mfcc(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) ;
int cfload_matchproc(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) ;

/*======================== Functions ==============================*/

SProfiler * load_config(int argc,char * argv[]) 
{
    HCFGFILE cfgfile ;
    SProfiler * profiler ;

/*
    char file_name[FILENAME_MAX] ;
*/

    if (argc > 1)
    {
        /* try to open configuration file */
        if ((cfgfile = cfOpen(argv[1],CFILE_READ))==NULL)
        {
            printf("Error: Can't open configuration file\n") ;
            return NULL ;
        }

    } else {

        /* try to open default configuration file */
        if ((cfgfile = cfOpen("sprofiler.cfg",CFILE_READ))==NULL)
        {
            printf("Error: Can't open configuration file\n") ;
            return NULL ;
        }

        printf("warning: default configuration file will be used\n") ;

    }

    /* allocate memory */
    if ((profiler = (SProfiler*) malloc(sizeof(SProfiler)))==NULL)
    {
        fprintf(stderr,"Fatal error: Not enough memory to continue\n") ;
        cfClose(cfgfile) ;
        return NULL ;
    }

    /* set to zero all parameters */
    memset(profiler,0,sizeof(SProfiler)) ;

    /* read logging parameters */
    if (cfload_logging(profiler,cfgfile,argc,argv) == ESR_FAILED)
    {
        cfClose(cfgfile) ;
        close_profiler(profiler) ;
        return NULL ;
    }

    /* read general parameters */
    if (cfload_general(profiler,cfgfile,argc,argv) == ESR_FAILED)
    {
        cfClose(cfgfile) ;
        close_profiler(profiler) ;
        return NULL ;
    }

    /* read feature extraction parameters */
    if (cfload_fextr(profiler,cfgfile,argc,argv) == ESR_FAILED)
    {
        cfClose(cfgfile) ;
        close_profiler(profiler) ;
        return NULL ;
    }

    /* read modeling parameters */
    if (cfload_modeling(profiler,cfgfile,argc,argv) == ESR_FAILED)
    {
        cfClose(cfgfile) ;
        close_profiler(profiler) ;
        return NULL ;
    }

    /* read matching preprocessing parameters */
    if (cfload_matchproc(profiler,cfgfile,argc,argv) == ESR_FAILED)
    {
        cfClose(cfgfile) ;
        close_profiler(profiler) ;
        return NULL ;
    }

    /* close configuration file */
    cfClose(cfgfile) ;

    return profiler ;
}

void close_profiler(SProfiler * profiler) 
{
    int count ;

    if (profiler)
    {
        free(profiler->database_file) ;

        if (profiler->fe_params)
        {
            if (profiler->fe_params->destroy_func && profiler->fe_params->alg_handle)
                profiler->fe_params->destroy_func(profiler->fe_params->alg_handle) ;

            ppDestroy(profiler->fe_params->pp) ;

            free(profiler->fe_params->alg_params) ;
            free(profiler->fe_params) ;
        }

        if (profiler->model_params)
        {
            vqDestroy(profiler->model_params->alg_handle) ; /* FIX THIS TO USE POINTER !!! */

            free(profiler->model_params->spruning_params) ;
            free(profiler->model_params) ;
        }

        if (profiler->logfile && profiler->logfile != stdout && profiler->logfile != stderr) 
            fclose(profiler->logfile) ;

        if (profiler->input_list_size >= 1 && profiler->input_files)
        {
            for (count=0;count<profiler->input_list_size;count++)
            {
                free(profiler->input_files[count]) ;
                if (profiler->output_files) free(profiler->output_files[count]) ;
            }
        }

        free(profiler->speaker_ids) ;
        free(profiler->input_files) ;
        free(profiler->output_files) ;

        sdbClose(profiler->sdb) ;
        msDestroy(profiler->models) ;
/*
        sdbClose(profiler->task_list) ;
*/
        fvsDestroy(profiler->test_set) ;

        free(profiler) ;
    }
}

char * copy_text_param(HCFGFILE cfgfile) 
{
    int tmp_int ;
    char * ret_str , * tmp_str;

    if ((tmp_str = cfGetTextValue(cfgfile)) && (tmp_int = strlen(tmp_str)))
    {
        if ((ret_str = malloc(sizeof(char)*(tmp_int+1)))!=NULL)
            return strcpy(ret_str,tmp_str) ;

    }

    return NULL ;
}

char * copy_text(char * copystr) 
{
    int tmp_int ;
    char * ret_str ;

    if (copystr!=NULL && (tmp_int = strlen(copystr)))
    {
        if ((ret_str = malloc(sizeof(char)*(tmp_int+1)))!=NULL)
            return strcpy(ret_str,copystr) ;

    }

    return NULL ;
}

/* read general parameters */
int cfload_general(SProfiler * profiler,HCFGFILE cfgfile, 
                   int argc,char * argv[]) 
{
    char * tmp_str ;
    HSDB sdb ;
    int count ;

    if (cfSelectSection(cfgfile,"General"))
    {
        /* what action to do */
        if (cfSelectParam(cfgfile,"action"))
        {
            tmp_str = cfGetTextValue(cfgfile) ;

            if (strcmp(tmp_str,"FE") == 0)
                profiler->action = SPR_ACTION_FEXTR ;
            else if (strcmp(tmp_str,"MODEL") == 0)
                profiler->action = SPR_ACTION_MODEL ;
            else if (strcmp(tmp_str,"MATCH") == 0)
                profiler->action = SPR_ACTION_MATCH ;
            else if (strcmp(tmp_str,"VERIFY") == 0)
                profiler->action = SPR_ACTION_VERIFY ;
            else {

                fprintf(profiler->logfile,"Config error (section [General]): unknown action\n") ;
                return ESR_FAILED ;
            }
        
        } else {

            fprintf(profiler->logfile,"Config error (section [General]): action is not set\n") ;
            return ESR_FAILED ;
        }

        /* input type */
        if (cfSelectParam(cfgfile,"input type") && 
            strcmp(cfGetTextValue(cfgfile),"LIST") == 0)
        {
            profiler->flags |= SPR_INPUT_LIST ;
        }

        if (cfSelectParam(cfgfile,"input file") && 
            strcmp(cfGetTextValue(cfgfile),"STDIN") != 0)
        {
            /* load input list */
            if (profiler->flags & SPR_INPUT_LIST)
            {
                sdb = sdbOpen(cfGetTextValue(cfgfile),SDB_OPEN_READ_ONLY) ;

                profiler->input_list_size = sdbSize(sdb) ;

                profiler->input_files = malloc(sizeof(char*)*profiler->input_list_size) ;
                profiler->output_files = malloc(sizeof(char*)*profiler->input_list_size) ;
                profiler->speaker_ids = malloc(sizeof(int)*profiler->input_list_size) ;

                if (profiler->input_files == NULL || profiler->output_files == NULL)
                {
                    free(profiler->input_files) ;
                    profiler->input_files = NULL ;
                    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                    return ESR_FAILED ;
                }

                /* load file names */
                for (count=0;count < profiler->input_list_size; count++)
                {
                    profiler->speaker_ids[count] = sdbGetIntValue(sdb,1) ;
                    profiler->input_files[count] = copy_text(sdbGetTextValue(sdb,2)) ;
                    profiler->output_files[count] = copy_text(sdbGetTextValue(sdb,3)) ;

                    sdbSelectNext(sdb) ;
                }

                sdbClose(sdb) ;

            } else {

                /* allocate only one input file */
                profiler->input_files = malloc(sizeof(char*)) ;
                profiler->input_list_size = 1 ;

                if (profiler->input_files ==NULL)
                {
                    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                    return ESR_FAILED ;
                }

                profiler->input_files[0] = copy_text_param(cfgfile) ;
            }

        } else {

            /* load input list */
            if (profiler->flags & SPR_INPUT_LIST)
            {
                /* LOAD LIST FROM STDIN */

            } else {

                /* MARK STDIN AS INPUT */
            }
        }

        if (cfSelectParam(cfgfile,"output file") &&
            strcmp(cfGetTextValue(cfgfile),"STDOUT") != 0)
        {
            if (!(profiler->flags & SPR_INPUT_LIST))
            {
                /* allocate only one output file */
                profiler->output_files = malloc(sizeof(char*)) ;
                profiler->output_files[0] = copy_text_param(cfgfile) ;

                if (profiler->output_files ==NULL)
                {
                    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                    return ESR_FAILED ;
                }
            }

        } else {

            /* MARK STDOUT AS OUTPUT */
        }

        /* speaker database */
        if (cfSelectParam(cfgfile,"database"))
            profiler->database_file = copy_text_param(cfgfile) ;

    } else {

        fprintf(profiler->logfile,"Config error: section [General] is missed\n") ;
        return ESR_FAILED ;
    }

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

/* read logging parameters */
int cfload_logging(SProfiler * profiler,HCFGFILE cfgfile,
                   int argc,char * argv[]) 
{
    char file_name[FILENAME_MAX] ;

    char * tmp_str ;

    profiler->logfile = stdout ;
    profiler->log_level = SPR_LOG_FULL ;

    if (cfSelectSection(cfgfile,"Logging"))
    {
        /* check whether logging is needed */
        if (cfSelectParam(cfgfile,"log events") && 
            cfGetBoolValue(cfgfile))
        {
            /* read logging type */
            if (cfSelectParam(cfgfile,"type"))
            {
                tmp_str = cfGetTextValue(cfgfile) ;

                if (strcmp(tmp_str,"FILE") == 0)
                {
                    profiler->logfile = NULL ;

                    /* read log file name */
                    if (cfSelectParam(cfgfile,"file name"))
                        strcpy(file_name,cfGetTextValue(cfgfile)) ;
                    else 
                        sprintf(file_name,"%s.log",argv[0]) ;

                    /* read logging mode */
                    if (cfSelectParam(cfgfile,"file mode"))
                    {
                        tmp_str = cfGetTextValue(cfgfile) ;

                        /* open log file */

                        if (strcmp(tmp_str,"NEW") == 0)
                            profiler->logfile = fopen(file_name,"w") ;
                        else if (strcmp(tmp_str,"APPEND") == 0)
                            profiler->logfile = fopen(file_name,"a") ;
                        else 
                            profiler->logfile = fopen(file_name,"w") ;

                    } else profiler->logfile = fopen(file_name,"w") ;

                    if (profiler->logfile == NULL)
                    {
                        printf("Config warning (section [Logging]): can't open log file, standard output will be used\n") ;
                        profiler->logfile = stdout ;
                    }

                } else if (strcmp(tmp_str,"STDOUT") == 0)
                    profiler->logfile = stdout ;
                else if (strcmp(tmp_str,"STDERR") == 0)
                    profiler->logfile = stderr ;
                else 
                    printf("Config warning (section [Logging]): unknown type, standard output will be used\n") ;

            } else {
                printf("Config warning (section [Logging]): type is missed, standard output will be used\n") ;
            }

            /* read logging type */
            if (cfSelectParam(cfgfile,"log level"))
            {
                tmp_str = cfGetTextValue(cfgfile) ;
                
                if (strcmp(tmp_str,"FULL") == 0)
                    profiler->log_level = SPR_LOG_FULL ;
                else if (strcmp(tmp_str,"MODERATE") == 0)
                    profiler->log_level = SPR_LOG_MODERATE ;
                else if (strcmp(tmp_str,"LOW") == 0)
                    profiler->log_level = SPR_LOG_LOW ;
                else
                    printf("Config warning (section [Logging]): unknown logging level, full level will be used\n") ;
            }

        } else {

            profiler->logfile = NULL ; /* no logging will be used */
            profiler->log_level = SPR_LOG_NOTHING ;
        }

    } else {

        printf("Config warning: section [Logging] is missed, standard output will be used\n") ;

    }

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

/* matching and modeling parameters */
int cfload_modeling(SProfiler * profiler,HCFGFILE cfgfile,
                    int argc,char * argv[]) 
{
    VQParams *   vq_params ;
    char * tmp_str ;

    if (profiler->action != SPR_ACTION_FEXTR)
    {
        if ((profiler->model_params = malloc(sizeof(MODEL_params)))==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
            return ESR_FAILED ;
        }

        memset(profiler->model_params,0,sizeof(MODEL_params)) ;

        /* read what model type is needed */
        if (cfSelectSection(cfgfile,"Modeling"))
        {
            if (cfSelectParam(cfgfile,"model type"))
            {
                tmp_str = cfGetTextValue(cfgfile) ;

                if (strcmp(tmp_str,"VQ") == 0)
                    profiler->model_type =  SPR_MODEL_VQ ;
                else {

                    fprintf(profiler->logfile,"Config warning (section [Modeling]): unknown modeling type, VQ will be used\n") ;
                    profiler->model_type =  SPR_MODEL_VQ ;
                }
            
            }

        } else {

            fprintf(profiler->logfile,"Config warning: modeling type is missed, VQ will be used\n") ;
            profiler->model_type =  SPR_MODEL_VQ ;
        }

        /* set up modeling parameters */
        if (profiler->model_type == SPR_MODEL_VQ)
        {
            vq_params = NULL ;

            /* read VQ parameters */
            if (cfSelectSection(cfgfile,"VQ"))
            {
                if (cfSelectParam(cfgfile,"size"))
                    profiler->model_params->model_size = cfGetIntValue(cfgfile) ;
                else
                    profiler->model_params->model_size = 64 ; /* default */

                /* allocate memory */
                if ((vq_params = malloc(sizeof(VQParams)))==NULL)
                {
                    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                    return ESR_FAILED ;
                }

                memset(vq_params,0,sizeof(VQParams)) ;

                vq_params->algorithm = VQ_ALG_GLA ;

                vq_params->index_type = VQ_IDX_NONE ;

                if (cfSelectParam(cfgfile,"index codebook"))
                {
                    if (cfGetBoolValue(cfgfile))
                        vq_params->index_type = VQ_IDX_VPT ; /* set default indexing algorithm */
                }

                /* check what indexing algorithm will be used */
                if (vq_params->index_type != VQ_IDX_NONE)
                {
                    if (cfSelectParam(cfgfile,"index type"))
                    {
                        /* TODO: Add reading indexing type*/
                    }
                }

                /* select matching function for VQ and model set indexing */
                if (cfSelectParam(cfgfile,"matching func"))
                {
                    tmp_str = cfGetTextValue(cfgfile) ;

                    if (strcmp(tmp_str,"MSE") == 0)
                    {
                        vq_params->err_func = errMSE ;
                        vq_params->idxerr_func = idxMSE ;

                    } else if (strcmp(tmp_str,"SOAM") == 0)
                    {
                        vq_params->err_func = smSumOfMin ;
                        vq_params->idxerr_func = NULL ; /* TODO: Change function protorypes */
/*                        vq_params->idxerr_func = smSumOfMinIdx ; */

                    } else if (strcmp(tmp_str,"Hausdorff") == 0)
                    {
                        vq_params->err_func = smHausdorff ;
                        vq_params->idxerr_func = NULL ; /* TODO: Change function protorypes */
/*                        vq_params->idxerr_func = smHausdorffIdx ; */

                    } else {

                        fprintf(profiler->logfile,"Config warning (section [Modeling]): unknown matching function, MSE will be used\n") ;
                    }
                }
            }
            

            profiler->model_params->alg_handle = vqCreate(vq_params) ;
            free(vq_params) ;

            if (profiler->model_params->alg_handle == NULL) return ESR_FAILED ;

            profiler->model_params->match_func = (MODELMATCH) vqMatch ;
            profiler->model_params->train_func = (MODELTRAIN) vqTrain ;
            profiler->model_params->destroy_func = (MODELDESTR) vqDestroyModel ; 
            profiler->model_params->load_func = vqLoadModel ;

        }

    }

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

int cfload_fextr(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) 
{
    char * tmp_str ;

    if (cfSelectSection(cfgfile,"Feature extraction"))
    {
    /* check known feature types 
        load means don't extract, just load vectors */
        if (cfSelectParam(cfgfile,"feature type"))
        {
            tmp_str = cfGetTextValue(cfgfile) ;

            if (strcmp(tmp_str,"MFCC") == 0)
                profiler->feat_type = SPR_FEAT_MFCC ;
            else if (strcmp(tmp_str,"LOAD") == 0)
                profiler->feat_type = SPR_FEAT_LOAD ;

            else {

                fprintf(profiler->logfile,"Config warning (section [Feature extraction]): feature type is missed, MFCC will be used\n") ;
                profiler->feat_type = SPR_FEAT_MFCC ;
            }
        }

        /* dummy configuration */
        if (profiler->feat_type == SPR_FEAT_LOAD && profiler->action == SPR_ACTION_FEXTR)
        {
            fprintf(profiler->logfile,"Config error: incorrect action or feature type\n") ;
            return ESR_FAILED ;
        }

        /* if features should be extracted */
        if (profiler->feat_type != SPR_FEAT_LOAD)
        {
            if ((profiler->fe_params = malloc(sizeof(FE_params)))==NULL)
            {
                ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                return ESR_FAILED ;
            }

            memset(profiler->fe_params,0,sizeof(FE_params)) ;

            /* default values */
            profiler->fe_params->win_size_ms = 30 ;
            profiler->fe_params->win_shift_ms = 10 ;

            /* window size in ms */
            if (cfSelectParam(cfgfile,"window size"))
                profiler->fe_params->win_size_ms = cfGetIntValue(cfgfile) ;
            else
                fprintf(profiler->logfile,"Config warning (section [Feature extraction]): window size is missed, 30 ms will be used\n") ;

            /* window shift in ms */
            if (cfSelectParam(cfgfile,"window shift"))
                profiler->fe_params->win_shift_ms = cfGetIntValue(cfgfile) ;
            else
                fprintf(profiler->logfile,"Config warning (section [Feature extraction]): window shift is missed, 10 ms will be used\n") ;

            /* check if data preprocessing is needed */
            if (cfSelectParam(cfgfile,"preprocess") && cfGetBoolValue(cfgfile))
            {
                profiler->flags |= SPR_PREPROCESS ;

                if (cfSelectSection(cfgfile,"Preprocess"))
                {
                    if ((profiler->fe_params->pp = ppCreate())==NULL)
                    {
                        fprintf(profiler->logfile,"Fatal error: Not enough memory to continue\n") ;
                        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                        return ESR_FAILED ;
                    }

                    /* silence removal */
                    if (cfSelectParam(cfgfile,"silence removal"))
                    {
                        if (cfGetBoolValue(cfgfile))
                        {
                            profiler->flags |= SPR_REMOVE_SILENCE ;
                        }
                        
                        /* LOAD SILENCE REMOVAL PARAMETERS !!! */
                    }

                    /* dc removal */
                    if (cfSelectParam(cfgfile,"dc removal") && cfGetBoolValue(cfgfile))
                        ppAddFunction(profiler->fe_params->pp,dspDCRemove,NULL) ;

                    /* high emphasis filtering */
                    if (cfSelectParam(cfgfile,"high emphasis filtering") && cfGetBoolValue(cfgfile))
                        ppAddFunction(profiler->fe_params->pp,dspHighEmphFilter,NULL) ;
                    
                    /* empty section or just silence removal is needed */
                    if (profiler->fe_params->pp->size == 0)
                    {
                        ppDestroy(profiler->fe_params->pp) ;
                        profiler->fe_params->pp = NULL ;
                    }
                }
                
            }
        }

    }

    if (profiler->feat_type == SPR_FEAT_MFCC)
    {
        /* read MFCC parameters */
        if (cfload_mfcc(profiler,cfgfile,argc,argv) == ESR_FAILED)
        {
            return ESR_FAILED ;
        }

    }

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

/* load MFCC parameters */
int cfload_mfcc(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) 
{
    char * tmp_str ;
    MFCCParams * mfcc_params ;

    if ((mfcc_params = malloc(sizeof(MFCCParams)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return ESR_FAILED ;
    }

    /* default mfcc parameters */
    mfcc_params->order = 12 ;
    mfcc_params->num_mel_filters = 0 ; /* auto */
    mfcc_params->window_func = wfHamming ;
    mfcc_params->data_preproc = profiler->fe_params->pp ;
    mfcc_params->flags = MFCC_FILTER_DATA | MFCC_PROTECT_INPUT ;
    mfcc_params->spec_type = MFCC_SPECTRUM_MAGNITUDE ;

    /* set default sampling frequency */
    mfcc_params->sampling_freq = 8000 ;

    /* set window size in samples */
    mfcc_params->win_size = (profiler->fe_params->win_size_ms * mfcc_params->sampling_freq) / 1000 ;

    if (cfSelectSection(cfgfile,"MFCC"))
    {
        if (cfSelectParam(cfgfile,"order"))
            mfcc_params->order = cfGetIntValue(cfgfile) ;

        if (cfSelectParam(cfgfile,"mel filters"))
        {
            if (strcmp(cfGetTextValue(cfgfile),"auto") != 0)
                mfcc_params->num_mel_filters = cfGetIntValue(cfgfile) ;
        }

        if (cfSelectParam(cfgfile,"window function"))
        {
            tmp_str = cfGetTextValue(cfgfile) ;

            if (strcmp(tmp_str,"Hamming") == 0)
                mfcc_params->window_func = wfHamming ;
            else if (strcmp(tmp_str,"Barlett") == 0)
                mfcc_params->window_func = wfBarlett ;
            else if (strcmp(tmp_str,"Hanning") == 0)
                mfcc_params->window_func = wfHanning ;
            else if (strcmp(tmp_str,"Blackman") == 0)
                mfcc_params->window_func = wfBlackman ;
            else if (strcmp(tmp_str,"Gaussian") == 0)
                mfcc_params->window_func = wfGaussian ;
            else {                     
                
                fprintf(profiler->logfile,"Config warning (section [MFCC]): window function is missed, Hamming will be used\n") ;
                mfcc_params->window_func = wfHamming ;
            }
        }
    }

    /* compute number of mel-filters automatically if needed */
    if (mfcc_params->num_mel_filters == 0)
    {
        profiler->fe_params->mel_filters_auto = 1 ;
        mfcc_params->num_mel_filters = (int)ceil((double)(3 * log(mfcc_params->sampling_freq))) ;
    }

    /* create algorithm handle */
    if ((profiler->fe_params->alg_handle = mfccCreate(mfcc_params))==NULL)
    {
        fprintf(profiler->logfile,"Config error (section [MFCC]): Can't create feature extractor\n") ;
        free(mfcc_params) ;
        return ESR_FAILED ;
    }

    /* NOTE: if sampling frequency for input file will change then MFCC handle must be recreated
             and number of mel-filters re-computed (if it was set to "auto") */

    profiler->fe_params->alg_params = mfcc_params ;

    profiler->fe_params->extractor_func = mfccExtractVector ;       
    profiler->fe_params->destroy_func = mfccDestroy ;


    /* if ((profiler->fe_params->alg_handle = mfccCreate(mfcc_params))==NULL)
    fprintf(profiler->logfile,"Can't create feature extractor with this parameters") ; */

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

int cfload_matchproc(SProfiler * profiler,HCFGFILE cfgfile,
                 int argc,char * argv[]) 
{
    char * tmp_str ;

    if (profiler->action == SPR_ACTION_MATCH || 
        profiler->action == SPR_ACTION_VERIFY)
    {
        if (cfSelectSection(cfgfile,"Matching preprocess"))
        {
            if (cfSelectParam(cfgfile,"prequantization") && cfGetBoolValue(cfgfile))
                profiler->flags |= SPR_PREQUANTIZE ;

            if (cfSelectParam(cfgfile,"speaker pruning") && cfGetBoolValue(cfgfile))
            {
                profiler->flags |= SPR_SPRUNING ;

                if (cfSelectParam(cfgfile,"pruning type"))
                {
                    tmp_str = cfGetTextValue(cfgfile) ;

                    if (strcmp(tmp_str,"REORDERING") == 0)
                        profiler->pruning_type = SPR_PRUNING_REORD ;
                    else if (strcmp(tmp_str,"PQBASED") == 0)
                        profiler->pruning_type = SPR_PRUNING_PQBASED ;
                    else if (strcmp(tmp_str,"HIERARCH") == 0)
                        profiler->pruning_type = SPR_PRUNING_HIERARCH ;
                    else if (strcmp(tmp_str,"CFBASED") == 0)
                        profiler->pruning_type = SPR_PRUNING_CFBASED ;
                    else if (strcmp(tmp_str,"STATIC") == 0)
                        profiler->pruning_type = SPR_PRUNING_STATIC ;
                    else if (strcmp(tmp_str,"ADAPTIVE") == 0)
                        profiler->pruning_type = SPR_PRUNING_ADAPTIVE ;
                    else {

                        fprintf(profiler->logfile,"Config warning (section [Matching preprocess]): unknown pruning type, CFBASED will be used\n") ;
                        profiler->pruning_type = SPR_PRUNING_CFBASED ;
                    }
                     
                } else {

                    fprintf(profiler->logfile,"Config warning (section [Matching preprocess]): pruning type is missed, CFBASED will be used\n") ;
                    profiler->pruning_type = SPR_PRUNING_CFBASED ;
                } 
            }
        }

        /* read speaker pruning parameters */
        if (profiler->flags & SPR_SPRUNING)
        {
            /* read pruning parameters */

            if (profiler->pruning_type == SPR_PRUNING_REORD)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprReordPruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"Reordering pruning"))
                {
                }

            } else if (profiler->pruning_type == SPR_PRUNING_PQBASED)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprPqBasedPruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"PQ based pruning"))
                {
                }

            } else if (profiler->pruning_type == SPR_PRUNING_HIERARCH)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprHierchPruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"Hierarchical pruning"))
                {
                }

            } else if (profiler->pruning_type == SPR_PRUNING_CFBASED)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprConfBasedPruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"Confidence based pruning"))
                {
                }

            } else if (profiler->pruning_type == SPR_PRUNING_STATIC)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprStaticPruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"Static pruning"))
                {
                    profiler->model_params->spruning_params = malloc(sizeof(SPRStatParams)) ;

                    if (cfSelectParam(cfgfile,"interval"))
                        ((SPRStatParams*)profiler->model_params->spruning_params)->step = cfGetIntValue(cfgfile) ;
                    else
                        ((SPRStatParams*)profiler->model_params->spruning_params)->step = 20 ;

                    if (cfSelectParam(cfgfile,"number"))
                        ((SPRStatParams*)profiler->model_params->spruning_params)->number = cfGetIntValue(cfgfile) ;
                    else
                        ((SPRStatParams*)profiler->model_params->spruning_params)->number = 5 ;
                }

            } else if (profiler->pruning_type == SPR_PRUNING_ADAPTIVE)
            {
                profiler->model_params->spruning_func = (SPRUNING) sprAdaptivePruning ; 

                /* load algorithm parameters */
                if (cfSelectSection(cfgfile,"Adaptive pruning"))
                {
                    profiler->model_params->spruning_params = malloc(sizeof(SPRAdapParams)) ;

                    if (cfSelectParam(cfgfile,"interval"))
                        ((SPRAdapParams*)profiler->model_params->spruning_params)->step = cfGetIntValue(cfgfile) ;
                    else
                        ((SPRAdapParams*)profiler->model_params->spruning_params)->step = 20 ;

                    if (cfSelectParam(cfgfile,"threshold"))
                        ((SPRAdapParams*)profiler->model_params->spruning_params)->threshold = cfGetDoubleValue(cfgfile) ;
                    else
                        ((SPRAdapParams*)profiler->model_params->spruning_params)->threshold = 0.9 ;
                }

            }

            if (profiler->model_params->spruning_params == NULL)
                fprintf(profiler->logfile,"Config warning: pruning parameters are missed, default will be used\n") ;
        }
    }

    /* TODO: Add usage of command line parameters */

    return ESR_SUCCEED ;
}

/*======================== Functions ==============================*/
