/*-----------------------------------------------------------------*/
/* Module : Speaker profiler                                       */
/*                                                                 */
/* Description : Command line application working with speaker     */
/*               profiles                                          */
/*                                                                 */
/*-----------------------------------------------------------------*/

/*
    $Id: sprofiler.c,v 1.15 2003/12/19 06:59:31 ekarpov Exp $
*/


/* windows mechanism for memory leaks checking */
#if defined(_DEBUG) && defined(WIN32)
 #define _CRTDBG_MAP_ALLOC
 #include <stdlib.h>
 #include <crtdbg.h>
#else
 #include <stdlib.h>
#endif

/*-----------------------------------------------------------------*/

#include <stdio.h>
#include <time.h>

#include <audiostream.h>

#include "vq.h" /* TODO: get rid of this header */
#include "mfcc.h" /* TODO: get rid of this header */
#include <math.h> /* TODO: get rid of this header */

#include "sprcfg.h"

/*-----------------------------------------------------------------*/

double get_difftime(clock_t finish,clock_t start) ;
int write_time_string(FILE * pfile) ;

HFSET load_features(SProfiler * profiler,char * input_file,int curr_speaker,char * buff) ;
HFSET extract_features(SProfiler * profiler,char * input_file,int curr_speaker,char * buff) ;

int match_speaker(SProfiler * profiler) ;
int verify_speaker(SProfiler * profiler) ;

void * model_speaker(SProfiler * profiler) ;

int save_features(SProfiler * profiler,char * output_file,int curr_speaker,char * buff) ;
int save_model_and_add(SProfiler * profiler,char * output_file,int curr_speaker,char * buff, void * pmodel) ;

int load_models_timit(SProfiler * profiler,char * buff) ;

void * load_vq_model(char * file_name,void * params) ;
int save_vq_model(char * file_name,void * model) ;

/*-----------------------------------------------------------------*/

int main(int argc,char * argv[])
{
    SProfiler * profiler ;
    clock_t start_clock, stop_clock;
	clock_t app_start_clock, speaker_start_clock ;

    int speaker, total_errors, best_speaker, correct_answers ;

    double avg_feat_extr_time, avg_total_time, avg_action_time ;

    char tmp_file_name[FILENAME_MAX+1] ;

    void * pmodel ;

    total_errors = 0 ;
	avg_feat_extr_time = 0.0 ;
	avg_total_time = 0.0 ;
    avg_action_time = 0.0 ;
	correct_answers = 0 ;

    /* load configuration */
    profiler = load_config(argc,argv) ;

    /* check parameters */
    if (profiler == NULL)
    {
        fprintf(stderr,
			"Fatal error: Can't load configuration file\n") ;
        return 0 ; /* exit */
    }

    /* application started */
    app_start_clock = clock() ;

    fprintf(profiler->logfile,"\n") ;
    write_time_string(profiler->logfile) ;
    fprintf(profiler->logfile," Speaker profiler started\n") ;

    if (profiler->action == SPR_ACTION_MATCH || 
        profiler->action == SPR_ACTION_VERIFY)
    {
        fprintf(profiler->logfile,"\tLoading speaker database... ") ;

        /* measure time */
        start_clock = clock() ;

        profiler->sdb = sdbOpen(profiler->database_file,SDB_OPEN_READ_ONLY) ;

        /* measure time */
        stop_clock = clock() ;

        /* record time measurements */
        fprintf(profiler->logfile,"done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;

        if (profiler->sdb == NULL)
        {
            /* error should be reported already */
            close_profiler(profiler) ;
            return 0 ;
        }

        fprintf(profiler->logfile,"\tLoading speaker models... ") ;
        
        if (profiler->action == SPR_MODEL_VQ)
        {
            /* measure time */
            start_clock = clock() ;
            
            /* create set to store models */
            profiler->models = msCreate(sdbSize(profiler->sdb),NULL) ;
            
            for (speaker=1; speaker <= sdbSize(profiler->sdb); speaker++)
            {
                msAddModel(profiler->models,sdbLoadModel(profiler->sdb,load_vq_model,profiler),speaker,0,NULL) ;
                sdbSelectNext(profiler->sdb) ;
            }
            
            /* measure time */
            stop_clock = clock() ;
            
        } else {
            
            fprintf(profiler->logfile,"\n\tUnknown model type: terminate\n") ;
            close_profiler(profiler) ;
            return 0 ;
            
        }
        
        fprintf(profiler->logfile,
            "done, %i models loaded,took %.3f seconds",
            msGetSize(profiler->models),
            get_difftime(stop_clock,start_clock)) ;
        
        /* TODO: insert number of errors if any */
        
        fprintf(profiler->logfile,"\n") ;
        
        if (profiler->models == NULL)
        {
            /* error should be reported already */
            close_profiler(profiler) ;
            return 0 ;
        }


    } else if (profiler->action == SPR_ACTION_MODEL)
    {
        /* try to open existing database first */

        fprintf(profiler->logfile,"\tOpening speaker database... ") ;

        /* measure time */
        start_clock = clock() ;

        profiler->sdb = sdbOpen(profiler->database_file,SDB_OPEN_CHANGE) ;

        if (profiler->sdb == NULL)
        {
            /* open new database */
            profiler->sdb = sdbOpen(profiler->database_file,SDB_OPEN_NEW) ;
        }

        /* measure time */
        stop_clock = clock() ;

        /* record time measurements */
        fprintf(profiler->logfile,"done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;

        if (profiler->sdb == NULL)
        {
            /* error should be reported already */
            close_profiler(profiler) ;
            return 0 ;
        }
    }


    speaker = 1 ;

    /* main cycle */
    while (!(profiler->flags & SPR_STATE_FINISHED))
    {
        fprintf(profiler->logfile,"\n") ;
        write_time_string(profiler->logfile) ;
        fprintf(profiler->logfile," Start speaker number %i \n",profiler->speaker_ids[speaker-1]) ;
        fflush(profiler->logfile) ;

        /* measure time */
        speaker_start_clock = clock() ;

        /* check if feature extraction is needed */
        if (profiler->feat_type == SPR_FEAT_LOAD)
        {
            fprintf(profiler->logfile,"\tloading features... ") ;

            /* load feature vectors */
            profiler->test_set = load_features(profiler,profiler->input_files[speaker-1],speaker,tmp_file_name) ;

            if (profiler->test_set)
            {
                fprintf(profiler->logfile," done, %i vectors loaded, ",fvsGetSize(profiler->test_set)) ;

            } else {
                
                total_errors ++ ;

                /* continue ; */ /* skip further processing */
            }

        } else {

            fprintf(profiler->logfile,"\textracting features... ") ;

            /* extract features */
            profiler->test_set = extract_features(profiler,profiler->input_files[speaker-1],speaker,tmp_file_name) ;
        }

        /* measure time */
        stop_clock = clock() ;

        /* measure statistics */
        avg_feat_extr_time += get_difftime(stop_clock,speaker_start_clock) ;

        if (profiler->test_set)
            fprintf(profiler->logfile,"took %.3f seconds\n",get_difftime(stop_clock,speaker_start_clock)) ;

        /* measure time */
        start_clock = clock() ;

        stop_clock = start_clock ;

        /* start actions */
        if (profiler->test_set == NULL)
        {
            /* skip processing */
            fprintf(profiler->logfile,"\tprocessing is skiped\n") ;

        } else if (profiler->action == SPR_ACTION_MATCH)
        {        
            fprintf(profiler->logfile,"\tmatching... ") ;

            /* match feature vectors against database */
			best_speaker = match_speaker(profiler);
            if (0 != best_speaker)
            {
                /* measure time */
                stop_clock = clock() ;

                fprintf(profiler->logfile," done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;

                /* report result */
                fprintf(profiler->logfile,"\tMatching results:\n") ;
                fprintf(profiler->logfile,"\t\tidentified as speaker: %i\n",best_speaker) ;
                fprintf(profiler->logfile,"\t\tmatching score: %.5f\n",profiler->curr_score) ;
                fprintf(profiler->logfile,"\t\tcorrect answer: ") ;

                if (best_speaker == profiler->speaker_ids[speaker-1])
                {
                    fprintf(profiler->logfile,"Yes\n") ;
                    correct_answers++ ;

                } else
                    fprintf(profiler->logfile,"No\n") ;

            } else {

                fprintf(profiler->logfile," failed\n") ;
            }

        } else if (profiler->action == SPR_ACTION_MODEL) 
        {
            fprintf(profiler->logfile,"\tcreate new model... ") ;

            /* create speaker model */
            if ((pmodel = model_speaker(profiler))!=NULL)
            {
                /* measure time */
                stop_clock = clock() ;

                /* save model */
                if (save_model_and_add(profiler,profiler->output_files[speaker-1],
                                    profiler->speaker_ids[speaker-1],tmp_file_name,pmodel))
                {
                    /* measure time */
                    stop_clock = clock() ;

                    fprintf(profiler->logfile," done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;
                }

                /* destroy model */
                profiler->model_params->destroy_func(profiler->model_params->alg_handle,pmodel) ;

            }

        } else if (profiler->action == SPR_ACTION_VERIFY) 
        {
            fprintf(profiler->logfile,"\tverify speaker... ") ;

            /* verify speaker */
			best_speaker = verify_speaker(profiler);
            if (0 != best_speaker)
            {
                /* measure time */
                stop_clock = clock() ;

                fprintf(profiler->logfile," done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;

                /* report result */

            }

        } else if (profiler->action == SPR_ACTION_FEXTR) 
        {
            fprintf(profiler->logfile,"\tsave vectors... ") ;

            /* save vectors */
            if (save_features(profiler,profiler->output_files[speaker-1],speaker,tmp_file_name))
            {
                /* measure time */
                stop_clock = clock() ;

                fprintf(profiler->logfile," done, took %.3f seconds\n",get_difftime(stop_clock,start_clock)) ;
            }
        }

        avg_action_time += get_difftime(stop_clock,start_clock) ;

        write_time_string(profiler->logfile) ;
        fprintf(profiler->logfile," Stop speaker number %i ",profiler->speaker_ids[speaker-1]) ;
        fprintf(profiler->logfile,"(time spent %.5f seconds)\n",get_difftime(stop_clock,speaker_start_clock)) ;

        avg_total_time += get_difftime(stop_clock,speaker_start_clock) ;

        fvsDestroy(profiler->test_set) ;
        profiler->test_set = NULL ;

        if (speaker >= profiler->input_list_size) 
            profiler->flags |= SPR_STATE_FINISHED ; 
        else
            speaker++ ;
    }

    /* measure time */
    stop_clock = clock() ;

    /* report final time */
    fprintf(profiler->logfile,"\n") ;
    write_time_string(profiler->logfile) ;
    fprintf(profiler->logfile," Speaker profiler stopped, time spent %.5f ",get_difftime(stop_clock,app_start_clock)) ;

    if (total_errors)
        fprintf(profiler->logfile,"seconds with %i errors",total_errors) ;
    else
        fprintf(profiler->logfile,"seconds\n") ;

    /* print statistics */
    if (profiler->action != SPR_ACTION_FEXTR)
    {
        fprintf(profiler->logfile,"\nWorking statistics:\n") ;
        fprintf(profiler->logfile,"\tcorrect answers: \t%i from %i\n",correct_answers,speaker) ;
        fprintf(profiler->logfile,"\tidentification rate: \t%.3f(%%)\n",(double)correct_answers/(double)speaker) ;
        fprintf(profiler->logfile,"\terror rate: \t\t%.3f(%%)\n",(double)(speaker - correct_answers)/(double)speaker) ;
    }

    fprintf(profiler->logfile,"\nTime statistics:\n") ;
    fprintf(profiler->logfile,"\tfeature extraction:\t%.5f(s)\n",avg_feat_extr_time/speaker) ;
    fprintf(profiler->logfile,"\taction:\t\t\t%.5f(s)\n",avg_action_time/speaker) ;
    fprintf(profiler->logfile,"\ttotal time:\t\t%.5f(s)\n",avg_total_time/speaker) ;

    /* destroy models */
    for (speaker=1;speaker<=msGetSize(profiler->models);speaker++)
        profiler->model_params->destroy_func(profiler->model_params->alg_handle,
                                             msGetModel(profiler->models,speaker)) ;

    /* release resources */
    close_profiler(profiler) ;

/* windows mechanism for memory leaks checking */
#if defined(_DEBUG) && defined(WIN32) 
    if (_CrtDumpMemoryLeaks()) printf("\nMEMORY LEAKS DETECTED !!!\n\n") ; /* memory leaks, A-A-A !!! */
#endif

    return 0 ;
} 

/*======================== Functions ==============================*/

double get_difftime(clock_t finish,clock_t start) 
{
    return (double)(finish - start) / CLOCKS_PER_SEC ;
}

int write_time_string(FILE * pfile) 
{
    time_t tmp_time;
    struct tm * curr_time ;
    int ret_val ;
    
    time(&tmp_time) ;
    curr_time = localtime(&tmp_time) ;

    ret_val = fprintf(pfile,"[%2i:%2i:%2i]",curr_time->tm_hour,
        curr_time->tm_min,curr_time->tm_sec) ;

    fflush(pfile) ;

    return ret_val ;
}

/* load features */
HFSET load_features(SProfiler * profiler,char * input_file,int curr_speaker,char * buff) 
{
    if (input_file == NULL) return NULL ;

    sprintf(buff,"%s",input_file) ;

    return fvsLoad(buff) ;
}

/* extract features */
HFSET extract_features(SProfiler * profiler,char * input_file,int curr_speaker,char * buff) 
{
    HASTREAM file_stream ;
    AudioParams audio_params ;
    MFCCParams * mfcc_params ;
    HFSET ret_set = NULL ;

    int window_size ;
    int shift_size ;
    int read_samples, count ;

    double * window_data ;

    if (input_file == NULL) return NULL ;

    if ((file_stream = asOpen(ASTREAM_TYPE_FILE,ASTREAM_MODE_READ,&audio_params,input_file))==NULL)
    {
        return NULL ;
    }

    /* check parameters for MFCC */
    if (profiler->feat_type == SPR_FEAT_MFCC)
    {
        /* get current parameters */
        mfcc_params = (MFCCParams*) profiler->fe_params->alg_params ;

        /* check is sampling rate has changed or handle was not created */
        if (mfcc_params->sampling_freq != audio_params.samples_per_sec || 
            profiler->fe_params->alg_handle == NULL)
        {
            /* destroy old handler */
            profiler->fe_params->destroy_func(profiler->fe_params->alg_handle) ;

            /* set new frequency */
            mfcc_params->sampling_freq = audio_params.samples_per_sec ;

            /* update number of mel-filters if needed */
            if (profiler->fe_params->mel_filters_auto)
                mfcc_params->num_mel_filters = (int)ceil((double)(3 * log(mfcc_params->sampling_freq))) ;

            /* update window size in samples */
            mfcc_params->win_size = (mfcc_params->sampling_freq * profiler->fe_params->win_size_ms) / 1000 ;

            /* create new handler */
            profiler->fe_params->alg_handle = mfccCreate(mfcc_params) ;

            if (profiler->fe_params->alg_handle == NULL)
            {
                asClose(file_stream) ;
                return NULL ;
            }
        }
        
        /* create vector size */
        ret_set = fvsCreate(0,mfcc_params->order) ;

    } else {

        asClose(file_stream) ;
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    /* calculate */
    window_size = (audio_params.samples_per_sec * profiler->fe_params->win_size_ms) / 1000 ;
    shift_size = (audio_params.samples_per_sec * profiler->fe_params->win_shift_ms) / 1000 ;

    /* allocate memory for window */
    window_data = malloc(sizeof(double)*window_size) ;

    if (audio_params.channels == 1) /* simple work with mono signal */
    {
        /* first time reading */
        read_samples = asReadDouble(file_stream,window_data,shift_size) ;

        read_samples = window_size - shift_size ; /* simple assumption - FIX THIS */

        while (read_samples == window_size - shift_size)
        {
            read_samples = asReadDouble(file_stream,window_data + shift_size,window_size - shift_size) ;

            /* skip processing if there is not enough data */
            if (read_samples < window_size - shift_size) continue ;

            /* extract features here */
            fvsAddVector(ret_set,profiler->fe_params->extractor_func(profiler->fe_params->alg_handle,
                            window_data,window_size)) ;

            /* move overlapping part at the beggining */
            for (count = 0 ; count < shift_size;  count++)
                window_data[count] = window_data[window_size - shift_size + count] ;
        }

    } else { /* create temporary buffer and use only first channel */
     
        /* TODO: Add working with multiple channels */
        
    }

    if (file_stream == NULL) return NULL ;

    free(window_data) ;
    asClose(file_stream) ;

    return ret_set ;
}

/* match speaker */
int match_speaker(SProfiler * profiler) 
{
    int speaker,size ;
    int best_speaker = 0 ;

    double min_score ;

    /* check whether speaker pruning is used */
    if (profiler->flags & SPR_SPRUNING)
    {
        /* start speaker pruning */
        best_speaker = profiler->model_params->spruning_func(profiler->model_params->alg_handle,
                profiler->models, profiler->test_set,&(profiler->curr_score),
                profiler->model_params->spruning_params) ;

    } else {

        min_score = 0.0 ; best_speaker = 0 ;

        size = msGetSize(profiler->models) ;

        /* loop trough all speakers */
        for (speaker=1; speaker <= size; speaker++)
        {
            profiler->curr_score = 
                profiler->model_params->match_func(
                            profiler->model_params->alg_handle,
                            msGetModel(profiler->models,speaker),
                            profiler->test_set) ;

            if (best_speaker == 0 || min_score > profiler->curr_score)
            {
                best_speaker = speaker ;
                min_score = profiler->curr_score ;
            }
        }
        
        profiler->curr_score = min_score ;
    }

    return best_speaker ;
}

/* verify speaker */
int verify_speaker(SProfiler * profiler) 
{
    return 0 ;
}

/* create speaker model */
void * model_speaker(SProfiler * profiler) 
{
    return profiler->model_params->train_func(profiler->model_params->alg_handle,
                                    profiler->test_set,profiler->model_params->model_size) ;
}

/* save feature vectors in file */
int save_features(SProfiler * profiler,char * output_file,int curr_speaker,char * buff) 
{
    if (output_file != NULL)
        return fvsSave(profiler->test_set,output_file) ;
    
    sprintf(buff,"spkr%1.fe",curr_speaker) ;

    return fvsSave(profiler->test_set,buff) ; 
}

/* save model in file */
int save_model_and_add(SProfiler * profiler,char * output_file,int curr_speaker,char * buff, void * pmodel) 
{
    char spkr_name[20] ;

    sprintf(spkr_name,"Spkr%i",curr_speaker) ;

    if (output_file != NULL)
    {
        if (vqSaveModel(profiler->model_params->alg_handle,pmodel,output_file))
        {
            sdbAddSpeaker(profiler->sdb,curr_speaker,spkr_name,output_file) ;

            return 1 ;
        }

    } else {
        
        sprintf(buff,"spkr%1.fe",curr_speaker) ;

        if (vqSaveModel(profiler->model_params->alg_handle,pmodel,buff))
        {
            sdbAddSpeaker(profiler->sdb,curr_speaker,spkr_name,buff) ;

            return 1 ;
        }
    }

    return 0 ;

}

void * load_vq_model(char * file_name,void * params) 
{
    SProfiler * profiler = params ;

    return profiler->model_params->load_func(profiler->model_params->alg_handle,file_name) ;
}

int save_vq_model(char * file_name,void * pmodel) 
{
    return 1 ;
}

/*======================== Functions ==============================*/
