/*! 
\file 
\brief Set of models

Functionality to work with set of models

\date 18 November 2003
\author Evgeny Karpov
*/

/*
    $Id: modelset.c,v 1.4 2003/11/18 19:05:47 ekarpov Exp $
*/

#include <stdlib.h>
#include <errno.h>

#include "modelset.h"

/*======================== Structures =============================*/

/*! Keeps model and its parameters */
typedef struct {

    int id ;        /*!< Model ID */
    int flags ;     /*!< Model flags */

    void * model ;  /*!< Pointer to a model */

    void * params ; /*!< Model parameters */

} ModelRef ;

/*! Set of models */
typedef struct {

    int model_number ;  /*!< amount of models in set */
    int array_size ;    /*!< buffer size */

    SMFREE free_func ;  /*!< Pointer to a function what will be 
                             used to destroy models when set will be
                             destroyed */

    ModelRef ** models ; /*!< Array of models */

} ModelSet ;

/*======================== Structures =============================*/

#define MSET_INITIAL_SIZE       20

/*======================== Functions ==============================*/

/*!
\brief Creates model set

Allocates initial space for model set

\param size Initial size
\param free_func Function pointer what will be used to destroy
models
\return Handle to model set
\sa msDestroy
\author Evgeny Karpov
*/
HMS msCreate(
	int size,
	SMFREE free_func) 
{
    ModelSet * pmset ;

    if ((pmset = malloc(sizeof(ModelSet)))==NULL)
    {
        errno = ENOMEM ;
        return NULL ;
    }

    /* default size */
    if (size<=0) size = MSET_INITIAL_SIZE ;

    if ((pmset->models = malloc(sizeof(ModelRef*)*size))==NULL)
    {
        errno = ENOMEM ;
        free(pmset) ;
        return NULL ;
    }

    pmset->model_number = 0 ;
    pmset->array_size = size ;

    /* default freeing function */
    /* if (free_func==NULL) free_func = free ; */

    pmset->free_func = free_func ;

    return (HMS) pmset ;
}


/*!
\brief Destroys model set and models

Destroys model set and models

\param ms Handle to model set
\return Nothing
\sa msCreate
\author Evgeny Karpov
*/
void msDestroy(HMS ms) 
{
    ModelSet * pmset ;
    int i ;

    if ((pmset = (ModelSet*) ms)==NULL) return ;

    for (i=0;i<pmset->model_number;i++)
    {
        if (pmset->free_func && pmset->models[i])
            pmset->free_func(pmset->models[i]->model) ;

        free(pmset->models[i]) ;
    }

    free(pmset->models) ;
    free(pmset) ;
}


/*-----------------------------------------------------------------*/



/*!
\brief Add new model to a model set

\param ms Model set handle
\param model The model that needs to be added
\param id The new model identifier
\param flags Flags associated with model
\param params Parameters associated with model
\return model number if succesful, otherwise returns 0
\sa HMS
\author Evgeny Karpov
*/
int msAddModel(HMS ms,void * model,int id,int flags,void * params) 
{
    ModelSet * pmset ;
    ModelRef ** ptmp ;

    if ((pmset = (ModelSet*) ms)==NULL)
    {
        errno = EINVAL ;
        return 0 ;
    }

    if (pmset->model_number >= pmset->array_size)
    {
        /* increase array size */
        if ((ptmp = realloc(pmset->models,sizeof(ModelRef*)*pmset->array_size*2))==NULL)
        {
            errno = ENOMEM ;
            return 0 ;
        }

        pmset->models = ptmp ;
        pmset->array_size *= 2 ;
    }

    pmset->models[pmset->model_number] = malloc(sizeof(ModelRef)) ;

    if (pmset->models[pmset->model_number] == NULL)
    {
        errno = ENOMEM ;
        return 0 ;
    }

    /* copy all parameters */
    pmset->models[pmset->model_number]->id = id ;
    pmset->models[pmset->model_number]->flags = flags ;
    pmset->models[pmset->model_number]->model = model ;
    pmset->models[pmset->model_number]->params = params ;

    pmset->model_number++ ;

    return pmset->model_number ;
}


/*!
\brief Remove model from a model set


\param ms The models set
\param id The identifier of the model that should be removed
\return TODO: describe function return value
\note Not Implemented
\todo Implement it
\sa HMS
\author Evgeny Karpov
*/
int msRemoveModel(HMS ms,int id) 
{
    /* NOT IMPLEMENTED YET */

    return 0 ;
}



/*!
\brief Get model set size

Get the number of models in a model set
\param ms The model set whose size is requested
\return Returns the number of model in the model set
\sa HMS
\author Evgeny Karpov
*/
int msGetSize(HMS ms) 
{
    ModelSet * pmset ;

    if ((pmset = (ModelSet*) ms)==NULL)
    {
        errno = EINVAL ;
        return 0 ;
    }

    return pmset->model_number ;
}

/*-----------------------------------------------------------------*/



/*!
\brief Get model idendtifier for given index

\param ms The model set
\param num The model index
\return The model identifier
\note These functions are supposed to be used in
time consuming  operations. To be more fast
they do not check correctness of incoming parameters.
\sa HMS
\author Evgeny Karpov
*/
int msGetModelId(HMS ms,int num) 
{
    return ((ModelSet*)ms)->models[num-1]->id ;
}


/*!
\brief Get model flags

\param ms The model set
\param num The model index
\return The flags of the model at index \p num
\sa HMS
\author Evgeny Karpov
*/
int msGetModelFlags(HMS ms,int num) 
{
    return ((ModelSet*)ms)->models[num-1]->flags ;
}

int msSetModelFlags(HMS ms,int num,int flags) 
{
    ((ModelSet*)ms)->models[num-1]->flags = flags ;
    return 1 ;
}

void * msGetModelParams(HMS ms,int num) 
{
    return ((ModelSet*)ms)->models[num-1]->params ;
}


/*!
\brief Get model with given index

\param ms The model set where the model is fetched
\param num The number of the model
\return Pointer to the model at the index \p num
\sa HMS
\author Evgeny Karpov
*/
void * msGetModel(HMS ms,int num)
{
    return ((ModelSet*)ms)->models[num-1]->model ;
}

/*-----------------------------------------------------------------*/


/*!
\brief Merge two model sets

\param set1 First model set to be merged
\param set2 The other model set to be merged
\return Handle to the merged model set
\note Not implemented
\todo Implement it!
\sa HMS
\author Evgeny Karpov
*/
HMS msMerge(HMS set1,HMS set2) 
{
    /* NOT IMPLEMENTED YET */

    return NULL ;
}

/*======================== Functions ==============================*/
