/*! 
\file speakerdb.c
\brief Speaker database handling

Provides mechanism for working with simple speaker database.
Speaker database in this realization consits of simple ASCII
file, one record per a row, fields are separated by comma.

$Id: speakerdb.c,v 1.11 2003/12/31 13:05:44 juhani Exp $

\date 03 November 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <csv.h>
#include <errorapi.h>

#include "speakerdb.h"

/*======================== Structures =============================*/

/*! Speaker database handle */
typedef struct {

    int selected ;  /*!< Selected speaker (record) */
    int size ;      /*!< Database size (number of records) */

    FILE * pfile ;  /*!< File pointer to database */

    CSV ** records ;    /*!< Records */
    int array_size ;    /*!< Amount of records what can be hold 
                             without reallocating memory */

    SdbOpenMode mode ;  /*!< Open mode */

    char * file_name ; /*!< file name */
    int state ;     /*!< Databsae state */

} SDBHandle ;

/*======================== Structures =============================*/

#define SDB_ARRAY_INITIAL           20

#define SDB_FIELD_ID                1
#define SDB_FIELD_NAME              2
#define SDB_FIELD_MODEL_PATH        3
#define SDB_FIELD_INDEX_PATH        4

#define SDB_INT_IN_CHAR             7

#define SDB_STATE_NOT_CHANGED       0
#define SDB_STATE_CHANGED           1

#define SDB_STRING_SIZE_INIT        30

/*======================== Functions ==============================*/

/*!
\brief Checks for end of line marker

Checks for '\r' '\n' '\r\n' EOF and destroy them

\param filep File pointer
\param ch Last read charater
\return Zero if end of line was not reached and not zero otherwise
\note For internal use only
\author Evgeny Karpov
*/
static int endofline(FILE * filep, int ch)
{
	int eol ;

	eol = (ch=='\r' || ch=='\n') ;

	if (ch=='\r')
	{
		ch=getc(filep) ;
		if (ch!='\n' && ch!=EOF)
			ungetc(ch,filep) ; /* read too much -> return 'ch' back to stream  */
	}

	return eol ;
}

/*!
\brief Loads speaker database

Parses and loads into memory speaker records saved in
comma separated format.

\param sdb Handle to already opened database
\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note This function can not be used outside of speakerdb module
\sa sdbOpen sdbSave sdbClose
\author Evgeny Karpov
*/
int sdbLoad(SDBHandle * psdb) 
{
    CSV * pcsv ;
    CSV ** tmp_buffer ;

    char * read_str, * tmp_str ;
    int count ;
    char ch ;

    int read_str_size = SDB_STRING_SIZE_INIT ;

    /* allocate initial temporary string */
    if ((read_str = malloc(sizeof(char)*read_str_size))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return ESR_FAILED ;
    }

    /* read data from file */
    while (! feof(psdb->pfile))
    {
        count = 0 ;

        /* read one line into a buffer */
        while ((ch=getc(psdb->pfile)) != EOF && !endofline(psdb->pfile,ch))
        {
            if (count >= read_str_size)
            {
                read_str_size *= 2 ;

                /* allocate new space */
                if ((tmp_str = realloc(read_str,sizeof(char)*read_str_size))==NULL)
                {
                    free(read_str) ;
                    ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;                
                    return ESR_FAILED ;
                }

                read_str = tmp_str ;
            }

            read_str[count] = ch ;

            count++ ;
        }

        /* 6 is minimum number of symbols (3 commas plus 3 values) */
        if (count < 6) continue ;

        /* parse CSV structure */
        if ((pcsv = csvParseLine(read_str,count+1))==NULL)
        {
            free(read_str) ;
            return ESR_FAILED ; /* reason of error is lack of memory */
        }

        if (psdb->size >= psdb->array_size)
        {
            /* allocate new buffer */
            tmp_buffer = realloc(psdb->records,sizeof(CSV*)*(2*psdb->array_size)) ;

            if (tmp_buffer == NULL)
            {
                free(read_str) ;
                csvDestroy(pcsv) ;
                ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
                return ESR_FAILED ;
            }

            /* set new buffer */
            psdb->records = tmp_buffer ;

            /* double array size */
            psdb->array_size *= 2 ;
        }


        /* save new record */
        psdb->records[psdb->size] = pcsv ;

        /* increase size */
        psdb->size++ ;
    }

    free(read_str) ;

    return ESR_SUCCEED ;
}

/*!
\brief Saves speaker database

Saves speaker database into a file, one record per a row,
fields are separated by comma.

\param sdb Handle to already opened database

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note This function can not be used outside of speakerdb module
\sa sdbOpen sdbLoad sdbClose
\author Evgeny Karpov
*/
int sdbSaveDB(SDBHandle * psdb) 
{
    int count,field ;

    if (psdb->mode == SDB_OPEN_READ_ONLY || psdb->pfile == NULL) return ESR_FAILED ;

    /* reopen file */
    if (psdb->mode == SDB_OPEN_CHANGE) 
    {
        fclose(psdb->pfile) ;

        psdb->pfile = fopen(psdb->file_name,"w") ;
    }

    /* write records */
    for (count=0; count<psdb->size; count++)
    {
        /* save first record */
        if (psdb->records[count]->count > 0)
            fprintf(psdb->pfile,"%s",psdb->records[count]->fields[0]) ;

        /* save all other records */
        for (field = 1; field < psdb->records[count]->count ; field++)
            fprintf(psdb->pfile,",%s",psdb->records[count]->fields[field]) ;

        /* write end of string marker */
        if (count < psdb->size - 1) fprintf(psdb->pfile,"\n") ;

    }

    fflush(psdb->pfile) ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Opens speaker database

Opens speaker database stored in a file, one record per a row, fileds are
separated by comma. Parses and loads all values into a memory.

\param file_name Database file name
\param mode Opening mode, can be read only, change or new
\return Returns handle to database or NULL in case of error. Reason of 
error is reported using errorapi.

\sa sdbClose
\author Evgeny Karpov
*/

HSDB sdbOpen(char * file_name,SdbOpenMode mode) 
{
    SDBHandle * psdb ;
    char open_mode[5] ;

    /* create handle */
    if ((psdb = (SDBHandle*) malloc(sizeof(SDBHandle)))==NULL) 
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* set all to zero */
    memset(psdb,0,sizeof(SDBHandle)) ;

    psdb->array_size = SDB_ARRAY_INITIAL ;
    psdb->mode = mode ;
    psdb->file_name = NULL ;

    /* allocate initial memory for records */
    if ((psdb->records = (CSV**) malloc(sizeof(CSV*)*psdb->array_size))==NULL)
    {
        free(psdb) ;
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* file mode */
    if (psdb->mode == SDB_OPEN_READ_ONLY)
        strcpy(open_mode,"r") ;
    else if (psdb->mode == SDB_OPEN_CHANGE)
        strcpy(open_mode,"r+") ;
    else 
        strcpy(open_mode,"w") ;

    /* open file */
    if ((psdb->pfile = fopen(file_name,open_mode))==NULL)
    {
        sdbClose(psdb) ;
        ReportError(ESR_CANT_OPEN_FILE,file_name) ;
        return NULL ;
    }

    /* load database */
    if (psdb->mode == SDB_OPEN_READ_ONLY || psdb->mode == SDB_OPEN_CHANGE)
        if (sdbLoad(psdb) == ESR_FAILED)
        {
            sdbClose(psdb) ;
            ReportError(ESR_CANT_LOAD_FILE,file_name) ;
            return NULL ;
        } ;


    /* close file if we only read data */
    if (psdb->mode == SDB_OPEN_READ_ONLY)
    {
        fclose(psdb->pfile) ;
        psdb->pfile = NULL ;
    }

    /* copy file name if we are in "change" mode*/
    if (psdb->mode == SDB_OPEN_CHANGE)
    {
        psdb->file_name = malloc(sizeof(char)*(strlen(file_name)+1)) ;

        if (psdb->file_name == NULL)
        {
            sdbClose(psdb) ;
            ReportError(ESR_NOT_ENOUGH_MEMORY,file_name) ;
            return NULL ;
        }

        strcpy(psdb->file_name,file_name) ;
    }

    return (HSDB) psdb ;
}

/*!
\brief Closes speaker database

Closes speaker database and frees all allocated resources

\param sdb Handle to already opened database
\return Nothing
\note This function saves data before cleaning
\sa sdbOpen
\author Evgeny Karpov
*/

void sdbClose(HSDB sdb) 
{
    SDBHandle * psdb = sdb;
    int count ;

    if (psdb)
    {
        /* save data if there were changes */
        if (psdb->mode != SDB_OPEN_READ_ONLY && psdb->state != SDB_STATE_NOT_CHANGED) sdbSaveDB(psdb) ;
        
        /* close file if it was opened */
        if (psdb->pfile) fclose(psdb->pfile) ;

        /* free records */
        for (count=0; count<psdb->size; count++)
        {
           if (psdb->records[count]) csvDestroy(psdb->records[count]) ;
        }

        free(psdb->records) ;
        free(psdb->file_name) ;
        free(psdb) ;
    }
}

/*-----------------------------------------------------------------*/

/*!
\brief Loads speaker model

Loads model for selected speaker. For model loading handler to 
model loading routine must be supplied. If loading requires 
additional parameters they can be passed to loading routine also.

\param sdb Handle to already opened database
\param load_func Loading routine
\param load_params Loading params
\return Pointer to a model or NULL in a case of error
\sa SDBMLOAD
\author Evgeny Karpov
*/
void * sdbLoadModel(HSDB sdb,SDBMLOAD load_func,void * load_params) 
{
    /* assert(sdb != NULL) ; */

    return load_func(sdbGetSpeakerModelPath(sdb),load_params) ; 
}

/*-----------------------------------------------------------------*/

int sdbSave(HSDB sdb) 
{
    return sdbSaveDB(sdb) ;
}

int sdbSaveAs(HSDB sdb, char * file_name) 
{
    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Selects record in a database

Selects record in a database.

\param sdb Handle to already opened database
\param num Record number to select

\return In success returns number of selected record, in case of error
returns ESR_FAILED

\sa sdbSelectNext sdbReset
\author Evgeny Karpov
*/
int sdbSelect(HSDB sdb,int num) 
{
    SDBHandle * psdb = sdb;

    if (psdb != NULL && num <= psdb->size && num > 0) 
    {
        psdb->selected = num - 1 ;

        return num ;
    }

    return ESR_FAILED ;
}

/*!
\brief Selects next record

Selects next record in database

\param sdb Handle to already opened database

\return In success returns number of selected record, in case of error
returns ESR_FAILED

\sa sdbSelect sdbReset
\author Evgeny Karpov
*/
int sdbSelectNext(HSDB sdb) 
{
    return sdbSelect(sdb,((SDBHandle*)sdb)->selected+2) ;
}

/*!
\brief Resets current selection to initial

Resets current selection to initial

\param sdb Handle to already opened database

\return In success returns zero , in case of error
returns ESR_FAILED

\sa sdbSelect sdbSelectNext
\author Evgeny Karpov
*/
int sdbReset(HSDB sdb) 
{
    SDBHandle * psdb = sdb;

    if (psdb) return psdb->selected = 0 ;

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Returns the size of database

Returns the number of records loaded into memory

\param sdb Handle to already opened database

\return In success returns size of database, in case of error
returns ESR_FAILED

\author Evgeny Karpov
*/
int sdbSize(HSDB sdb) 
{
    SDBHandle * psdb = sdb;

    if (psdb) return psdb->size ;

    return ESR_FAILED ;
}

/*!
\brief Returns integer value of fields

Returns integer value of fields for selected record

\param sdb Handle to already opened database
\param field Number of field

\return In success returns integer value of field, in case of error
returns ESR_FAILED

\sa sdbGetTextValue sdbSetIntValue sdbSetTextValue
\author Evgeny Karpov
*/
int sdbGetIntValue(HSDB sdb,int field) 
{
    SDBHandle * psdb = sdb;

    if (psdb && psdb->records[psdb->selected] 
             && psdb->records[psdb->selected]->count >= field && field>0)
        return atoi(psdb->records[psdb->selected]->fields[field-1]) ;

    return ESR_FAILED ;
}

/*!
\brief Returns text value of fields

Returns textual value of fields for selected record

\param sdb Handle to already opened database
\param field Number of field

\return In success returns textual value of field, in case of error
returns NULL

\note Returned value must not be freed

\sa sdbGetIntValue sdbSetIntValue sdbSetTextValue
\author Evgeny Karpov
*/
char * sdbGetTextValue(HSDB sdb,int field) 
{
    SDBHandle * psdb = sdb;

    if (psdb && psdb->records[psdb->selected] 
             && psdb->records[psdb->selected]->count >= field && field>0)
        return psdb->records[psdb->selected]->fields[field-1] ;

    return NULL ;
}

/*!
\brief Sets integer value for a field

Sets integer value for a field

\param sdb Handle to already opened database
\param field Number of field
\param value Value to set

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\sa sdbSetTextValue sdbGetIntValue sdbGetTextValue
\author Evgeny Karpov
*/
int sdbSetIntValue(HSDB sdb,int field,int value) 
{
    SDBHandle * psdb = sdb;
    char * tmp_str ;

    if (psdb && psdb->mode != SDB_OPEN_READ_ONLY && psdb->records[psdb->selected] 
             && psdb->records[psdb->selected]->count >= field && field>0)
    {
        tmp_str = realloc(psdb->records[psdb->selected]->fields[field-1],(sizeof(char))*(SDB_INT_IN_CHAR+1)) ;

        if (tmp_str == NULL) return ESR_FAILED ;
        
        psdb->records[psdb->selected]->fields[field-1] = tmp_str ;

        sprintf(psdb->records[psdb->selected]->fields[field-1],"%i",value) ;

        psdb->state = SDB_STATE_CHANGED ;

        return ESR_SUCCEED ;
    }

    return ESR_FAILED ;
}

/*!
\brief Sets textual value for a field

Sets textual value for a field

\param sdb Handle to already opened database
\param field Number of field
\param value Value to set

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note In case if value to set is NULL empty string is assigned to a field.

\note Assigned value is copied and may be freed

\sa sdbSetIntValue sdbGetIntValue sdbGetTextValue
\author Evgeny Karpov
*/
int sdbSetTextValue(HSDB sdb,int field,char * value) 
{
    SDBHandle * psdb = sdb;
    char * tmp_str ;

    if (psdb && psdb->mode != SDB_OPEN_READ_ONLY && psdb->records[psdb->selected] 
             && psdb->records[psdb->selected]->count >= field && field>0)
    {
        if (value)
        {
            tmp_str = realloc(psdb->records[psdb->selected]->fields[field-1],(sizeof(char))*(strlen(value)+1)) ;

            if (tmp_str == NULL) return ESR_FAILED ;

            psdb->records[psdb->selected]->fields[field-1] = tmp_str ;

            strcpy(psdb->records[psdb->selected]->fields[field-1],value) ;

        } else {

            tmp_str = realloc(psdb->records[psdb->selected]->fields[field-1],(sizeof(char))*(strlen(value)+1)) ;
            
            if (tmp_str == NULL) return ESR_FAILED ;

            psdb->records[psdb->selected]->fields[field-1] = tmp_str ;

            strcpy(psdb->records[psdb->selected]->fields[field-1],"") ;
        }

        psdb->state = SDB_STATE_CHANGED ;

        return ESR_SUCCEED ;
    }

    return ESR_FAILED ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Returns speaker ID for selected record

Returns speaker ID for selected record. This function must 
be used to get speaker ID instead of low-level sdbGetIntValue.

\param sdb Handle to already opened database

\return In success returns speaker ID, in case of error
returns ESR_FAILED

\sa sdbGetSpeakerName sdbGetSpeakerModelPath
\author Evgeny Karpov
*/
int sdbGetSpeakerID(HSDB sdb) 
{
    return sdbGetIntValue(sdb,SDB_FIELD_ID) ;
}

/*!
\brief Returns speaker name

Returns speaker name for selected record. This function must 
be used to get speaker name instead of low-level sdbGetTextValue.

\param sdb Handle to already opened database

\return In success returns speaker name, in case of error
returns ESR_FAILED

\note Returned value must not be freed

\sa sdbGetSpeakerID sdbGetSpeakerModelPath
\author Evgeny Karpov
*/
char * sdbGetSpeakerName(HSDB sdb) 
{
    return sdbGetTextValue(sdb,SDB_FIELD_NAME) ;
}

/*!
\brief Returns path to speaker model

Returns path to model for selected record. This function must 
be used to get path instead of low-level sdbGetTextValue.

\param sdb Handle to already opened database

\return In success returns path to speaker model, in case of error
returns ESR_FAILED

\note Returned value must not be freed

\sa sdbGetSpeakerID sdbGetSpeakerName
\author Evgeny Karpov
*/
char * sdbGetSpeakerModelPath(HSDB sdb) 
{
    return sdbGetTextValue(sdb,SDB_FIELD_MODEL_PATH) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Changes speaker ID for selected speaker

Changes speaker ID for selected speaker. This function must be used
instead of low-level sdbSetTextValue. Performs also cheking of 
uniqueness of assigned ID in database. In case if such ID is already 
present in database no changes will be made and error will be returned.

\param sdb Handle to already opened database
\param ID Updated speaker ID

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbSetSpeakerName sdbSetSpeakerModelPath
\author Evgeny Karpov
*/
int sdbSetSpeakerID(HSDB sdb, int ID) 
{
    return sdbSetIntValue(sdb,SDB_FIELD_ID,ID) ; 
}

/*!
\brief Changes speaker name for selected speaker

Changes speaker name for selected speaker. This function must be used
instead of low-level sdbSetTextValue.

\param sdb Handle to already opened database
\param speaker_name Updated speaker name

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbSetSpeakerID sdbSetSpeakerModelPath
\author Evgeny Karpov
*/
int sdbSetSpeakerName(HSDB sdb,char * speaker_name) 
{
    return sdbSetTextValue(sdb,SDB_FIELD_NAME,speaker_name) ; 
}

/*!
\brief Changes path to speaker model for selected speaker

Changes path to speaker model for selected speaker. This function 
must be used instead of low-level sdbSetTextValue.

\param sdb Handle to already opened database
\param model_path Updated path to speaker model

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbSetSpeakerID sdbSetSpeakerName
\author Evgeny Karpov
*/
int sdbSetSpeakerModelPath(HSDB sdb,char * model_path) 
{
    return sdbSetTextValue(sdb,SDB_FIELD_MODEL_PATH,model_path) ; 
}

/*-----------------------------------------------------------------*/

/*!
\brief Adds new record to database

Adds new record to speaker database. Record values must be field
after using standard sdbSetIntValue and sdbSetTextValue functions.

\param sdb Handle to already opened database
\param field_num Number of fields

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbRemoveRecord sdbSetIntValue sdbSetTextValue
\author Evgeny Karpov
*/
int sdbAddRecord(HSDB sdb,int field_num) 
{
    SDBHandle * psdb = (SDBHandle*) sdb ;
    CSV ** tmp_buffer ;
    CSV * pcsv ;
    int field ;

    if (psdb == NULL || psdb->mode == SDB_OPEN_READ_ONLY) 
    {
        return ESR_FAILED ; 
    }

    /* check if we have space for new record */
    if (psdb->size + 1 >= psdb->array_size)
    {
        /* allocate more space */           
        tmp_buffer = realloc(psdb->records,sizeof(CSV*)*(2*psdb->array_size)) ;

        if (tmp_buffer == NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
            return ESR_FAILED ;
        }
        
        /* set new buffer */
        psdb->records = tmp_buffer ;

        /* double array size */
        psdb->array_size *= 2 ;
    }

    /* create new record */
    if ((pcsv = csvCreate(field_num))==NULL)
    {
        return ESR_FAILED ;
    }

    /* fill all values with empty string */
    for (field=0; field < pcsv->count; field++)
    {
        if ((pcsv->fields[field] = malloc(sizeof(char)*2))==NULL)
        {
            ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
            free(pcsv) ;
            return ESR_FAILED ;
        }

        strcpy(pcsv->fields[field],"") ;
    }

    /* save new record */
    psdb->records[psdb->size] = pcsv ;

    /* increase size */
    psdb->size++ ;

    psdb->selected = psdb->size - 1;

    return psdb->size ; 
}

/*!
\brief Removes selected record from database

Removes selected record from speaker database.

\param sdb Handle to already opened database

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbAddRecord
\author Evgeny Karpov
*/
int sdbRemoveRecord(HSDB sdb) 
{
    SDBHandle * psdb = (SDBHandle*) sdb ;
    int count ;

    if (psdb == NULL || psdb->mode == SDB_OPEN_READ_ONLY || psdb->size == 0) 
    {
        return ESR_FAILED ; 
    }

    /* delete selected item */
    csvDestroy(psdb->records[psdb->selected]) ;

    /* copy all values from selected */
    if (psdb->selected + 1 < psdb->size)
    {
        for (count = psdb->selected; count < psdb->size; count++)
            psdb->records[count] = psdb->records[count + 1] ;
    }

    if (psdb->selected > 0) psdb->selected-- ;

    /* decrease size */
    psdb->size-- ;

    psdb->state = SDB_STATE_CHANGED ;

    return ESR_SUCCEED ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Adds new speaker to database

Adds new speaker to database. This function must be used instead 
low-level sdbAddRecord.

\param sdb Handle to already opened database
\param id Speaker ID
\param name Speaker name
\param model_path Path to speaker model

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbRemoveSpeaker
\author Evgeny Karpov
*/
int sdbAddSpeaker(HSDB sdb,int id,char * name,char * model_path) 
{
    SDBHandle * psdb = (SDBHandle*) sdb ;

    if (psdb == NULL || psdb->mode == SDB_OPEN_READ_ONLY) 
    {
        return ESR_FAILED ; 
    }

    if (sdbAddRecord(sdb,3) == ESR_FAILED) return ESR_FAILED ;

    /* fill values */
    if (sdbSetSpeakerID(sdb,id) == ESR_SUCCEED)
        if (sdbSetSpeakerName(sdb,name) == ESR_SUCCEED)
            if (sdbSetSpeakerModelPath(sdb,model_path) == ESR_SUCCEED) 
                return ESR_SUCCEED ;

    psdb->state = SDB_STATE_CHANGED ;

    return psdb->selected + 1;

}


/*!
\brief Removes speaker from database

Removes speaker from database.This function must be used instead 
low-level sdbRemoveRecord if speaker ID is known.

\param sdb Handle to already opened database
\param speaker_num Speaker ID

\return In success returns ESR_SUCCEED, in case of error
returns ESR_FAILED

\note sdbSave must be used to save changes before closing
\note Not implemented yet
\todo To implement

\sa sdbRemoveRecord
\author Evgeny Karpov
*/
int sdbRemoveSpeaker(HSDB sdb,int speaker_id) 
{
    return ESR_FAILED ; /* NOT IMPLEMENTED YET */
}

/*======================== Functions ==============================*/

