/*! 
\file errfunc.c
\brief Clustering error function implementations

This module implements several different
error measures that can be used in clustering algorithms.\n
$Id: errfunc.c,v 1.14 2003/11/20 11:36:29 ekarpov Exp $
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>


#include <errorapi.h>


#include "errfunc.h"

/*-----------------------------------------------------------------*/

#ifndef max
/*! \brief Resolve maximum of two items */
#define max(a,b)    (((a) > (b)) ? (a) : (b))
/*! \brief Resolve minimum of two items */
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

/*-----------------------------------------------------------------*/


/*======================== Functions ==============================*/

/*!
\brief Compute Mean Squared Error (MSE)

Compute Mean Squared Error (MSE) for a vector set
of cetroids used for classifying a test set.
\param codebook Codebook used for NN-classification
\param test_set Test set that is calssified
\param dist_func Distance function used for computing the MSE
\return Mean Squared Error or negative value in case of failure
\todo Explain in more detail what is computed
\todo There is no need to divide with set size
in the main loop to avoid floating point overflow
\todo Implementation is not optimal:
use fvsGetVectorBuff to get the vector pointer
table before loop instead of calling fvsGetVector in the loop
or use FVS_GET_ELEMENT macro
\todo Check subroutine call results
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
double errMSE(
	HFSET codebook,
	HFSET test_set,
	DISTFUNC dist_func,
    void * dfunc_params) 
{
    int vector_size, test_size, cb_size ;
    int  i,j;

    double ret_val, min_dist, dist ;

    /* check input parameters */
    if (codebook==NULL || test_set==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    /* set default distance function */
    if (dist_func==NULL)
    {
        dist_func = dfEuclidean ;
    }

    ret_val = 0.0 ;
    
    /* copy values to local variables
       to avoid unnecessary function calls */
    test_size = fvsGetSize(test_set) ;
    cb_size = fvsGetSize(codebook) ;

    vector_size = min(fvsGetVectorSize(codebook),
                      fvsGetVectorSize(test_set)) ;

    /* check sizes */
    if (test_size==0 || cb_size==0)
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED ;
    }

    for (i=1;i<=test_size;i++)
    {
        min_dist = dist_func(fvsGetVector(codebook,1),
                             fvsGetVector(test_set,i),vector_size,dfunc_params) ;

        for (j=2;j<=cb_size;j++)
        {
            dist = dist_func(fvsGetVector(codebook,j),
                             fvsGetVector(test_set,i), vector_size,dfunc_params) ;

            min_dist = min(dist, min_dist) ;
        } 

        /* divide by test_size to avoid floating number overflow */
        ret_val += (min_dist*min_dist)/test_size ;
    }

    return ret_val ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Compute Total Squared Error (MSE)

Compute Total Squared Error (MSE)
between test set and centroids stored in codebook.
The TSE is the MSE but without dividing
by the test set size???
\param codebook Codebook
\param test_set Test set that is compared with codebook
\param dist_func Distance function used
\return Total Squared Error or negative value in case of failure
\todo Implementation can be done by calling errMSE
and multiplying result with size of test_set, or
in fact the code of errMSE should be here
without the unnecessary dividing and then
errMSE should call this function and divide.
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
double errTSE(
	HFSET codebook,
	HFSET test_set,
	DISTFUNC dist_func,
    void * dfunc_params) 
{
    ReportError(ESR_NOT_IMPLEMENTED,
                "errTSE (Total Squared Error)");
    return -1.0 ;
}

/*======================== Functions ==============================*/
