/*! 
\file gla.c
\brief Generalized Lloyd Algorithm implementation

An implementation of the Generalized Lloyd Algorithm
(GLA algorithm).
TODO: explain here shortly what is the GLA
algorithm and what it is used for.\n
$Id: gla.c,v 1.15 2004/01/08 09:40:17 juhani Exp $
\todo Add reference to litterature in here.
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>


#include <errorapi.h>


#include "gla.h"

/*-----------------------------------------------------------------*/


/*-----------------------------------------------------------------*/


int make_new_partition(HFSET cb, HFSET fvs, int * partition, 
                       DISTFUNC dist_func, void * dfunc_params) ;

int compute_new_centroids(HFSET cb, HFSET fvs, int * partition) ;


/*======================== Functions ==============================*/


/*!
\brief Adjust codebook centroids with GLA iteration

The generalized Lloyd algorithm (GLA) is a method
for iteratively clustering a set of vectors.
Starting with a set of centroids, the algorithm
iteratively computes
<ul>
<li>new NN-partition of the vectors based on codebook, and
<li>new codebook as averages of partition subsets,
</ul>
until the iteration converges or maximum number
of iterations have been performed.
\param init_cb Starting point for the iteration:
a codebook, or a set of centroids of a (NN-)partition
\param fvs Set of vectors whose clustering is needed
\param params Data for the algorithm including
maximum number of iterations and convergence criterion,
or NULL for using the default parameters.
\param ret_err Error measure of the final clustering
\return 1 for success, 0 for failure
\todo If GLA_USE_INDEXING flag is set, then nothing is computed.
Fix this.
\todo Check return values from function calls
\todo Use macros for hard coded default values
or some other mechanism
\sa HFSET (a pointer to FVSet)
\sa GLAParams
\author Evgeny Karpov
*/
int glaRun(HFSET init_cb, HFSET fvs,
           GLAParams * params, double * ret_err) 
{
    GLAParams gla ;

    int iteration, vector_size ;
    double curr_err, tmp_err ;

    int * partition ;

    if ((init_cb==NULL) || (fvs==NULL) || (NULL == ret_err))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL) ;
        return 0 ;
    }

    /* default functions */
    gla.dist_func = dfEuclidean ;
    gla.dfunc_params = NULL ;
    gla.err_func = errMSE ;
    gla.idxerr_func = idxMSE ;

    if (params)
    {
        gla.max_iteration = params->max_iteration ;
        gla.threshold = params->threshold ;
        gla.flags = params->flags ;

        if (params->flags & GLA_DISTANCE_FUNC)
            gla.dist_func = params->dist_func ;

        if (params->flags & GLA_ERROR_FUNC)
            gla.err_func = params->err_func ;

        if (params->flags & GLA_FERROR_FUNC)
            gla.idxerr_func = params->idxerr_func ;

    } else {

        /* default values */
        gla.max_iteration = 100 ;
        gla.threshold = 1e-05 ;
        gla.flags = GLA_USE_INDEXING ;
    }

    vector_size = fvsGetVectorSize(fvs) ;

    partition = malloc(sizeof(int)*fvsGetSize(fvs));
    if (NULL == partition)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL) ;
        return 0 ;
    }

    /* evaluate current solution */
    if (gla.flags & GLA_USE_INDEXING)
    {
        /* create temporary codebook */

        /* curr_err = gla.idxerr_func(init_cb,fvs,gla.dist_func,NULL,NULL) ; */
        curr_err = DBL_MAX;

    } else {

        curr_err = gla.err_func(init_cb,fvs,gla.dist_func,gla.dfunc_params) ;
    }

    tmp_err = 0.0 ; /* just to avoid compiler warnings */

    /* refine current solution */
    for (iteration=1;
         iteration<=gla.max_iteration &&
           (fabs(tmp_err/curr_err-1)>gla.threshold || iteration==1);
         iteration++)
    {
        /* make new partition */
        make_new_partition(init_cb, fvs, partition, gla.dist_func,gla.dfunc_params) ;

        /* compute new centroids */
        if (compute_new_centroids(init_cb, fvs, partition)==ESR_FAILED)
        {
            ReportError(ESR_GLA_CENTROID_COMPUTATION, NULL) ;
            free(partition) ;
            return 0 ;        
        }

        /* evaluate clustering */
        tmp_err = curr_err ;

        if (gla.flags & GLA_USE_INDEXING)
        {
            /* create temporary codebook */

            /* curr_err = gla.idxerr_func(init_cb,fvs,gla.dist_func,NULL,NULL) ; */

        } else {

            curr_err = gla.err_func(init_cb,fvs,gla.dist_func,gla.dfunc_params) ;
        }
    }

    if (ret_err) *ret_err = curr_err ;

    free(partition) ;

    return 1 ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Compute new NN-based partition for a vector set

Compute new partition for a vector set by finding
the nearest vector in a set of centroids for all
input vectors. The distance computations are done
with a given distance function.
\param cb Input centroid vector set
\param fvs Input vector set that is partitioned
\param partition Output partition
\param dist_func Distance function used in computation
\return ESR_SUCCEED indicating success
or ESR_FAILED in case of failure
\todo Check all return values of subroutine calls.
\todo ADD INDEXING GENERATION FOR CODEBOOK IF NEEDED !!!
\todo Implementation is not optimal:
use fvsGetVectorBuff to get the vector pointer
table before loop instead of calling fvsGetVector in the loop
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
int make_new_partition(HFSET cb, HFSET fvs,
                       int * partition, DISTFUNC dist_func,void * dfunc_params) 
{
    int cb_size ; /* centroid vector set size */
    int set_size; /* input vector set size */
    int vector_size; /* length of vectors in the input vector set */

    int vector_num, centr_num, min_centr ;

    double min_dist, curr_dist ;
    double * curr_vect ;
    double * curr_centr ;

    if ((NULL == cb) || (NULL == fvs)
          || (NULL == partition) || (NULL == dist_func))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* copy all sizes to local variables
       to avoid unnecessary function calls */
    set_size = fvsGetSize(fvs) ;
    cb_size = fvsGetSize(cb) ;
    vector_size = fvsGetVectorSize(fvs) ;
    if ((ESR_FAILED == set_size) || (ESR_FAILED == cb_size)
          || (ESR_FAILED == vector_size)
          || (0 == set_size) || (0 == cb_size) || (0 == vector_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* ADD INDEXING GENERATION FOR CODEBOOK IF NEEDED !!! */

    min_dist = 0.0 ; min_centr = 0 ; /* just to avoid compiler warnings */

    for (vector_num=1;vector_num<=set_size;vector_num++)
    {
        curr_vect = fvsGetVector(fvs,vector_num) ;
        if (NULL == curr_vect)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return ESR_FAILED;
        }

        /* find nearest centroid */
        for (centr_num = 1; centr_num <= cb_size; centr_num++)
        {
            curr_centr = fvsGetVector(cb, centr_num);
            if (NULL == curr_centr)
            {
                ReportError(ESR_VECTOR_SET_GET, NULL);
                return ESR_FAILED;
            }

            /* calculate distance to current centroid */
            curr_dist = dist_func(curr_vect,
                                  curr_centr,vector_size,dfunc_params) ;

            /* check if it is less than minimum distance */
            if (min_dist>curr_dist || centr_num==1)
            {
                min_dist = curr_dist ;
                min_centr = centr_num ;
            }
        }

        /* set partition number */
        partition[vector_num-1] = min_centr ;
    }

    return ESR_SUCCEED ;
}


/*!
\brief Compute centroids of a vector set partition

Compute average of each subset of feature vectors
in the partition and set the codebook vectors
to these averages.
\param cb Output vector set of computed centroids
\param fvs Input vector set
\param partition Partition of the input vector set
\param dist_func ???
\return ESR_SUCCEED indicating success
or ESR_FAILED in case of failure
\pre The array partition must be a partition
of the feature vector set fvs, i.e. if
N is size of fvs and K the number of classes in
the partition, then for all i=1,...,N must be
\f$0 \leq \texttt{partition[i]} < K\f$.
\todo Check that partition really satisfies the pre-condition
\todo Implementation is not optimal:
<ul>
<li> divide with the number of vectors only after the main loop.
<li> use fvsGetVectorBuff to get the vector pointer
table before loop instead of calling fvsGetVector in the loop
</ul>
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
int compute_new_centroids(HFSET cb, HFSET fvs, int * partition) 
{
    int cb_size ; /* number of centroids in cb */
    int set_size; /* number of feature vectors in fvs */
    int vector_size; /* length of vectors in fvs */

    int vector_num, i ;

    int * part_sizes ;

    double * curr_vector, * curr_centr ;

    if ((NULL == cb) || (NULL == fvs)
          || (NULL == partition))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* copy all sizes in local variables
       to avoid unnecessary function calls */
    set_size = fvsGetSize(fvs) ;
    cb_size = fvsGetSize(cb) ;
    vector_size = fvsGetVectorSize(fvs) ;
    if ((ESR_FAILED == set_size) || (ESR_FAILED == cb_size)
          || (ESR_FAILED == vector_size)
          || (0 == set_size) || (0 == cb_size) || (0 == vector_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return ESR_FAILED;
    }

    /* allocate memory for set sizes in the partition */
    part_sizes = malloc(sizeof(int)*cb_size);
    if (part_sizes == NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY, NULL);
        return ESR_FAILED ;
    }

    memset(part_sizes, 0, sizeof(int)*cb_size) ;

    /* compute sizes of the sets in the partition */
    for (vector_num=1; vector_num<=set_size; vector_num++)
    {
        part_sizes[partition[vector_num-1]-1]++ ;
    }

    /* set centroids to zero */
    for (vector_num=1;vector_num<=cb_size;vector_num++)
    {
        memset(fvsGetVector(cb, vector_num),
               0, sizeof(double)*vector_size) ;
    }

    /* compute average vector of each set in the partition */
    for (vector_num=1; vector_num<=set_size; vector_num++)
    {
        curr_vector = fvsGetVector(fvs, vector_num) ;
        curr_centr = fvsGetVector(cb, partition[vector_num-1]) ;
        if ((NULL == curr_vector) || (NULL == curr_centr))
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            free(part_sizes);
            return ESR_FAILED;
        }

        for (i=0;i<vector_size;i++)
        {
            curr_centr[i] += curr_vector[i] / part_sizes[partition[vector_num-1]-1] ;
        }
    }
    /*
    for (vector_num=1; vector_num<=cb_size; vector_num++)
    {
        if (part_sizes[vector_num-1] == 0)
        {
            continue;
        }
        curr_centr = fvsGetVector(cb, vector_num) ;
        for (i=0;i<vector_size;i++)
        {
            curr_centr[i] = curr_centr[i] / part_sizes[vector_num-1] ;
        }
    }
    */

    free(part_sizes) ;

    return ESR_SUCCEED ;
}


/*======================== Functions ==============================*/
