/*! 
\file idxerrfunc.c
\brief Implementations of clustering error functions

Several error functions for clustering.
These functions work using indexing.
??? Be more specific: what is indexed? how? \n
$Id: idxerrfunc.c,v 1.14 2003/11/11 22:40:36 juhani Exp $
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>


#include <errorapi.h>
#include <msindex.h>


#include "idxerrfunc.h"


/*======================== Functions ==============================*/

/*!
\brief Compute Mean Squared Error with help of indexing

Compute the mean squared error (MSE) for a vector set
when NN-classified with given set of centroids, a codebook.
Use metric space index generated for the codebook.
If the index does not exist it will be computed first
with the functions msidxCreate and msidxTrain.
\param codebook Input codebook used classification
and error measurement
\param test_set Input vector set whose classification
error is computed
\param msidx Metric space indexer used in the computation
\param indexes Metric space indexing of codebook
\return Mean squared error of the codebook when
used for NN-classification of the given vector set.
\todo There is no need to divide with set size
in the main loop to avoid floating point overflow
\todo Implementation is not optimal:
use fvsGetVectorBuff to get the vector pointer
table before loop instead of calling fvsGetVector in the loop
or use FVS_GET_ELEMENT macro
\todo Check subroutine call results
\sa FVSet
\sa MSIndex
\sa DISTFUNC
\author Evgeny Karpov
*/
double idxMSE(
    HFSET codebook,
    HFSET test_set,
    HMSIDX msidx,
    HMSINDEX indexes) 
{
    int i, test_size ;
    double min_dist, ret_dist ;

    if (codebook==NULL || test_set==NULL || 
        msidx == NULL || indexes == NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    test_size = fvsGetSize(test_set) ;

    ret_dist = 0.0 ; 

    for (i=1;i<=test_size;i++)
    {
        /* find distance to NN vector using indexes */
        msidxNNSearch(msidx,indexes,codebook,fvsGetVector(test_set,i),&min_dist) ;

        /* divide by test_size to avoid floating number overflow */
        ret_dist += (min_dist*min_dist)/test_size ;
    }

    return ret_dist ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Compute Total Squared Error with help of indexing

Compute the total squared error (TSE) for a vector set
when NN-classified with given set of centroids, a codebook.
Use metric space index generated for the codebook.
If the index does not exist it will be computed first
with the functions msidxCreate and msidxTrain.
\param codebook Input codebook used classification
and error measurement
\param test_set Input vector set whose classification
error is computed
\param msidx Metric space indexer used in the computation
\param indexes Metric space indexing of codebook
\return Total squared error of the codebook when
used for NN-classification of the given vector set.
\todo This can be implemented by calling idxMSE
and multiplying the result with size of the test_set.
In fact the code should be in here and idxMSE
should call this to save unnecessary division
in the code.
\sa FVSet
\sa MSIndex
\sa DISTFUNC
\author Evgeny Karpov
*/
double idxTSE(
    HFSET codebook,
    HFSET test_set,
    HMSIDX msidx,
    HMSINDEX indexes) 
{
    ReportError(ESR_NOT_IMPLEMENTED,
                "idxTSE (Total Squared Error with indexing");
    exit(1);
    return 0.0 ;
}

/*======================== Functions ==============================*/
