/*! 
\file pq.c
\brief Pre-quantization implementation

This module consists of several vector set reduction
algorithms for producing from a given vector set a new vector set
with less vectors within the set. The generated
vector set has fever vectors but it is hoped that
the reduction keeps set characteristics of certain kind:
we hope that there would be a representative close to all
original vectors in the new reduced set.
This means that for all original vectors x there is
a vector in the new set near to x.\n
$Id: pq.c,v 1.11 2003/11/13 19:25:24 juhani Exp $
\note No vector reduction algorithm can guarantee
a good modeling of the original set but the quality
of the modeling can be measured, for example, with
help of NN-classification and MSE error measure.
\todo Add ability to set parameters
for selecting algorithm and parameters
to use for codebook pre-quantization
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <string.h>
#include <time.h>


#include <errorapi.h>


#include "gla.h"
#include "vq.h"


#include "pq.h"


/*-----------------------------------------------------------------*/


/*======================== Structures =============================*/


/*======================== Structures =============================*/


HFSET average_clustering(HFSET test_set, int clust_num, void * params) ;
HFSET random_clustering(HFSET test_set, int clust_num, void * params) ;
HFSET subsampl_clustering(HFSET test_set, int clust_num, void * params) ;
HFSET codebook_clustering(HFSET test_set, int clust_num, void * params) ;


/*======================== Functions ==============================*/


/*!
\brief Apply pre-quantization algorithm to a vector set

Use the algorithm and its parameters stored
in a pre-quantizer to quantize input data
\param pq Pre-quantization parameters
\param test_set Input set of vectors that is quantized
\return Vector set after quantization
\author Evgeny Karpov
*/
HFSET pqQuantize(PQParams pq, HFSET test_set) 
{
    if (pq.algorithm == PQTYPE_RANDOM)
    {
        return random_clustering(
                 test_set, pq.output_size_spec.interval_size, NULL) ;
    }
    else if (pq.algorithm == PQTYPE_AVERAGE)
    {
        return average_clustering(
                 test_set, pq.output_size_spec.interval_size, NULL) ;
    }
    else if (pq.algorithm == PQTYPE_SUBSAMPL)
    {
        return subsampl_clustering(
                 test_set, pq.output_size_spec.interval_size, NULL) ;
    }
    else if (pq.algorithm == PQTYPE_CODEBOOK)
    {
        return codebook_clustering(
                 test_set,
                 pq.output_size_spec.number_of_clusters, NULL) ;
    }

    ReportError(ESR_INVALID_PARAMETER, NULL) ;
    return NULL ;
}


/*-----------------------------------------------------------------*/


/*!
\brief Reduce vector set size by using average of moving block

Divide vector set indices in consecutive blocks
and compute new vectors as the averages of the blocks.
\param test_set Vector set that needs to be reduced
\param clust_num Desired number of vectors in
the reduced set represented as integer percentage
\param params Not used by this quantization method
\return Pointer to the reduced vector set, or NULL
in case of failure
\pre 0 < clust_num < 100, and clust_num times test_set size
divided by 100 should be positive
\bug If size of the input vector set is less than desired
number of model vectors this should return with error
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET average_clustering(HFSET test_set, int clust_num, void * params)
{
    HFSET ret_set ;
    double * pvector, * ptmp ;
    int vect_size, test_size, ret_size ;
    int used, avg_size ;
    int i,j ;

    int vect_num ;

    if (test_set==NULL || clust_num==0)
    {
        return NULL ;
    }

    test_size = fvsGetSize(test_set) ;
    vect_size = fvsGetVectorSize(test_set) ;

    if (test_size==0)
    {
        return NULL ;
    }

    /* convert percent to size */
    /* vect_num = test_size - (test_size*(100-clust_num))/100 ; */
    vect_num = (test_size * clust_num) / 100;

    /* create set */
    if ((ret_set = fvsCreate(vect_num,vect_size))==NULL)
    {
        return NULL ;
    }

    /* if test size is less than requested amount of vectors */
    if (test_size <= vect_num)
    {
        /* just copy vectors */
        for (i=1;i<=vect_num;i++)
        {
            fvsAddVector(ret_set, fvsCopyVector(test_set,i)) ;

            if (i==test_size)
            {
                i=1 ; 
            }
        }

        return ret_set ;
    }

    used = 0 ;
    ret_size = 0 ;
    avg_size = test_size / vect_num;

    while (used < test_size)
    {
        /* create new vector */
        if ((pvector = fvCreate(vect_size))==NULL) 
        {
            fvsDestroy(ret_set) ;
            return NULL ; /* error was set already */
        }

        /* initialize to zero */
        memset(pvector,0,sizeof(double)*vect_size) ;

        /* if there are less vectors left than needed */
        if (test_size - used < avg_size)
        {
            avg_size = test_size - used ;
        }

        /* average "avg_size" vectors */
        for (j = 1; j <= avg_size; j++)
        {
            ptmp = fvsGetVector(test_set, used+j) ;

            /* add vector to average sum */
            for (i=0;i<vect_size;i++)
            {
                pvector[i] += ptmp[i] / avg_size ;
            }
        }

        used += avg_size ;

        fvsAddVector(ret_set,pvector) ;
        ret_size ++ ;

        /* adapt average size */
        if (((test_size - used) / (avg_size+1))
              >= (vect_num - ret_size))
        {
            avg_size++ ;
        }

    }

    return ret_set ;
}


/*!
\brief Reduce vector set size by using random picking in moving block

Create a subset of the input vector set by picking
wanted number of vectors from original set. Vecotrs are
selected by dividing vector indices in approximately
equally sized continuous subsets (1-k, k+1-2*k, ..., n*k+1-N)
and picking one vector from each subset.
\param test_set Vector set that needs to be reduced
\param clust_num Desired number of vectors in
the reduced set represented as integer percentage
\param params Not used by this quantization method
\return Pointer to the reduced vector set, or NULL
in case of failure
\pre 0 < clust_num < 100, and clust_num times test_set size
divided by 100 should be positive
\bug If size of the input vector set is less than desired
number of model vectors this should return with error
\todo Why intervals are used, why not just pick
test_set size * clust_num / 100 random vectors???
Is the vector set index analogous with time axis when
this function is called? If it is then this kind
of pick-from-frame approach may be sensible and
the argumentation should be made in the documentation.
Completely random approach can be implemented as:
<ul>
<li> decide/compute how many random vectors are needed,
<li> compute that many unique indices to test set,
<li> pick the selected vectors
</ul>
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET random_clustering(HFSET test_set, int clust_num, void * params)
{
    HFSET ret_set ;
    int vect_size,test_size,ret_size ;
    int used, avg_size ;
    int rand_vect ;
    int i ;

    int vect_num ;

    if (test_set==NULL || clust_num==0)
    {
        return NULL ;
    }

    test_size = fvsGetSize(test_set) ;
    vect_size = fvsGetVectorSize(test_set) ;

    /* convert percent to size */
    /* vect_num = test_size - (test_size*(100-clust_num))/100 ; */
    vect_num = (test_size * clust_num) / 100;

    if (test_size==0)
    {
        return NULL ;
    }

    /* create set */
    ret_set = fvsCreate(vect_num,vect_size);
    if (ret_set == NULL)
    {
        return NULL ;
    }

    /* if test size is less than requested amount of vectors */
    if (test_size <= vect_num)
    {
        /* just copy vectors */
        for (i=1;i<=vect_num;i++)
        {
            fvsAddVector(ret_set,fvsCopyVector(test_set,i)) ;

            if (i==test_size)
            {
                i=1 ;
            }
        }

        return ret_set ;
    }

    used = 0 ; ret_size = 0 ;
    avg_size = test_size / vect_num;

    while (used < test_size)
    {
        /* if there are less vectors left than needed */
        if (test_size - used < avg_size)
        {
            avg_size = test_size - used ;
        }

        /* seed the random-number generator with current time */
        srand((unsigned)time(NULL));

        /* take random vector from "avg_size" vectors */
        rand_vect = rand() % avg_size + 1 ;

        /* copy random vector */
        fvsAddVector(ret_set,
                     fvsCopyVector(test_set, used + rand_vect));
        
        used += avg_size ;

        ret_size ++ ;

        /* adapt average size */
        if (((test_size - used) / (avg_size+1))
              >= (vect_num - ret_size))
        {
            avg_size++ ;
        }
    }

    return ret_set ;
}


/*!
\brief Reduce vector set size by using equally spaced subsampling

Create a subset of the input vector set by picking
wanted number of vectors from original set with equally
spaced indices.
\param test_set Vector set that needs to be reduced
\param clust_num Desired number of vectors in
the reduced set represented as integer percentage
\param params Not used by this quantization method
\return Pointer to the reduced vector set, or NULL
in case of failure
\pre 0 < clust_num < 100, and
clust_num times test_set size divided by 100
should be positive
\bug If size of the input vector set is less than desired
number of model vectors this should return with error
\bug The size of output set should be 
<tt>(test_size * clust_num) / 100</tt>,
not <tt>test_size / clust_num</tt>
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET subsampl_clustering(HFSET test_set, int clust_num, void * params)
{
    HFSET ret_set ;
    int vect_size,test_size ;
    int used ;

    if (test_set == NULL) return NULL ;

    test_size = fvsGetSize(test_set) ;
    vect_size = fvsGetVectorSize(test_set) ;

    /* create set */
    ret_set = fvsCreate(test_size/clust_num, vect_size);
    if (ret_set == NULL)
    {
        return NULL ;
    }

    for (used = 1; used <= test_size; used += clust_num)
    {
        fvsAddVector(ret_set, fvsCopyVector(test_set,used)) ;
    }

    return ret_set ;
}


/*!
\brief Reduce vector set size by using Generalized Lloyd Algorithm

Create a codebook with given number of clusters,
train it by using the generalized Lloyd algorithm (GLA),
and return the codebook as the reduced data set.
\param test_set Vector set that needs to be reduced
\param number_of_clusters Desired number of vectors in
the reduced set represented as integer percentage
\param params Not used by this quantization method
\return Pointer to the reduced vector set, or NULL
in case of failure
\pre 0 < clust_num < 100, and clust_num times test_set size
divided by 100 should be positive
\bug size of vector set compared to input percentage is not
checked. Moreover, clust_num is not interpreted as
percentage here but given directly to vqCreate as
number of codebook vectors!!!
\bug If size of the input vector set is less than desired
number of model vectors this should return with error
\sa VQ_ALG_GLA
\sa HFSET (a pointer to FVSet)
\author Evgeny Karpov
*/
HFSET codebook_clustering(
    HFSET test_set,
    int number_of_clusters,
    void * params)
{
    HFSET cb ;

    HVQ vq ;
    HVQMODEL vqmodel ;
    VQParams vq_params ;

    vq_params.algorithm = VQ_ALG_GLA;
    vq_params.algorithm_params = NULL;
    vq_params.dist_func = dfEuclidean;
    vq_params.index_type = VQ_IDX_NONE;
    vq_params.err_func = errMSE;
    vq_params.idxerr_func = idxMSE;
    vq_params.flags = 0;
    vq_params.gla_num_initial = 1;

    /*
    vq_params.num_clusters = number_of_clusters ;
    vq_params.gla_num_initial = 1 ;
    vq_params.flags = VQ_GLA_NUM_INITIAL ;
    */

    /* create codebook */
    vq = vqCreate(&vq_params) ;
    if (NULL == vq)
    {
        ReportError(ESR_VECTOR_QUANTIZER_CREATE, NULL);
        return NULL;
    }

    /* adjust codebook to data */
    vqmodel = vqTrain(vq, test_set, number_of_clusters) ;
    if (NULL == vqmodel)
    {
        ReportError(ESR_VECTOR_QUANTIZER_TRAIN, NULL);
        return NULL;
    }

    /* grab codebook vector set */
    cb = vqmodel->codebook;
    vqmodel->codebook = NULL;

    vqDestroyModel(vq, vqmodel) ;

    return cb ;

    /* return NULL ; */
}


/*======================== Functions ==============================*/

