/*! 
\file smetric.c
\brief Set metrics implementation

Implementation of computing distances
of pair of sets using several different
metrics.\n
$Id: smetric.c,v 1.13 2003/11/20 11:44:34 ekarpov Exp $
\date 2003
\author Evgeny Karpov
*/


#include <stdlib.h>
#include <float.h>


#include <errorapi.h>


#include "smetric.h"

/*-----------------------------------------------------------------*/

#ifndef max
/*! \brief Resolve maximum of two items */
#define max(a,b)    (((a) > (b)) ? (a) : (b))
/*! \brief Resolve minimum of two items */
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

/*======================== Functions ==============================*/

/*!
\brief Compute Hausdorff set distance

Compute Hausdorff set distance for two sets
using a given distance function to measure distances of points.
The Hausdorff distance of sets \f$A,B\subset X\f$ is
\f[
d_{\mathrm{H}}(A,B) = \max(h(A,B), h(B,A)),
\f]
where
\f[
h(A,B) = \max_{a\in A}\min_{b\in B}\mathop\mathrm{d}(a,b).
\f]
Here \f$\mathrm{d}\,:\,X\times X\to \mathbb{R}\f$
is metric defined on \f$X\f$.
Note that \f$\mathrm{d}_{\mathrm{H}}\f$ is a metric
defined on nonempty subsets of \f$X\f$.
\param left_set First set
\param right_set Second set
\param dist_func Distance function used
for computing distances of points (\f$d\f$ above)
or NULL for using the default euclidean distance
\return Hausdorff set distance of the two sets,
negative value is returned in case of failure
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
double smHausdorff(
    HFSET left_set,
    HFSET right_set,
    DISTFUNC dist_func, 
    void * dfunc_params) 
{
    double left_dist,right_dist,tmp_dist,min_dist ;
    int left_size,right_size ;
    int i,j,vector_size ;
    double *left_vec;
    double *right_vec;

    if (left_set==NULL || right_set==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    if (dist_func==NULL)
    {
        dist_func = dfEuclidean ;
    }

    left_size = fvsGetVectorSize(left_set) ;
    right_size = fvsGetVectorSize(right_set) ;
    if ((ESR_FAILED == left_size) || (ESR_FAILED == right_size)
          || (left_size != right_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }
    vector_size = left_size;

    left_size = fvsGetSize(left_set) ;
    right_size = fvsGetSize(right_set) ;
    if ((left_size <= 0) || (right_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    /* find dist from left set to right */
    for (i = 1, left_dist = 0.0; i <= left_size; i++)
    {
        left_vec = fvsGetVector(left_set,i);
        if (NULL == left_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }
        min_dist = DBL_MAX ;

        for (j=1;j<=right_size;j++)
        {
            right_vec = fvsGetVector(right_set,j);
            if (NULL == right_vec)
            {
                ReportError(ESR_VECTOR_SET_GET, NULL);
                return -1.0;
            }

            tmp_dist = dist_func(left_vec, right_vec, vector_size,dfunc_params) ;
            if (tmp_dist < 0.0)
            {
                ReportError(ESR_DISTANCE_FUNCTION, NULL);
                return -1.0;
            }

            min_dist = min(tmp_dist,min_dist) ;
        }
        
        left_dist = max(min_dist,left_dist) ;
    }

    /* find dist from right set to left */
    for (i = 1, right_dist = 0.0; i <= right_size; i++)
    {
        right_vec = fvsGetVector(right_set,i);
        if (NULL == right_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }
        min_dist = DBL_MAX ;

        for (j=1;j<=left_size;j++)
        {
            left_vec = fvsGetVector(left_set,j);
            if (NULL == left_vec)
            {
                ReportError(ESR_VECTOR_SET_GET, NULL);
                return -1.0;
            }

            tmp_dist = dist_func(right_vec, left_vec, vector_size, dfunc_params) ;
            if (tmp_dist < 0.0)
            {
                ReportError(ESR_DISTANCE_FUNCTION, NULL);
                return -1.0;
            }

            min_dist = min(tmp_dist,min_dist) ;
        }
        
        right_dist = max(min_dist,right_dist) ;
    }

    return max(left_dist, right_dist) ;
}

/*!
\brief Compute Sum of Minimum distances set distance

Compute Sum of Minimum set distance for two sets using
a given distance function to measure distance of points.
The <i>Sum of minimum distances</i> set distance function is
\f[
\mathop\mathrm{d}_{\mathrm{md}}(A,B)
 = \frac{1}{2}\left(
 \sum_{a\in A} \Delta_m(a,B) + \sum_{b\in B} \Delta_m(b,A)
 \right),
\f]
where \f$\Delta_m(a,B) = \min_{b\in B} \mathop\mathrm{d}(a,b)\f$.
Here \f$\mathrm{d}\,:\,X\times X\to \mathbb{R}\f$
is metric defined on \f$X\f$.
Note that \f$\mathrm{d}_{\mathrm{md}}\f$ is a set distance
function defined on nonempty subsets of \f$X\f$
but not a metric.
\param left_set First set
\param right_set Second set
\param dist_func Distance function used
for computing distances of points (\f$d\f$ above)
or NULL for using the default euclidean distance
\return Sum of Minimum set distance of the two sets,
negative value is returned in case of failure
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
double smSumOfMin(
    HFSET left_set,
    HFSET right_set,
    DISTFUNC dist_func, 
    void * dfunc_params) 
{
    double left_dist,right_dist,tmp_dist,min_dist ;
    int left_size,right_size ;
    int i,j,vector_size ;
    double *left_vec;
    double *right_vec;

    if (left_set==NULL || right_set==NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    if (dist_func==NULL)
    {
        dist_func = dfEuclidean ;
    }

    left_size = fvsGetVectorSize(left_set) ;
    right_size = fvsGetVectorSize(right_set) ;
    if ((ESR_FAILED == left_size) || (ESR_FAILED == right_size)
          || (left_size != right_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }
    vector_size = left_size;

    left_size = fvsGetSize(left_set) ;
    right_size = fvsGetSize(right_set) ;
    if ((left_size <= 0) || (right_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    /* find sum of min distances from left set to right */
    for (i = 1, left_dist = 0.0; i <= left_size; i++)
    {
        left_vec = fvsGetVector(left_set,i);
        if (NULL == left_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }
        min_dist = DBL_MAX ;

        for (j = 1; j <= right_size; j++)
        {
            right_vec = fvsGetVector(right_set,j);
            if (NULL == right_vec)
            {
                ReportError(ESR_VECTOR_SET_GET, NULL);
                return -1.0;
            }

            tmp_dist = dist_func(left_vec, right_vec, vector_size, dfunc_params) ;
            if (tmp_dist < 0.0)
            {
                ReportError(ESR_DISTANCE_FUNCTION, NULL);
                return -1.0;
            }

            min_dist = min(tmp_dist,min_dist) ;
        }

        left_dist += min_dist ;
    }

    /* find dist from right set to left */
    for (i = 1, right_dist = 0.0; i <= right_size; i++)
    {
        right_vec = fvsGetVector(right_set,i);
        if (NULL == right_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }
        min_dist = DBL_MAX ;

        for (j=1;j<=left_size;j++)
        {
            left_vec = fvsGetVector(left_set,j);
            if (NULL == left_vec)
            {
                ReportError(ESR_VECTOR_SET_GET, NULL);
                return -1.0;
            }

            tmp_dist = dist_func(right_vec, left_vec, vector_size, dfunc_params) ;
            if (tmp_dist < 0.0)
            {
                ReportError(ESR_DISTANCE_FUNCTION, NULL);
                return -1.0;
            }

            min_dist = min(tmp_dist,min_dist) ;
        }

        right_dist += min_dist ;

    }

    return (left_dist + right_dist)/2 ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Compute Hausdorff set distance with help of indexing

Compute Hausdorff set distance for two sets
using a given distance function to measure distances of points.
Use the input indexes defined for the point sets
to speed up computation.
\param left_set First set
\param right_set Second set
\param indexer Metric space indexer used
for computing distances of points
\param left_idx Indexing of left_set
\param right_idx Indexing of right_set
\return Hausdorff set distance of the two sets,
negative value is returned in case of failure
\sa smHausdorff for definition
of the Hausdorff set distance.
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\sa HMSIDX (a pointer to MSIndex)
\author Evgeny Karpov
*/
double smHausdorffIdx(
    HFSET left_set,
    HFSET right_set,
    HMSIDX indexer,
    HMSINDEX left_idx,
    HMSINDEX right_idx) 
{
    double left_dist,right_dist,min_dist ;
    int left_size,right_size ;
    int i ;
    double *left_vec;
    double *right_vec;
    RetStatus result;

    if (left_set==NULL || right_set==NULL ||
        left_idx == NULL || right_idx == NULL || indexer == NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    left_size = fvsGetVectorSize(left_set) ;
    right_size = fvsGetVectorSize(right_set) ;
    if ((ESR_FAILED == left_size) || (ESR_FAILED == right_size)
          || (left_size != right_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    left_size = fvsGetSize(left_set) ;
    right_size = fvsGetSize(right_set) ;
    if ((left_size <= 0) || (right_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    /* find dist from left set to right */
    for (i = 1, left_dist = 0.0; i <= left_size; i++)
    {
        left_vec = fvsGetVector(left_set,i);
        if (NULL == left_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }

        result = msidxNNSearch(
                     indexer, right_idx, right_set,
                     left_vec, &min_dist) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_MSINDEX_NN_SEARCH_FUNCTION, NULL);
            return -1.0;
        }

        left_dist = max(min_dist,left_dist) ;
    }

    /* find dist from right set to left */
    for (i = 1, right_dist = 0.0; i <= right_size; i++)
    {
        right_vec = fvsGetVector(right_set,i);
        if (NULL == right_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }

        result = msidxNNSearch(
                     indexer, left_idx, left_set,
                     right_vec, &min_dist) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_MSINDEX_NN_SEARCH_FUNCTION, NULL);
            return -1.0;
        }

        right_dist = max(min_dist,right_dist) ;
    }

    return max(left_dist,right_dist) ;
}

/*!
\brief Compute Sum of Minimum distances set distance
with help of indexing

Compute Sum of Minimum distances set distance for two sets using
a given distance function to measure distance of points.
Use the input indexes defined for the point sets
to speed up computation.
\param left_set First set
\param right_set Second set
\param indexer Metric space indexer used
for computing distances of points
\param left_idx Indexing of left_set
\param right_idx Indexing of right_set
\return Sum of Minimum distances set distance of the two sets,
negative value is returned in case of failure
\sa smSumOfMin for definition
of the sum of minimum distance set distance.
\sa HFSET (a pointer to FVSet)
\sa DISTFUNC
\author Evgeny Karpov
*/
double smSumOfMinIdx(
    HFSET left_set,
    HFSET right_set,
    HMSIDX indexer,
    HMSINDEX left_idx,
    HMSINDEX right_idx) 
{
    double left_dist,right_dist,min_dist ;
    int left_size,right_size ;
    int i ;
    double *left_vec;
    double *right_vec;
    RetStatus result;

    if (left_set==NULL || right_set==NULL ||
        left_idx == NULL || right_idx == NULL || indexer == NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    left_size = fvsGetVectorSize(left_set) ;
    right_size = fvsGetVectorSize(right_set) ;
    if ((ESR_FAILED == left_size) || (ESR_FAILED == right_size)
          || (left_size != right_size))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    left_size = fvsGetSize(left_set) ;
    right_size = fvsGetSize(right_set) ;
    if ((left_size <= 0) || (right_size <= 0))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return -1.0;
    }

    /* find dist from left set to right */
    for (i = 1, left_dist = 0.0; i <= left_size; i++)
    {
        left_vec = fvsGetVector(left_set,i);
        if (NULL == left_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }

        result = msidxNNSearch(
                     indexer, right_idx, right_set,
                     left_vec, &min_dist) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_MSINDEX_NN_SEARCH_FUNCTION, NULL);
            return -1.0;
        }
        
        left_dist += min_dist ;
    }

    /* find dist from right set to left */
    for (i = 1, right_dist = 0.0; i <= right_size; i++)
    {
        right_vec = fvsGetVector(right_set,i);
        if (NULL == right_vec)
        {
            ReportError(ESR_VECTOR_SET_GET, NULL);
            return -1.0;
        }

        result = msidxNNSearch(
                     indexer, left_idx, left_set,
                     right_vec, &min_dist) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_MSINDEX_NN_SEARCH_FUNCTION, NULL);
            return -1.0;
        }

        right_dist += min_dist ;
    }

    return (left_dist + right_dist)/2 ; 
}

/*======================== Functions ==============================*/

