/*! 
\file spruning.c
\brief Speaker pruning implementation

Implementation of the framework for speaker
pruning  algorithms.\n
$Id: spruning.c,v 1.4 2003/11/26 08:01:27 ekarpov Exp $
\todo Change all error handling to use errorapi
\date 2003
\author Evgeny Karpov
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "spruning.h"
#include "vq.h"

/*-----------------------------------------------------------------*/

#define SPR_MODEL_PRUNED        1
#define SPR_MODEL_STABILIZED    2

/*======================== Functions ==============================*/

HSPR sprCreate() 
{
    return NULL ;
}

void sprDestroy(HSPR spr) 
{
}

/*-----------------------------------------------------------------*/

int sprReordPruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRReordParam * alg_params) 
{
    HFSET tmp_set ;
    SPRReordParam params ;

    int vector, set_size, vector_left ;
    int iteration ;
    int start, stop ;
    int model_num, model, pruned, model_left ;

    double ** pvect_buff ;
    double * scores ;

    int curr_model ;
    double curr_score ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        params.init_interval = 4 ;
        params.max_iteration = 3 ;
        params.prune_num     = msGetSize(ms) / 3 ; /* 1/3 of models */

    } else memcpy(&params,alg_params,sizeof(SPRReordParam)) ;

    set_size = fvsGetSize(test_set) ;
    model_num = msGetSize(ms) ;

    vector_left = set_size ;
    model_left = model_num ;

    /* create temporary set */
    tmp_set = fvsCreate(set_size,fvsGetVectorSize(test_set)) ;

    /* allocate temporary buffer for vectors */
    pvect_buff = malloc(sizeof(double*)*set_size) ;

    /* allocate buffer for scores */
    scores = malloc(sizeof(double)*model_num) ;

    if (tmp_set==NULL || pvect_buff==NULL || scores==NULL)
    {
        errno = ENOMEM ;
        free(pvect_buff) ;
        free(scores) ;
        fvsDestroy(tmp_set) ;
        return 0 ;
    }

    /* copy vector references to temporary buffer */
    for (vector=1; vector <= set_size; vector ++)
        pvect_buff[vector-1] = fvsGetVector(test_set,vector) ;

    /* intialize model flags */
    for (model=1; model<=model_num; model++)
        msSetModelFlags(ms,model,0) ;

    /* intialize scores */
    memset(scores,0,sizeof(double)*model_num) ;
    
    /* start algorithm iterations */
    for(iteration = 1 ;iteration <= params.max_iteration && model_left > 1 && vector_left > 0; iteration++)
    {
        /* find current iteration test set */
        if (iteration == 1)
        {
            for (vector=1; vector <= set_size; vector += params.init_interval)
            {
                fvsAddVector(tmp_set,pvect_buff[vector-1]) ;
                pvect_buff[vector-1] = NULL ; /* mark vector as used */

                vector_left-- ;
            }

        } else {

            /* find middle points for every interval */
            stop = 0 ; start = 0 ;
            while (start < set_size && stop < set_size && vector_left > 0)
            {
                start = stop ; stop = start + 1 ;

                /* find null vector (end of interval) */                
                while (pvect_buff[stop] != NULL && stop < set_size) stop ++ ;

                /* find interval middle */
                if (stop - start <= 1) continue ; /* interval is empty */

                /* copy vector from the middle of interval */
                vector = start + (stop - start) / 2 ;

                fvsAddVector(tmp_set,pvect_buff[vector]) ;
                pvect_buff[vector] = NULL ; /* mark vector as used */

                vector_left-- ;
            }
        }

        if (fvsGetSize(tmp_set) == 0) continue ; /* set is empty */

        /* macth set against models */
        for (model=1; model <= model_num; model++)
        {
            if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
                scores[model-1] += vqMatch(vq,msGetModel(ms,model),tmp_set) ;
        }

        /* prune models */
        for (pruned=0; pruned < params.prune_num && model_left > 1; pruned++,model_left--)
        {
            /* find first not pruned model */
            for (model=1; model <= model_num && 
                    msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

            curr_score = scores[model-1] ;
            curr_model = model ;

            /* find not pruned model with maximum score */
            for (model = curr_model + 1;model <= model_num; model++)
            {
                if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
                {
                    if (scores[model-1] > curr_score)
                    {
                        curr_score = scores[model-1] ;
                        curr_model = model ;
                    }
                }
            }

            /* mark model as pruned */
            msSetModelFlags(ms,curr_model,SPR_MODEL_PRUNED) ;
        }

        /* keep test vectors untouched */
        fvsSetSize(tmp_set,0) ;
    }

    /* use remained vectors */
    if (vector_left!=0 && model_left>1)
    {
        for (vector=1; vector <= set_size; vector++)
        {
            if (pvect_buff[vector-1] != NULL)
                fvsAddVector(tmp_set,pvect_buff[vector-1]) ;
        }

        /* macth set against models */
        for (model=1; model <= model_num; model++)
        {
            if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
                scores[model-1] += vqMatch(vq,msGetModel(ms,model),tmp_set) ;
        }

    }

    /* keep test vectors untouched */
    fvsSetSize(tmp_set,0) ;

    /* find first not pruned model */
    for (model=1; model <= model_num && 
        msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

    curr_score = scores[model-1] ;
    curr_model = model ;

    /* find not pruned model with minimum score */
    for (model = curr_model + 1;model <= model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
        {
            if (scores[model-1] < curr_score)
            {
                curr_score = scores[model-1] ;
                curr_model = model ;
            }
        }
    }

    /* find correct match score for the best model */
    if (ret_score) *ret_score = vqMatch(vq,msGetModel(ms,curr_model),test_set) ;

    /* free allocated resources */
    free(pvect_buff) ;
    free(scores) ;
    fvsDestroy(tmp_set) ;

    return curr_model ;
}

int sprPqBasedPruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRPqParams * alg_params) 
{
    HFSET tmp_set ;
    SPRPqParams params ;

    int model, model_num, tmp, pruned ;

    int curr_model ;
    double curr_score ;

    double * scores ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        tmp = 16 ; /* codebook size for pre-quantization */
/*        params.pq = pqCreate(PQTYPE_CODEBOOK,&tmp) ; /* create default algorihm */
        params.prune_num     = 90 * msGetSize(ms) / 100 ; /* 90 percent of models */

    } else memcpy(&params,alg_params,sizeof(SPRPqParams)) ;

    model_num = msGetSize(ms) ;

    /* allocate buffer for scores */
    scores = malloc(sizeof(double)*model_num) ;

/*    if (scores == NULL || params.pq==NULL)
    {
        if (scores==NULL) errno = ENOMEM ;
        if (alg_params==NULL) pqDestroy(params.pq) ;
        free(scores) ;
        return 0 ;
    } */

    /* intialize model flags */
    for (model=1; model<=model_num; model++)
        msSetModelFlags(ms,model,0) ;

    /* reduce number of test data by pre-quantization */
    tmp_set = pqQuantize(params.pq,test_set) ;

    /* macth models against reduced test */
    for (model=1; model<=model_num; model++)
        scores[model-1] = vqMatch(vq,msGetModel(ms,model),tmp_set) ;

    /* prune models */
    for (pruned=0; pruned < params.prune_num && model_num - pruned > 1; pruned++)
    {
        /* find first not pruned model */
        for (model=1; model <= model_num && 
            msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

        curr_score = scores[model-1] ;
        curr_model = model ;

        /* find not pruned model with maximum score */
        for (model = curr_model + 1;model <= model_num; model++)
        {
            if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
            {
                if (scores[model-1] > curr_score)
                {
                    curr_score = scores[model-1] ;
                    curr_model = model ;
                }
            }
        }

        /* mark model as pruned */
        msSetModelFlags(ms,curr_model,SPR_MODEL_PRUNED) ;
    }

    /* destroy temporary set */
    fvsDestroy(tmp_set) ;

    /* full search through all not pruned models */
    for (model=1; model<=model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
            scores[model-1] = vqMatch(vq,msGetModel(ms,model),test_set) ;
    }

    /* find first not pruned model */
    for (model=1; model <= model_num && 
        msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

    curr_score = scores[model-1] ;
    curr_model = model ;

    /* find not pruned model with minimum score */
    for (model = curr_model + 1;model <= model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
        {
            if (scores[model-1] < curr_score)
            {
                curr_score = scores[model-1] ;
                curr_model = model ;
            }
        }
    }

    /* copy scores */
    if (ret_score) *ret_score = curr_score ;

    /* free allocated resources */
/*    if (alg_params==NULL) pqDestroy(params.pq) ; */
    free(scores) ;

    return curr_model ;
}

int sprHierchPruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRHrchParams * alg_params) 
{
    SPRHrchParams params ;

    int model, model_num, pruned ;

    int curr_model;
    double curr_score ;

    double * scores ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    model_num = msGetSize(ms) ;

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        params.cb_size = 8 ;
        params.prune_num = 90 * msGetSize(ms) / 100 ; /* 90 percent of models */
        params.ms = NULL ;

    } else memcpy(&params,alg_params,sizeof(SPRHrchParams)) ;

    /* allocate buffer for scores */
    if ((scores = malloc(sizeof(double)*model_num))==NULL)
    {
        errno = ENOMEM ;
        return 0 ;
    }

    /* intialize model flags */
    for (model=1; model<=model_num; model++)
        msSetModelFlags(ms,model,0) ;

    /* match small models */
    if (params.ms)
    {
        for (model=1; model<=model_num; model++)
            /* macth small model */
            scores[model-1] = vqMatch(vq,msGetModel(params.ms,model),test_set) ;

    }

    /* prune models */
    for (pruned=0; pruned < params.prune_num && model_num - pruned > 1; pruned++)
    {
        /* find first not pruned model */
        for (model=1; model <= model_num && 
            msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

        curr_score = scores[model-1] ;
        curr_model = model ;

        /* find not pruned model with maximum score */
        for (model = curr_model + 1;model <= model_num; model++)
        {
            if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
            {
                if (scores[model-1] > curr_score)
                {
                    curr_score = scores[model-1] ;
                    curr_model = model ;
                }
            }
        }

        /* mark model as pruned */
        msSetModelFlags(ms,curr_model,SPR_MODEL_PRUNED) ;
    }

    /* match not pruned models against test set */
    for (model=1; model<=model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
            scores[model-1] = vqMatch(vq,msGetModel(ms,model),test_set) ;
    }

    /* find first not pruned model */
    for (model=1; model <= model_num && 
        msGetModelFlags(ms,model) == SPR_MODEL_PRUNED; model++) ;

    curr_score = scores[model-1] ;
    curr_model = model ;

    /* find not pruned model with minimum score */
    for (model = curr_model + 1;model <= model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
        {
            if (scores[model-1] < curr_score)
            {
                curr_score = scores[model-1] ;
                curr_model = model ;
            }
        }
    }

    /* copy scores */
    if (ret_score) *ret_score = curr_score ;

    /* free allocated resources */
    free(scores) ;

    return curr_model ;
}

int sprConfBasedPruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRCfParams * alg_params) 
{
    SPRCfParams params ;

    double score_min,score_max,pruning_thr ;

    double * scores, * distances ;
    double ** tmp ;

    int model, model_num, models_pruned,min_model ;
    int vector_num,curr_vector, curr_dist, count ;
    int slope_center ;

    double old_score ;
    double mse_mean, mean , std_dev, variance ;
    double slope, slope_avg ;

    HFSET past_dist ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        params.vector_step  = 20 ;
        params.stab_thr_type = SPR_STBTHR_SIMPLE ;
        params.pr_thr_type = SPR_CFBTHR_MINMAX ;        
        params.variance_thr = 0.01 ;
        params.pruning_thr  = 0.05 ;

    } else memcpy(&params,alg_params,sizeof(SPRCfParams)) ;

    model_num = msGetSize(ms) ;

    /* allocate buffer for scores */
    if ((scores = malloc(sizeof(double)*model_num))==NULL)
    {
        errno = ENOMEM ;
        return 0 ;
    }

    /* set all scores to zero */
    memset(scores,0,sizeof(double)*model_num) ;

    /* intialize model flags */
    if (params.stab_thr_type != SPR_STBTHR_NONE)
    {
        for (model=1; model<=model_num; model++)
            msSetModelFlags(ms,model,0) ;

    } else {

        for (model=1; model<=model_num; model++)
            msSetModelFlags(ms,model,SPR_MODEL_STABILIZED) ;
    }

    vector_num = fvsGetSize(test_set) ;

    past_dist = NULL ; curr_dist = 0 ;
    curr_vector = 0 ; models_pruned = 0 ; min_model = 0 ;

    /* just to make gcc compiler happy */
    score_min = 0.0 ; slope_center = 0 ; slope_avg = 0.0 ;

    /* store previous pointer */
    tmp = test_set->vectors ;

    /* create temprorary set to store past distances */
    if (params.stab_thr_type == SPR_STBTHR_SLOPE || 
        params.stab_thr_type == SPR_STBTHR_VARIANCE)
    {
        /* check window size */
        if (params.short_term_win <= 1) params.short_term_win = 9 ; /* default */

        if ((past_dist = fvsCreate(model_num,params.short_term_win))==NULL)
        {
            free(scores) ;
            return 0 ;
        }

        /* add zero distances */
        for (model=1; model<=model_num; model++)
        {
            if ((distances = fvCreate(params.short_term_win))==NULL)
            {
                fvsDestroy(past_dist) ;
                free(scores) ;
                return 0 ;
            }

            memset(distances,0,sizeof(double)*params.short_term_win) ;

            fvsAddVector(past_dist,distances) ;
        }

        if (params.stab_thr_type == SPR_STBTHR_SLOPE)
        {
            slope_center = params.short_term_win / 2 + 1 ;

            slope_avg = 0.0 ;

            /* pre-compute slope formula */
            for(count=1;count<=params.short_term_win;count++)
                slope_avg += (count-slope_center)*(count-slope_center) ;
        }
    }

    /* main cycle untill we have vectors and not all models are pruned */
    while (curr_vector < vector_num && models_pruned < model_num - 1)
    {
        /* check do we have enough vectors */
        if (curr_vector + params.vector_step >= vector_num)
            params.vector_step = vector_num - curr_vector ;

        /* change set size */
        test_set->vector_number = params.vector_step ;

        /* IT IS ASSUMED THAT MSE IS USED => ADD CHECKING !!! */

        score_min = 0.0 ; score_max = 0.0 ; min_model = 0 ;
        mse_mean = 0.0 ;

        /* update scores */
        for (model=1; model<=model_num; model++)
        {
            if (!(msGetModelFlags(ms,model) & SPR_MODEL_PRUNED))
            {
                /* store old score */
                old_score = scores[model-1] ;

                /* update score */
                scores[model-1] = (scores[model-1]*curr_vector + vqMatch(vq,msGetModel(ms,model),test_set)*test_set->vector_number)/
                                 (curr_vector + test_set->vector_number) ;

                /* === check whether score is already stabilized === */

                if (params.stab_thr_type == SPR_STBTHR_SLOPE)
                {
                    distances = fvsGetVector(past_dist,model) ;

                    slope = 0.0 ;

                    /* compute slope */
                    for(count=1;count<=params.short_term_win;count++)
                        slope += (count-slope_center)*distances[count-1]/slope_avg ;

                    /* compare with threshold */
                    if (distances[0] > 0.0) /* SIMPLE ASSUMPTION - REFINE !!! */
                    {
                        if (fabs(slope) <= params.variance_thr)
                            msSetModelFlags(ms,model,SPR_MODEL_STABILIZED) ; /* note: model can't be pruned here */
                    }

                    /* store score for next usage */
                    for(count=1;count < params.short_term_win;count++)
                        distances[count-1] = distances[count] ;

                    distances[params.short_term_win - 1] = scores[model-1] ;

                } else if (params.stab_thr_type == SPR_STBTHR_VARIANCE || 
                           params.stab_thr_type == SPR_STBTHR_STDDEV)
                {
                    distances = fvsGetVector(past_dist,model) ;

                    mean = 0.0 ;

                    /* compute mean */
                    for(count=1;count<=params.short_term_win;count++)
                        mean += distances[count-1]/params.short_term_win ;

                    variance = 0.0 ;

                    /* compute variance */
                    for(count=1;count<=params.short_term_win;count++)
                        variance += (distances[count-1]-mean)*(distances[count-1]-mean)/
                                    (params.short_term_win-1) ;

                    if (params.stab_thr_type == SPR_STBTHR_STDDEV)
                    {
                        /* compute standard deviation */
                        std_dev = sqrt(variance) ;

                        /* match threshold against standard deviation */
                        if (std_dev <= params.variance_thr)
                            msSetModelFlags(ms,model,SPR_MODEL_STABILIZED) ; /* note: model can't be pruned here */

                    } else {

                        /* match threshold against variance */
                        if (variance <= params.variance_thr)
                            msSetModelFlags(ms,model,SPR_MODEL_STABILIZED) ; /* note: model can't be pruned here */
                    }

                    /* store score for next usage */

                } else if (params.stab_thr_type != SPR_STBTHR_NONE) { /* default thresholding */

                    if (fabs(old_score/scores[model-1] - 1) <= params.variance_thr)
                        msSetModelFlags(ms,model,SPR_MODEL_STABILIZED) ; /* note: model can't be pruned here */
                }

                /* find the minimum model */
                if (scores[model-1] < score_min || min_model == 0)
                {
                    score_min = scores[model-1] ;
                    min_model = model ;
                }

                if (params.pr_thr_type == SPR_CFBTHR_MEANSTD)
                {
                    /* compute mean */
                    mse_mean += scores[model-1]/(model_num-models_pruned) ;

                } else { /* default thresholding */

                    if (scores[model-1] > score_max)
                        score_max = scores[model-1] ;    
                    
                }
            }
        }

        /* === update threshold === */
        if (params.pr_thr_type == SPR_CFBTHR_MEANSTD)
        {
            variance = 0.0 ;

            if (model_num-models_pruned > 1)
            {
                /* compute standard deviation */
                for (model=1; model<=model_num; model++)
                    if (!(msGetModelFlags(ms,model) & SPR_MODEL_PRUNED))
                        variance += (scores[model-1] - mse_mean)*(scores[model-1] - mse_mean)/
                                    (model_num-models_pruned-1) ;

                std_dev = sqrt(variance) ;
            
            } else std_dev = 0.0 ;

            pruning_thr = mse_mean + params.pruning_thr*std_dev ;

        } else /* default thresholding */
            pruning_thr = score_min + params.pruning_thr*(score_max - score_min) ;

        /* === prune models === */
        for (model=1; model<=model_num && models_pruned < model_num - 1; model++)
        {
            if (msGetModelFlags(ms,model) & SPR_MODEL_STABILIZED)
            {
                /* check if model score above threshold */
                if (scores[model-1] > pruning_thr && model != min_model)
                {
                    msSetModelFlags(ms,model,SPR_MODEL_PRUNED) ;
                    models_pruned++ ;
                }
            }
        }

        /* move to other interval */
        curr_vector += params.vector_step ;

        /* move pointer to current vector */
        test_set->vectors += params.vector_step ;
        
    }

    /* REMOVE */
    /* printf("%i vectors were used, %i models were pruned\n",curr_vector,models_pruned) ; */

    /* restore pointer */
    test_set->vectors = tmp ;

    /* restore size */
    test_set->vector_number = vector_num ;

    if (ret_score) *ret_score = scores[min_model-1] ;

    /* free allocated resources */
    fvsDestroy(past_dist) ;
    free(scores) ;

    return min_model ;
}

int sprStaticPruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRStatParams * alg_params) 
{
    SPRStatParams params ;

    int model, model_num, vector_num, curr_vector, models_pruned ;
    int min_model, max_model,models_pruned_iter ;

    double score_min,score_max ;

    double * scores ;
    double ** tmp ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        params.step = 20 ;
        params.number = 5 ;

    } else memcpy(&params,alg_params,sizeof(SPRStatParams)) ;

    model_num = msGetSize(ms) ;
    vector_num = fvsGetSize(test_set) ;

    /* allocate buffer for scores */
    if ((scores = malloc(sizeof(double)*model_num))==NULL)
    {
        errno = ENOMEM ;
        return 0 ;
    }

    /* set all scores to zero */
    memset(scores,0,sizeof(double)*model_num) ;

    /* intialize model flags */
    for (model=1; model<=model_num; model++)
        msSetModelFlags(ms,model,0) ;

    models_pruned = 0 ; curr_vector = 0 ; min_model = 0 ; score_min = 0.0 ; 

    /* store previous pointer */
    tmp = test_set->vectors ;

    /* main cycle untill we have vectors and not all models are pruned */
    while (curr_vector < vector_num && models_pruned < model_num - 1)
    {
        /* check do we have enough vectors */
        if (curr_vector + params.step >= vector_num)
            params.step = vector_num - curr_vector ;

        /* change set size */
        test_set->vector_number = params.step ;

        /* IT IS ASSUMED THAT MSE IS USED => ADD CHECKING !!! */

        /* update scores */
        for (model=1; model<=model_num; model++)
        {
            if (!(msGetModelFlags(ms,model) & SPR_MODEL_PRUNED))
            {
                /* update score */
                scores[model-1] = (scores[model-1]*curr_vector + vqMatch(vq,msGetModel(ms,model),test_set)*test_set->vector_number)/
                                 (curr_vector + test_set->vector_number) ;

            }
        }

        models_pruned_iter = 0 ;

        /* prune models */
        while (models_pruned_iter < params.number && models_pruned < model_num - 1)
        {
            score_max = 0.0 ; max_model = 0 ;

            /* find not pruned model with maximum score */
            for (model = 1;model <= model_num; model++)
            {
                if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
                {
                    if (max_model == 0 || scores[model-1] > score_max)
                    {
                        score_max = scores[model-1] ;
                        max_model = model ;
                    }
                }
            }

            /* prune model */
            msSetModelFlags(ms,max_model,SPR_MODEL_PRUNED) ;

            models_pruned_iter++ ;
            models_pruned++ ;
        }

        /* move to other interval */
        curr_vector += params.step ;

        /* move pointer to current vector */
        test_set->vectors += params.step ;
    }

    /* find best speaker */
    score_min = 0.0 ; min_model = 0 ;

    /* find not pruned model with minimum score */
    for (model = 1;model <= model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
        {
            if (min_model == 0 || scores[model-1] < score_min)
            {
                score_min = scores[model-1] ;
                min_model = model ;
            }
        }
    }

    /* restore pointer */
    test_set->vectors = tmp ;

    /* restore size */
    test_set->vector_number = vector_num ;

    if (ret_score) *ret_score = scores[min_model-1] ;

    /* release resources */
    free(scores) ;

    return min_model ;
}

int sprAdaptivePruning(HVQ vq,HMS ms,HFSET test_set,double * ret_score,SPRAdapParams * alg_params) 
{
    SPRAdapParams params ;

    int model, model_num, vector_num, curr_vector, models_pruned ;
    int min_model ;

    double score_min, std_dev, mean, prune_thr ;

    double * scores ;
    double ** tmp ;

    /* check parameters */
    if (ms == NULL || test_set == NULL || 
        msGetSize(ms) == 0 || fvsGetSize(test_set) == 0)
    {
        errno = EINVAL ;
        return 0 ;
    }

    /* copy algorithm parameters */
    if (alg_params == NULL)
    {
        params.step = 20 ;
        params.threshold = 0.2 ;

    } else memcpy(&params,alg_params,sizeof(SPRAdapParams)) ;

    model_num = msGetSize(ms) ;
    vector_num = fvsGetSize(test_set) ;

    /* allocate buffer for scores */
    if ((scores = malloc(sizeof(double)*model_num))==NULL)
    {
        errno = ENOMEM ;
        return 0 ;
    }

    /* set all scores to zero */
    memset(scores,0,sizeof(double)*model_num) ;

    /* intialize model flags */
    for (model=1; model<=model_num; model++)
        msSetModelFlags(ms,model,0) ;

    models_pruned = 0 ; curr_vector = 0 ; min_model = 0 ; score_min = 0.0 ; 

    /* store previous pointer */
    tmp = test_set->vectors ;

    /* main cycle untill we have vectors and not all models are pruned */
    while (curr_vector < vector_num && models_pruned < model_num - 1)
    {
        /* check do we have enough vectors */
        if (curr_vector + params.step >= vector_num)
            params.step = vector_num - curr_vector ;

        /* change set size */
        test_set->vector_number = params.step ;

        /* IT IS ASSUMED THAT MSE IS USED => ADD CHECKING !!! */

        mean = 0.0 ; std_dev = 0.0 ;

        /* update scores */
        for (model=1; model<=model_num; model++)
        {
            if (!(msGetModelFlags(ms,model) & SPR_MODEL_PRUNED))
            {
                /* update score */
                scores[model-1] = (scores[model-1]*curr_vector + vqMatch(vq,msGetModel(ms,model),test_set)*test_set->vector_number)/
                                 (curr_vector + test_set->vector_number) ;

                /* compute mean */
                mean += scores[model-1] / (model_num-models_pruned) ;

            }
        }

        /* compute standard deviation */
        for (model=1; model<=model_num; model++)
            if (!(msGetModelFlags(ms,model) & SPR_MODEL_PRUNED))
                std_dev += (scores[model-1]-mean)*(scores[model-1]-mean)/(model_num-models_pruned-1) ;

        /* compute threshold */
        prune_thr = mean + params.threshold*std_dev ;

        /* prune models */
        for (model = 1;model <= model_num && models_pruned < model_num - 1; model++)
        {
            if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED && scores[model-1] > prune_thr)
            {
                /* prune model */
                msSetModelFlags(ms,model,SPR_MODEL_PRUNED) ;

                models_pruned++ ;
            }
        }

        /* move to other interval */
        curr_vector += params.step ;

        /* move pointer to current vector */
        test_set->vectors += params.step ;
    }

    /* find best speaker */
    score_min = 0.0 ; min_model = 0 ;

    /* find not pruned model with minimum score */
    for (model = 1;model <= model_num; model++)
    {
        if (msGetModelFlags(ms,model) != SPR_MODEL_PRUNED)
        {
            if (min_model == 0 || scores[model-1] < score_min)
            {
                score_min = scores[model-1] ;
                min_model = model ;
            }
        }
    }

    /* restore pointer */
    test_set->vectors = tmp ;

    /* restore size */
    test_set->vector_number = vector_num ;

    if (ret_score) *ret_score = scores[min_model-1] ;

    /* release resources */
    free(scores) ;

    return min_model ;
}

/*======================== Functions ==============================*/
