/*!
\file vq.c
\brief Vector quantization implementation

Implementation of an abstract interface to
vector quantization algorithms.\n
$Id: vq.c,v 1.28 2003/12/19 06:57:35 ekarpov Exp $
\date 2003
\author Evgeny Karpov
*/


#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <time.h>


#include <errorapi.h>
#include <msindex.h>


#include "vq.h"
#include "gla.h"

/*-----------------------------------------------------------------*/


#ifndef max
/*! \brief Resolve maximum of two items */
#define max(a,b)    (((a) > (b)) ? (a) : (b))
/*! \brief Resolve minimum of two items */
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif


/*-----------------------------------------------------------------*/


/*!
\brief Clustering function pointer type

All clustering functions should have this type
of parameters and same return value type.
\param pvq Pointer to quantizer
\param train_set Vector set whose cluster analysis is done
\param num_clusters Number of clusters
\return Cluster centroids
*/
typedef HFSET (*CLUSTFUNC)(
    HVQ pvq,
    HFSET train_set,
    int num_clusters) ;


/*======================== Structures =============================*/

/*! 
\brief Vector quantized speaker modeling algorithms

Collection of data and function pointers needed by
vector quantization speaker modeling and matching.
*/
typedef struct {

    VQParams params ;        /*!< vector quantization params */

    HMSIDX msidx ;           /*!< codebook indexing algorithm */
    CLUSTFUNC clust_func ;   /*!< codebook creation algorithm */

} VQHandle ;

/*======================== Structures =============================*/


HFSET gla_clustering(
    HVQ pvq,
    HFSET train_set,
    int num_clusters) ;


/*======================== Functions ==============================*/

/*!
\brief Create a new vector quantizer

Allocate a new vector quantization abstract handle that
will use given vector quantization algorithm type
and given algorithm parameters.
\param params Vector quantization algorithm used
by the new vector quantizer
\return Pointer to the created vector quantizer
or NULL if creation fails.
\todo Describe in more detail what are quantizer
parameters and algorithm parameters (why
they are separated in such a manner).
\sa ClustAlgType
\sa VQParams
\author Evgeny Karpov
*/
HVQ vqCreate(VQParams * params) 
{
    VQHandle * pvq ;
    GLAParams * gla_params ;

    /* allocate memory for field */
    if ((pvq = malloc(sizeof(VQHandle)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* initialize to zero */
    memset(pvq, 0, sizeof(VQHandle)) ;
    
    /* set default parameters */
    pvq->params.algorithm = VQ_ALG_GLA ;
    pvq->params.flags = VQ_INDEX_QUANTIZE ;
    pvq->params.dist_func = dfEuclidean ;
    pvq->params.dfunc_params = NULL ;
    pvq->params.index_type = VQ_IDX_VPT ;
    pvq->params.err_func = errMSE ;
    pvq->params.idxerr_func = idxMSE ;
    pvq->params.gla_num_initial = 3 ;

    /* copy parameters if they are set */
    if (params != NULL)
    {
        pvq->params.algorithm = params->algorithm ;
        pvq->params.flags = params->flags ;
        pvq->params.index_type = params->index_type ;

        if (params->gla_num_initial != 0)
            pvq->params.gla_num_initial = params->gla_num_initial ;

        if (params->dist_func)
            pvq->params.dist_func = params->dist_func ;

        if (pvq->params.index_type == VQ_IDX_NONE)
            pvq->params.err_func = params->err_func ;
        else
            pvq->params.idxerr_func = params->idxerr_func ;

    }

    /* set algorithm specific pointers and parameters */
    if (pvq->params.algorithm == VQ_ALG_GLA)
    {
        /* GLA specific parameters */
        if ((gla_params = malloc(sizeof(GLAParams)))==NULL)
        {
            free(pvq) ;
            ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
            return NULL ;
        }

        pvq->clust_func = gla_clustering ;
        pvq->params.algorithm_params = gla_params ;

        if (params != NULL && params->algorithm_params != NULL)
        {
            /* copy parameters if they are set */
            memcpy(gla_params,params->algorithm_params,sizeof(GLAParams)) ;

        } else {

            /* set default parameters */
            gla_params->max_iteration = 100 ;
            gla_params->threshold = 1e-05 ;
            gla_params->flags = GLA_DISTANCE_FUNC   | 
                                GLA_ERROR_FUNC      | 
                                GLA_FERROR_FUNC ;

             gla_params->dist_func = pvq->params.dist_func ;

            /* copy function pointers */
            if (pvq->params.flags & VQ_INDEX_QUANTIZE)
            {
                gla_params->flags |= GLA_USE_INDEXING ;
                gla_params->idxerr_func = pvq->params.idxerr_func ;

            } else {

                gla_params->err_func = pvq->params.err_func ;
            }

        }

    } else {

        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    /* create indexer if needed */
    if (pvq->params.index_type == VQ_IDX_VPT)
    {
        pvq->msidx = msidxCreate(MSINDEX_ALGTYPE_VPT) ;
    }

    return pvq ;
}

/*!
\brief Destroy a vector quantizer

Deallocate resources allocated to a given
vector quantizer
\param vq Pointer to vector quantizer that will be destroyed
\sa HVQ
\author Evgeny Karpov
*/
void vqDestroy(HVQ vq) 
{
    VQHandle * pvq = (VQHandle*) vq;

    if (pvq == NULL) return ;

    /* destroy indexer */
    msidxDestroy(pvq->msidx) ;

    free(pvq->params.algorithm_params) ;
    free(pvq) ;
}

/*!
\brief Destroy a vector quantized speaker model

Deallocate resources allocated to a given
vector quantized model
\param vq Pointer to vector quantizer that was used for model creation
\param vqmodel Pointer to vector quantized speaker model 
that will be destroyed
\sa HVQ
\author Evgeny Karpov
*/
void vqDestroyModel(HVQ vq, HVQMODEL vqmodel) 
{
    if (vqmodel)
    {
        fvsDestroy(vqmodel->codebook) ;

        /* destroy indexes using appropriate indexer */
        if (vq) msidxDestroyIndexes(((VQHandle*)vq)->msidx,vqmodel->indexes) ;
    }

    free(vqmodel) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Train a vector quantized speaker model

Use a given training data set to adjust
the codebook of the vector quantized speaker model.
\param vq Vector quantizer 
\param train_data Input data for training
\param num_clusters Number of clusters in quantized data
\return Handle to speaker model
\todo Explain in detail what happens
when a vector quantized speaker model is trained and
how/why it is trained (for using in classification?)
\sa HVQ
\sa HFSET
\author Evgen Karpov
*/
HVQMODEL vqTrain(HVQ vq, HFSET train_data,int num_clusters) 
{
    VQHandle * pvq = (VQHandle*) vq;
    VQModel * vqmodel ;

    /* check paprameters */
    if (pvq == NULL || train_data == NULL || 
        num_clusters <= 0 || num_clusters > fvsGetSize(train_data))
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return NULL ;
    }

    if ((vqmodel = malloc(sizeof(VQModel)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    /* initialize to zero */
    memset(vqmodel,0,sizeof(VQModel)) ;

    /* train model */
    if ((vqmodel->codebook = pvq->clust_func(vq,train_data,num_clusters)) == NULL)
    {
        free(vqmodel) ;
        ReportError(ESR_MODEL_CONSTR_FAIL,NULL) ;
        return NULL ;
    }

    /* create indexes if needed */
    if (pvq->msidx)
    {
        vqmodel->indexes = msidxTrain(pvq->msidx,vqmodel->codebook,
                                      pvq->params.dist_func, pvq->params.dfunc_params) ;
    }

    return vqmodel ;
}

/*!
\brief Match a vector set against a vector quantized speaker model

Compute the error measure for given feature vector
vector set with respect to a vector quantized speaker model.
\param vq Vector quantizer
\param vqmodel Vector quantized speaker model
\param test_data Feature vector set whose matching is done
\return Non-negative value of the matching error
or -1 in case of failure.
\author Evgeny Karpov
*/
double vqMatch(HVQ vq, HVQMODEL vqmodel, HFSET test_data) 
{
    VQHandle * pvq = (VQHandle*) vq;

    /* check paprameters */
    if (pvq == NULL || test_data == NULL || vqmodel == NULL)
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return -1.0 ;
    }

    /* compute matching score  */
    if (pvq->params.index_type != VQ_IDX_NONE)
    {
        return pvq->params.idxerr_func(
                 vqmodel->codebook, test_data,
                 pvq->msidx, vqmodel->indexes) ;
    }
    else
    {
        return pvq->params.err_func(
                 vqmodel->codebook, test_data, pvq->params.dist_func,
                 pvq->params.dfunc_params) ;
    }

}

/*-----------------------------------------------------------------*/

/*!
\brief Load vector set from a file to a speaker model codebook

Read a stored vector set from a file to a vector
quantized speaker model's codebook and create
an index for the codebook if needed.
\param vq Vector quantizer
\param file_name Name of the file where data is stored
\return Pointer to loaded speaker model
\sa fvsLoad
\sa VQHandle
\author Evgeny Karpov
*/
HVQMODEL vqLoadModel(HVQ vq, char * file_name) 
{
    VQHandle * pvq = (VQHandle*) vq;
    VQModel * vqmodel ;

    if ((vqmodel = malloc(sizeof(VQModel)))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return NULL ;
    }

    vqmodel->indexes = NULL ;

    /* load feature vectors */
    if ((vqmodel->codebook = fvsLoad(file_name))==NULL)
    {
        free(vqmodel) ;
        ReportError(ESR_CANT_LOAD_FILE,NULL) ;
        return NULL ;
    }

    /* train indexes if needed */
    if (pvq && pvq->params.index_type != VQ_IDX_NONE)
    {
        vqmodel->indexes = msidxTrain(pvq->msidx,vqmodel->codebook,
                                      pvq->params.dist_func, 
                                      pvq->params.dfunc_params) ;

        if (vqmodel->indexes == NULL)
        {
            vqDestroyModel(NULL,vqmodel) ;
            return NULL ; /* error should be reported before */
        }
    }

    return vqmodel ;
}

/*!
\brief Save vector quantized speaker model in a file

\param vq Vector quantizer
\param vqmodel Vector quantized speaker model
\param file_name File where model is stored
\return ESR_SUCCEED for successful processing 
or ESR_FAILED in failure.
\todo NOT IMPLEMENTED YET BECAUSE FORMAT IS NOT DEFINED
\sa fvsSave
\author Evgeny Karpov
*/
int vqSaveModel(HVQ vq, HVQMODEL vqmodel, char * file_name) 
{
    if (vqmodel == NULL) 
    {
        ReportError(ESR_INVALID_PARAMETER,NULL) ;
        return ESR_FAILED ;
    }

    return fvsSave(vqmodel->codebook,file_name) ;
}

/*-----------------------------------------------------------------*/

/*!
\brief Pick random vectors from feature vector set into codebook

Select vector randomly from a feature vector set and
copy them to a codebook.
\param cb Codebook where to put randomly selected vectors
\param fvs Feature vector set where vecotr are selected from
\param num How many vectors are randomly selected
\return ESR_SUCCEED or ESR_FAILED depending on
the outcome of the processing.
\bug If the number of random vectors requested
is bigger than number of vectors where to pick from
then this is not exactly copying of random vectors.
\todo Check that the number of requested vectors
is a lot less than number of vectors where to select from.
\author Evgeny Karpov
*/
int copy_random_vectors(HFSET cb,HFSET fvs,int num) 
{
    int * used_vectors ;
    int vector_num,rand_vect, set_size ;
    int i ;

    if ((used_vectors = malloc(sizeof(int)*num))==NULL)
    {
        ReportError(ESR_NOT_ENOUGH_MEMORY,NULL) ;
        return ESR_FAILED ;
    }

    memset(used_vectors,0,sizeof(int)*num) ;

    /* seed the random-number generator with current time */
    srand((unsigned)time(NULL));
    
    set_size = fvsGetSize(fvs) ;

    rand_vect = 0 ; /* just to avoid compiler warnings */

    if (set_size>num)
    {
        for (vector_num=1;vector_num<=num;vector_num++)
        {
            /* find random but unique vector from set */
            while (used_vectors[vector_num-1]==0)
            {
                /* random number */
                rand_vect = rand()%set_size+1 ;

                /* check if this vector already exist in set */
                for (i=0;i<vector_num-1;i++)
                    if (used_vectors[i]==rand_vect) break ;

                /* copy vector number if it is not present in set already */
                if (i==vector_num-1) 
                    used_vectors[vector_num-1] = rand_vect ;
            }

            /* copy vector */
            fvsAddVector(cb,fvsCopyVector(fvs,rand_vect)) ;
        }

    } else {


        /* copy all vectors first */
        for (vector_num=1;vector_num<=set_size;vector_num++)
            fvsAddVector(cb,fvsCopyVector(fvs,vector_num)) ;

        /* copy random vectors */
        for (vector_num=set_size + 1;vector_num<=num;vector_num++)
        {
            /* random number */
            rand_vect = rand()%set_size+1 ;

            /* copy vector */
            fvsAddVector(cb,fvsCopyVector(fvs,rand_vect)) ;
        }
    }

    free(used_vectors) ;

    return ESR_SUCCEED ;
}

/*!
\brief Vector quantization using GLA algorithm

Performs clustering using GLA and several initial
random solutions.
\param pvq Vector quantizer handle
\param train_set Feature vector set used for training
\param num_clusters The size of the resulting codebook
\return Pointer to speaker codebook or NULL in case of error.
\author Evgeny Karpov
*/
HFSET gla_clustering(
    HVQ pvq,
    HFSET train_set,
    int num_clusters) 
{
    HFSET ret_cb, tmp_cb ;
    int solution ;
    double best_err, curr_err ;
    int result;
    VQParams* vq;

    if ((NULL == pvq) || (NULL == train_set)
          || (num_clusters <= 2))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }
    vq = (VQParams*) pvq;

    /*
    if ((pvq->params.gla_num_initial < 1)
    */
    if ((vq->gla_num_initial < 1)
          || (fvsGetVectorSize(train_set) < 1)
          || (fvsGetSize(train_set) <= num_clusters))
    {
        ReportError(ESR_INVALID_PARAMETER, NULL);
        return NULL;
    }

    /* apply GLA algorithm */
    for (solution = 1, best_err = DBL_MAX, ret_cb = NULL;
         /*
         solution <= pvq->params.gla_num_initial;
         */
         solution <= vq->gla_num_initial;
         solution++)
    {
        tmp_cb = fvsCreate(num_clusters, fvsGetVectorSize(train_set)) ;
        if (tmp_cb == NULL)
        {
            ReportError(ESR_VECTOR_SET_CREATE, NULL);
            return NULL ;
        }

        /* generate random solution first */
        result = copy_random_vectors(tmp_cb, train_set, num_clusters) ;
        if (ESR_FAILED == result)
        {
            ReportError(ESR_GLA_COPY_RANDOM_VECTORS, NULL);
            fvsDestroy(tmp_cb);
            if (ret_cb)
            {
                fvsDestroy(ret_cb) ;
            }
            return NULL;
        }

        /* apply GLA to this solution */
        result = glaRun(tmp_cb, train_set,
                        /*
                        pvq->params.algorithm_params, &curr_err) ;
                        */
                        vq->algorithm_params, &curr_err) ;
        if (0 == result)
        {
            ReportError(ESR_GLA_RUN, NULL);
            fvsDestroy(tmp_cb);
            if (ret_cb)
            {
                fvsDestroy(ret_cb) ;
            }
            return NULL;
        }

        /* check wheter new solution is better */
        if (solution==1 || best_err > curr_err)
        {
            best_err = curr_err ;

            if (ret_cb)
            {
                fvsDestroy(ret_cb) ;
            }

            ret_cb = tmp_cb ;

        }
        else
        {
            fvsDestroy(tmp_cb) ;
        }
    }

    return ret_cb ;
}

/*======================== Functions ==============================*/
