/*! 
\file 
\brief Vector quantization interface

\date 2003
\author Evgeny Karpov
*/

/*
    $Id: vq.h,v 1.15 2003/11/20 11:48:39 ekarpov Exp $
*/

#ifndef __VQ_H__
#define __VQ_H__

/*-----------------------------------------------------------------*/

#include <fvector.h>

#include "distfunc.h"
#include "errfunc.h"
#include "idxerrfunc.h"
    
/*-----------------------------------------------------------------*/

#if defined(__cplusplus)
extern "C" {
#endif

/*======================== Structures =============================*/

/*! 
\brief Type of clustering algorithm
*/
typedef enum {

    VQ_ALG_GLA = 1, /*!< generalized Lloyd algorithm */
    VQ_ALG_RLS      /*!< randomized local search (not implemented yet) 
                            \todo Implement randomized local search */

} ClustAlgType ;

/*! 
\brief Type of indexing algorithm for VQ
*/
typedef enum {

    VQ_IDX_NONE = 0,    /*!< No indexing */
    VQ_IDX_VPT = 1      /*!< Vantage Point Tree */

} IdxAlgType ;


/*!
\brief Parameters of vector quantization speaker modeling

This is collection of the data needed
in vector quantization process for speaker modeling.
 */
typedef struct {

    ClustAlgType algorithm ;  /*!< quantization algorithm */
    void * algorithm_params ; /*!< quantization algorithm parameters */

    IdxAlgType index_type ;   /*!< indexing type */

    DISTFUNC dist_func ;      /*!< distance function */
    void * dfunc_params ;     /*!< distance function parameters */

    ERRFUNC  err_func ;      /*!< normal error function */
    IDXERRFUNC idxerr_func ; /*!< fast error function */

    int flags ; /*!< flags */

    /* GLA algorithm specific parameters */
    int gla_num_initial ; /*!< number of different randomly picked
                               vector sets used with GLA clustering
                               as initial solutions */
    /* HFSET gla_init_cb ;*/   /* initial codebook */

} VQParams ;

/*! 
\brief Vector quantization model
*/
typedef struct {

    HFSET    codebook ;    /*!< codevectors */
    HMSINDEX indexes ;     /*!< indexes */

} VQModel ;

/*======================== Structures =============================*/

#define VQ_INDEX_QUANTIZE    1 /* use indexing during quantization */

/*-----------------------------------------------------------------*/

/*!
\brief Abstract handle to vector quantization speaker model
\sa VQParams
\sa VQHandle
 */
typedef VQModel * HVQMODEL ;   /* model */

/*!
\brief Abstract handle to vector quantization parameters
\sa VQParams
\sa VQHandle
 */
typedef void * HVQ ;        /* matcher/quantizer */

/*======================== Functions ==============================*/

/* creation/deletion */
HVQ vqCreate(VQParams * params) ;
void vqDestroy(HVQ vq) ;
void vqDestroyModel(HVQ vq, HVQMODEL vqmodel) ;

/* quantization/matching */
HVQMODEL vqTrain(HVQ vq, HFSET train_data, int num_clusters) ;
double   vqMatch(HVQ vq, HVQMODEL vqmodel, HFSET test_data) ;

/* save/load */
HVQMODEL vqLoadModel(HVQ vq, char * file_name) ;
int vqSaveModel(HVQ vq, HVQMODEL vqmodel, char * file_name) ;

/*======================== Functions ==============================*/

#if defined(__cplusplus)
}
#endif

#endif /* __VQ_H__ */
